(*
    Title: 	MathLib - Implementation
    LastEdit:	"Tue Sep 11 12:06:49 1984"
    Author: 	jd
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE MathLib;

(* These routines are transcribed from the VAX C versions. A restriction
   introduced by the transcription is that sin and cos will not work for
   real numbers which are outside INTEGER range. This should be fixed by
   providing the machine dependent C routine modf to grub around in the
   representation of real numbers. This library is machine-independent
   with the addition of the range restriction
*)

FROM Exceptions IMPORT RAISE;

CONST
    SinCosArgTooLarge = "SinCosArgTooLarge";
    ExpOverflow = "ExpOverflow";

VAR degreestoradians: REAL;
    epow2, TwoPow : ARRAY [-6..6] OF REAL ;


PROCEDURE Radians (degrees: REAL): REAL;
BEGIN
    RETURN degrees * degreestoradians;
END Radians;


PROCEDURE Degrees (radians: REAL): REAL;
BEGIN
    RETURN radians / degreestoradians;
END Degrees;

MODULE SinCos;
    IMPORT RAISE, SinCosArgTooLarge;
    EXPORT Sin, Cos;

    CONST twoopi    =  0.63661977236758134308;
    CONST p0        =  0.1357884097877375669092680E8;
    CONST p1        = -0.4942908100902844161158627E7;
    CONST p2        =  0.4401030535375266501944918E6;
    CONST p3        = -0.1384727249982452873054457E5;
    CONST p4        =  0.1459688406665768722226959E3;
    CONST q0        =  0.8644558652922534429915149E7;
    CONST q1        =  0.4081792252343299749395779E6;
    CONST q2        =  0.9463096101538208180571257E4;
    CONST q3        =  0.1326534908786136358911494E3;


    PROCEDURE Sinus(arg: REAL; quad: CARDINAL): REAL;
        VAR e, f, ysq, x, y, temp1, temp2: REAL;
        VAR k: CARDINAL;
    BEGIN
        x := arg;
        IF x < 0.0 THEN
            x := -x;
            quad := quad + 2;
        END;
        x := x * twoopi;    (* underflow? *)
        IF ABS (x - FLOAT (TRUNC (x))) > 1.0 THEN
            RAISE(SinCosArgTooLarge);
        END (* if *);
        k := TRUNC (x);
        y := x - FLOAT (k);
        quad := (quad + k) MOD 4;
        IF ODD (quad) THEN  y := 1.0 - y; END;
        IF quad > 1 THEN y := 0.0 - y; END;

        ysq := y*y;
        temp1 := ((((p4*ysq+p3)*ysq+p2)*ysq+p1)*ysq+p0)*y;
        temp2 := ((((ysq+q3)*ysq+q2)*ysq+q1)*ysq+q0);
        RETURN temp1/temp2;
    END Sinus;


    PROCEDURE Cos (arg: REAL): REAL;
    BEGIN
        IF arg < 0.0 THEN arg := 0.0 - arg END;
        RETURN Sinus (arg, 1);
    END Cos;


    PROCEDURE Sin (arg: REAL): REAL;
    BEGIN
        RETURN Sinus (arg, 0);
    END Sin;

END SinCos;


MODULE ArcTan;

    EXPORT Atan, Atan2;

    (* atan coefficients are #5077 from Hart & Cheney. (19.56D) *)

    CONST sq2p1  = 2.414213562373095048802e0;
    CONST sq2m1  = 0.414213562373095048802e0;
    CONST pio2   = 1.570796326794896619231e0;
    CONST pio4   = 0.785398163397448309615e0;
    CONST p4     = 0.161536412982230228262e2;
    CONST p3     = 0.26842548195503973794141e3;
    CONST p2     = 0.11530293515404850115428136e4;
    CONST p1     = 0.178040631643319697105464587e4;
    CONST p0     = 0.89678597403663861959987488e3;
    CONST q4     = 0.5895697050844462222791e2;
    CONST q3     = 0.536265374031215315104235e3;
    CONST q2     = 0.16667838148816337184521798e4;
    CONST q1     = 0.207933497444540981287275926e4;
    CONST q0     = 0.89678597403663861962481162e3;

    (* xatan evaluates a series valid in the range [-0.414...,+0.414...]. *)

    PROCEDURE xatan(arg: REAL): REAL;
        VAR argsq, value: REAL;
    BEGIN
        argsq := arg * arg;
        value := ((((p4 * argsq + p3) * argsq + p2) * argsq + p1) *
            argsq + p0);
        value := value / (((((argsq + q4) * argsq + q3) *
            argsq + q2) * argsq + q1) * argsq + q0);
        RETURN value*arg;
    END xatan;

    (* satan reduces its positive argument to the range [0,0.414...]
       and calls xatan.
    *)

    PROCEDURE satan (arg: REAL): REAL;
    BEGIN
        IF arg < sq2m1 THEN
            RETURN xatan(arg);
        ELSIF arg > sq2p1 THEN
            RETURN pio2 - xatan(1.0 / arg);
        ELSE RETURN pio4 + xatan ((arg - 1.0) / (arg + 1.0));
        END;
    END satan;

    (* atan makes its argument positive and calls satan. *)

    PROCEDURE Atan (arg: REAL): REAL;
    BEGIN
        IF arg > 0.0 THEN
            RETURN satan (arg);
        ELSE RETURN 0.0 - satan (0.0 - arg);
        END;
    END Atan;

    
    (* atan2 discovers what quadrant the angle is in and calls atan. *)

    PROCEDURE Atan2 (arg1, arg2: REAL): REAL;
    BEGIN
        IF arg2 = 0.0  THEN
            IF arg1 >= 0.0 THEN
                 RETURN pio2;
            ELSE RETURN 0.0 - pio2;
            END;
        ELSIF arg2 < 0.0 THEN
            IF arg1 >= 0.0 THEN
                RETURN pio2 + pio2 - satan(0.0 - arg1/arg2);
            ELSE RETURN 0.0 - pio2 - pio2 + satan(arg1/arg2);
            END;
        ELSIF arg1 > 0.0 THEN
            RETURN satan(arg1/arg2);
        ELSE RETURN 0.0 - satan(0.0 - arg1/arg2);
        END;
    END Atan2;


END ArcTan;

                 
PROCEDURE Exp (x : REAL) : REAL ;
(* Not the quickest code, but simple *)
VAR Power : INTEGER ;
    Reciprocate : BOOLEAN ;
    Result : REAL ;
    BEGIN
        IF x > 88.029
        THEN RAISE(ExpOverflow);
        ELSIF x < -88.0
        THEN RETURN 0.0
        ELSE
             IF x > 0.0
             THEN Reciprocate := FALSE
             ELSIF x = 0.0
             THEN RETURN 1.0
             ELSE Reciprocate := TRUE ;
                  x := - x
             END (* if *) ;
             
             Result := 1.0 ;
             
             FOR Power := 6 TO -6 BY -1
             DO
                IF x > TwoPow [Power]
                THEN x := x - TwoPow [Power] ;
                     Result := Result * epow2 [Power]
                END (* if *)
             END (* for *);
             
             (****   x is now less than 2 ** -6
                     so terms in the power series 
                     to x ** 4 will suffice         
             ****)
             
             Result := Result *
                         (1.0 + x *
                            (1.0 + x *
                               (0.5 + x *
                                  (0.16666667 + x *
                                     0.041666667)))) ;
                                     
             IF Reciprocate
             THEN RETURN 1.0 / Result
             ELSE RETURN Result
             END (* if *)
             
       END (* if *)
    END Exp ;
                 


BEGIN
    degreestoradians := (2.0 * Pi) / 360.0;
        
    (* epow2 [r] := exp (2 ** r) *)

   epow2 [-6] := 1.01574771E+00 ;
   epow2 [-5] := 1.03174341E+00 ;
   epow2 [-4] := 1.06449446E+00 ;
   epow2 [-3] := 1.13314845E+00 ;
   epow2 [-2] := 1.28402542E+00 ;
   epow2 [-1] := 1.64872127E+00 ;
   epow2 [0] := 2.71828183E+00 ;
   epow2 [1] := 7.38905610E+00 ;
   epow2 [2] := 5.45981500E+01 ;
   epow2 [3] := 2.98095799E+03 ;
   epow2 [4] := 8.88611052E+06 ;
   epow2 [5] := 7.89629602E+13 ;
   epow2 [6] := 6.23514908E+27 ; 

   (* TwoPow [r] := 2 ** r *)

   TwoPow [-6] := 0.015625 ;
   TwoPow [-5] := 0.031250 ;
   TwoPow [-4] := 0.062500 ;
   TwoPow [-3] := 0.125000 ;
   TwoPow [-2] := 0.250000 ;
   TwoPow [-1] := 0.500000 ;
   TwoPow [0] :=   1.000000 ;
   TwoPow [1] :=   2.000000 ;
   TwoPow [2] :=   4.000000 ;
   TwoPow [3] :=   8.000000 ;
   TwoPow [4] :=  16.000000 ;
   TwoPow [5] :=  32.000000 ;
   TwoPow [6] :=  64.000000 

    
END MathLib.
