(*
    Title: 	SCTermStream - Implementation
    LastEdit:	"Tue Oct 14 10:03:50 1986"
    Author: 	jd/mjj
		Cambridge University Computer Laboratory
                Acorn Research Centre

    Unix version
*)

IMPLEMENTATION MODULE SCTermStream;
(* $T-,$R- *)

FROM SYSTEM IMPORT WORD, ADR, UNIXCALL;
FROM Exceptions IMPORT RAISE;
FROM UnixCalls IMPORT stty, gtty, read, write;
FROM UnixMCI IMPORT errnoPtr;
FROM StreamRep IMPORT 
	DefaultProcs, SetStreamError, Procs, NullFlush;
FROM Streams IMPORT
        Create, Stream, StreamProcs, ErrorCode, OptionalFunction;
FROM CharCodes IMPORT
    CReturnCh, NewLineCh;


(******* Unix data structure used for stty, gtty - see man 4 tty **********)

TYPE SCBIT = (TANDEM, CBREAK, LCASE, ECHO, CRMOD, RAW);
    (* full definition of these bits in man 4 tty *)

TYPE SCMODE = SET OF SCBIT;

(* The following constants define what are probably the most useful settings
   of the terminal driver bits. You may of course create your own new and
   exciting combinations.
*)

CONST
    NOBUFF      = SCMODE {CBREAK, CRMOD};
                    (* like ordinary input except that no buffering is done.
                       CR is translated into LF, flow control is done, the
                       interrupt and kill characters work, but no reflection.
                    *)
    NOTRANS     = SCMODE {CBREAK};
                    (* as NOBUFF, but no translation of CR into LF is done on
                       input, and no translation of CR or LF into CR-LF on
                       output, so watch it if you try to type one of these two
                       when a stream of this mode is selected for input!
                    *)
    HANDSOFF    = SCMODE {RAW};
                    (* you are given all 8 bits with no interpretation at all.
                       This is the most useful mode, except that it has the
                       disadvantage that you can no longer kill a process from
                       the terminal. Terminal output is also 8-bits with no
                       translations whatsoever.
                    *)
    
    REFLECT     = SCMODE {ECHO};
                    (* a modifier which can be used with any mode to turn on
                       reflection: e.g. NOTRANS + REFLECT
                    *)

(* This code depends on the fact that an SCMODE set is aligned on a word in
   the same way as a BITSET, covering all WORDSIZE bits of the C structure
*)


TYPE
    SGTTYB = RECORD
        chars:          ARRAY [0..3] OF CHAR;
        (* respectively sgispeed, sgospeed, sgerase, sgkill *)
        sgflags:        SCMODE;
    END;
    
    CONST Mask = SCMODE {TANDEM, CBREAK, LCASE, ECHO, CRMOD, RAW};


(**************************************************************************)

(*
   To make this happen, an "stty" call to Unix is made to cause a
   change in the terminal driver behaviour. This will cause other
   terminal input streams to behave oddly, and you may not open
   another stream with different characteristics.			  
*)

VAR
    status: RECORD  (* only one instance, so data can be global *)
        CASE opened: BOOLEAN OF
          TRUE: instance: Stream;
                translateCRToNL: BOOLEAN;
                terminalWas: SCMODE; (* previous values of the bits *)
                bspPending: BOOLEAN;
                lastChar: CHAR;
        | FALSE:
        END (* case *);
        procs: Procs;
    END (* record *);


PROCEDURE TranslateCR(yes: BOOLEAN);
    BEGIN
        status.translateCRToNL := yes;
    END TranslateCR;


PROCEDURE SetFlags (s: Stream; value: SCMODE): SCMODE;
    VAR sgttyb: SGTTYB;
    VAR was: SCMODE;
    VAR rc: INTEGER;
BEGIN
    rc := UNIXCALL (gtty, 0, ADR (sgttyb)); (* 0 = terminal input *)
    IF rc = 0 THEN
        WITH sgttyb DO
            was := sgflags * Mask;
            sgflags := (sgflags - Mask) + value;
        END;
        rc := UNIXCALL (stty, 0, ADR (sgttyb));
    END;
    IF rc # 0 THEN
        SetStreamError(s, -errnoPtr^);
    END;
    RETURN was;
END SetFlags;

PROCEDURE Implements(s: Stream; f: OptionalFunction): BOOLEAN;
BEGIN
    RETURN FALSE;
END Implements;

PROCEDURE Delete(s: Stream);
VAR
    void: SCMODE;
BEGIN
    void := SetFlags (s, status.terminalWas);
END Delete;

PROCEDURE Get(s: Stream): WORD;
VAR
    rc: INTEGER;
    buf: ARRAY [0..0] OF CHAR;
BEGIN
    IF status.bspPending THEN
        status.bspPending := FALSE;
        RETURN WORD(status.lastChar)
    ELSE
    END (* if *);
    rc := UNIXCALL (read, 0, ADR (buf), 1);
    IF rc # 1 THEN
        IF rc < 0 THEN
            SetStreamError(s, -errnoPtr^);
        ELSE SetStreamError(s,  -rc);
        END;
        s^.status := StreamError;
        RETURN WORD (0);
    ELSE
        IF status.translateCRToNL AND (buf [0] = CReturnCh) THEN
            RETURN WORD (NewLineCh)
        ELSE RETURN WORD (buf [0]) END;
    END;
END Get;

PROCEDURE Backspace(s: Stream);
BEGIN
    status.bspPending := TRUE;
END Backspace;

PROCEDURE Put(s: Stream; arg: WORD);
VAR 
    rc, length: INTEGER;
    ch: CHAR;
    buf: ARRAY [0..1] OF CHAR;
BEGIN
    ch := CHAR (arg);
    IF status.translateCRToNL AND (ch=NewLineCh) THEN
        buf [0] := CReturnCh;
        buf [1] := NewLineCh;
        length := 2;
    ELSE
        buf [0] := ch;
        length := 1;
    END;
    rc := UNIXCALL (write, 1, ADR (buf), length);
    IF rc # length THEN
        IF rc < 0 THEN
            SetStreamError(s, -errnoPtr^);
        ELSE SetStreamError(s, -rc);
        END;
        s^.status := StreamError;
    END;
END Put;
    
PROCEDURE CreateInOutput(): Stream;
BEGIN
  WITH status DO
    IF NOT opened THEN
        instance := Create(StreamProcs(procs), 0);
        terminalWas := SetFlags (instance, NOTRANS);
        bspPending := FALSE;
        lastChar := 0C;
        opened := TRUE;
        
        (* To  avoid interfering with the buffered terminal stream *)
        (* in particular destroying the CR/NL mapping, we tell the *)
        (* other stream to do its own as we have just told the     *)
        (* system not to. DRM 14/10/86                             *)
        
        
    END (* if *);
    RETURN instance;
  END (* with *);
END CreateInOutput;


PROCEDURE StreamErrorCode(s: Stream): INTEGER;
BEGIN
    IF Procs(s^.procs) = status.procs THEN
        RETURN INTEGER(s^.streamError)
    ELSE
        RAISE("WrongClass");
    END (* if *);
END StreamErrorCode;

BEGIN
  WITH status DO
    procs := DefaultProcs(8);
    procs^.Get := Get; procs^.Put := Put;
    procs^.Flush := NullFlush;
    procs^.Implements := Implements; 
    procs^.Delete := Delete;
    procs^.Backspace := Backspace;
    opened := FALSE;
    translateCRToNL := TRUE;
  END (* with *);
END SCTermStream.
