(*
    Title: 	PageStream - Implementation
    LastEdit:	"Tue Sep 25 12:44:21 1984"
    Author: 	jd/mjj
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE PageStream;

FROM SYSTEM IMPORT WORD;
FROM SystemTypes IMPORT LongINTEGER, LongCARDINAL;

FROM Arithmetic IMPORT IMin, IMax;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM CharCodes IMPORT FormFeedCh, SpaceCh, TabCh, NewLineCh;
    

FROM Streams IMPORT 
    Create, StreamProcs, ErrorCode, SeekMode, Stream, OptionalFunction;
FROM StreamRep IMPORT
	DefaultProcs, SetStreamError, Procs;

CONST MaxWidth = 150;

VAR procs: Procs;

TYPE
    PAGEFUNCTION = (LineSet, PageSet);

    BUFFER = POINTER TO BUFFERDATA;
    
    BUFFERDATA = RECORD
        S:                        Stream;
        B:                        ARRAY [1..MaxWidth] OF CHAR;
        PageWidth, PageLength,
        Hwm, Char, Line, Page:    CARDINAL;
        Tab, Fold,
        PagePending, LinePending: BOOLEAN;
        LineProc, PageProc:       EventProc;
        stream: Stream;
    END;

CONST TabSize = 8;   (* ie at columns 1, 9, 17... *)
    NewLine = FALSE;
    NewPage = TRUE;
    
PROCEDURE Flush (b: BUFFER; page: BOOLEAN);
    VAR i, j, blanks: CARDINAL;
    Sprocs: Procs;
BEGIN
    WITH b^ DO
        INC (Hwm);
        IF page THEN 
            B[Hwm] := FormFeedCh;
            INC (Hwm);
            B [Hwm] := NewLineCh;
            Line := 1;
            INC (Page);
            PagePending := TRUE;
        ELSE B[Hwm] := NewLineCh;
            INC (Line);
            LinePending := TRUE;
        END;
        Sprocs := Procs(S^.procs);
        IF Tab THEN
            blanks := 0;
            FOR i := 1 TO Hwm DO
                IF (i MOD TabSize = 1) AND (blanks > 0) THEN
                    Sprocs^.Put(S, TabCh);
                    blanks := 0;
                END;
                IF B[i] = SpaceCh THEN
                    INC (blanks);
                ELSE IF blanks > 0 THEN
                        FOR j := 1 TO blanks DO 
                            Sprocs^.Put(S, SpaceCh);
                        END;
                        blanks := 0;
                    END;
                    Sprocs^.Put(S, B[i]);
                END;
            END;
        ELSE FOR i := 1 TO Hwm DO
                Sprocs^.Put(S, B[i]);
            END;
        END;
        FOR i := 1 TO Hwm  DO B[i] := SpaceCh; END;
        Char := 1;
        Hwm := 0;
    END;
END Flush;

PROCEDURE SFlush(s: Stream);
VAR
    b: BUFFER;
BEGIN
    b := BUFFER(s^.instanceData);
    Flush(b, NewLine);
END SFlush;

PROCEDURE Implements(s: Stream; f: OptionalFunction): BOOLEAN;
BEGIN
    RETURN (f=OptSeek) OR (f=OptTell);
END Implements;

PROCEDURE Put(s: Stream; arg: WORD);
VAR
    b: BUFFER;
    c: CHAR;
BEGIN
    b := BUFFER(s^.instanceData);
            WITH b^ DO
                c := CHAR (arg);
                IF PagePending THEN
                    PagePending := FALSE;
                    PageProc (s, Page);
                END;
                IF LinePending THEN
                    LinePending := FALSE;
                    LineProc (s, Line);
                END;
                IF c = FormFeedCh THEN
                    Flush (b, NewPage);
                ELSIF c = NewLineCh THEN
                    IF Line > PageLength THEN
                        Flush (b, NewPage);
                    ELSE Flush (b, NewLine);
                    END;
                ELSIF Char > PageWidth THEN
                    IF Fold THEN
                        IF Line > PageLength THEN
                            Flush (b, NewPage);
                        ELSE Flush (b, NewLine);
                        END;
                        Put(s, arg);
                    END;
                ELSIF c = TabCh THEN
                    WHILE Char MOD TabSize # 1 DO
                        Put(s, SpaceCh);
                    END;
                ELSE B[Char] := c;
                    IF Hwm < Char THEN Hwm := Char; END;
                    INC (Char);
                END;
            END;
END Put;

PROCEDURE Delete(s: Stream);
VAR
    b: BUFFER;
BEGIN
    b := BUFFER(s^.instanceData);
            Flush (b, NewLine);
            DISPOSE(b)
END Delete;


PROCEDURE Backspace(s: Stream);
BEGIN
    Seek(s, Current, -1);
END Backspace;

PROCEDURE Seek(s: Stream; sm: SeekMode; offset: LongINTEGER);
VAR
    b: BUFFER;
    pos: INTEGER;
BEGIN
    b := BUFFER(s^.instanceData);
            CASE sm OF
                Beginning:   pos := offset;
            |   Current:     pos := b^.Char;
                             pos := pos + offset;
            |   End:         pos := b^.PageWidth;
                             pos := pos + offset;
            END;
            b^.Char := IMax (1, IMin (pos, b^.PageWidth+1));
END Seek;

PROCEDURE Tell(s: Stream; VAR p: LongCARDINAL);
VAR
    b: BUFFER;
    pos: INTEGER;
BEGIN
    b := BUFFER(s^.instanceData);
    p := b^.Char;
END Tell;


PROCEDURE SeekChar (s: Stream; seekmode: SeekMode; offset: INTEGER);
BEGIN
    Seek(s, seekmode, LongINTEGER(offset));
END SeekChar;

PROCEDURE TellChar (s: Stream): CARDINAL;
VAR
    b: BUFFER;
BEGIN
    b := BUFFER(s^.instanceData);
    RETURN (b^.Char);
END TellChar;


PROCEDURE TellLine (s: Stream): CARDINAL;
VAR
    b: BUFFER;
BEGIN
    b := BUFFER(s^.instanceData);
    RETURN (b^.Line);
END TellLine;


PROCEDURE TellPage (s: Stream): CARDINAL;
    VAR b: BUFFER;
BEGIN
    b := BUFFER(s^.instanceData);
    RETURN b^.Page;
END TellPage;


PROCEDURE PageEvent (s: Stream; p: EventProc);
    VAR b: BUFFER;
BEGIN
    b := BUFFER(s^.instanceData);
    b^.PageProc := p;
END PageEvent;


PROCEDURE LineEvent (s: Stream; p: EventProc);
    VAR b: BUFFER;
BEGIN
    b := BUFFER(s^.instanceData);
    b^.LineProc := p;
END LineEvent;


PROCEDURE NullProc (s: Stream; i: CARDINAL);
END NullProc;

    
PROCEDURE CreateOutput (s: Stream; pagewidth, pagelength: CARDINAL;
    tab, fold: BOOLEAN): Stream;
    VAR i: CARDINAL;
    VAR b: BUFFER;
    VAR ns: Stream;
BEGIN
    NEW (b);
    WITH b^ DO
        S := s;
        FOR i := 1 TO MaxWidth DO B[i] := SpaceCh END;
        PageWidth := pagewidth;
        PageLength := pagelength;
        Char := 1;
        Hwm := 0;
        Line := 1;
        Page := 1;
        Tab := tab;
        Fold := fold;
        LineProc := NullProc;
        PageProc := NullProc;
        LinePending := FALSE;
        PagePending := TRUE;
    END;
    ns := Create(StreamProcs(procs), b);
    b^.stream := ns;
    RETURN ns;
END CreateOutput;

BEGIN
    procs := DefaultProcs(8);
    procs^.Put := Put;
    procs^.Delete := Delete;
    procs^.Seek := Seek;
    procs^.Tell := Tell;
    procs^.Backspace := Backspace;
    procs^.Flush := SFlush;
END PageStream.
