IMPLEMENTATION MODULE Storage;
(*
    Title: 	Storage Allocator
    LastEdit:	"Thu Jan 28 17:44:09 1988"
    Author: 	jd, mjj 
		Cambridge University Computer Laboratory
*)

FROM OSStorage IMPORT WordAlign, AddressesPerUnit, HeapAllocate;
FROM SYSTEM IMPORT 
    ADDRESS, WORD, MAXCARD, TSIZE;
FROM Exceptions IMPORT RAISE;

(* $T-,$R- *)

(* This is a machine-independent storage allocator based on the idea of
   keeping lists of free blocks of differing sizes.
   
   Each block in the heap has a free word at the front. When the block
   is free, this word contains the size of the block, and the next word
   is used as a chain pointer to blocks of similar size. If
   the block is in use, the first word contains either the negative of the
   size of the block, if the block is large enough to be on the overflow
   list, or the negative of the index of the block in the array of free lists,
   otherwise. The reason for the negative is to trap deallocations of already
   free objects. All sizes in this code are in the (arbitrary) storage units
   defined by TSIZE (WORD).

   The block sizes are allocated in intervals of about 20%. This means
   that fragmentation is about 10%. The last slot in the block list is used
   for very large allocations. This is the only slot for which the
   dependent list of free blocks will not all have the same size.
   
   To allocate a block of memory, we first go to the list of blocks of
   size just greater than that requested. If this is NIL, then we make a
   new block of this size by getting it from the end of the current block
   of heap. Otherwise, we chain down the list looking for a block which is
   large enough. In all cases but the overflow list, the first block on the
   list will be found.

   To dispose of a block, we look at the index word stored with it. This must
   be negative if the block is indeed in use, and if so, consider its
   absolute value. If this is small, it is the index of the list on which to
   chain the word. Otherwise it is the actual size of the block, which must
   be put on the overflow list. We insist that the argument given to 
   DEALLOCATE be at least as large as the actual size of the object.
   
*)

TYPE
    BLOCK = POINTER TO BLOCKLINK;

    BLOCKLINK = RECORD
        Size:   INTEGER;   (* or FreeList index in used blocks *)
        Block:  BLOCK;      (* only in free blocks *)
    END;

CONST Overflow = 40;

VAR FreeList: ARRAY [1..Overflow] OF BLOCKLINK;
    
VAR
    Base:        ADDRESS;               (* base of free space block     *)
    Free:        CARDINAL;              (* size of free space block     *)
    Floor:       ADDRESS;               (* highest reserved address     *)
(*    WordsAllocated,                     (* ALLOCATEd and not DEALLOCATEd*)*)
(*    WordsDeallocated: CARDINAL;         (* DEALLOCATEd and not ALLOCATEd*)*)

CONST
    AllocateFailed = "ALLOCATE failed";
    DeAllocateFailed = "DEALLOCATE failed";

VAR Mode: CARDINAL;


PROCEDURE SetMode (m: CARDINAL); BEGIN Mode := m END SetMode;


PROCEDURE ALLOCATE (VAR a: ADDRESS; n: CARDINAL);
    VAR index, half, need: INTEGER;
        rem: CARDINAL;
        b, p: BLOCK;
BEGIN
    IF WordAlign THEN
        rem := n MOD TSIZE (WORD);
        IF rem # 0 THEN INC (n, TSIZE (WORD) - rem) END
    END;
    need := n + TSIZE (INTEGER);  (* allow space for the size word *)

    (* now find the correct slot by binary chop, biased towards small sizes *)
    index := Overflow DIV 4;
    half := index DIV 2;
    LOOP
        IF need > FreeList [index].Size THEN
            INC (index, half);
            IF index = Overflow THEN EXIT END;
        ELSIF need > FreeList [index-1].Size THEN
            EXIT
        ELSE DEC (index, half);
            IF index = 1 THEN EXIT END;
        END;
        IF half > 1 THEN half := half DIV 2 END;
    END;

    (* now either choose a free block of the right size or get a new one *)
    b := FreeList [index].Block;
    p := NIL;
    WHILE (b # NIL) AND (b^.Size < need) DO 
        p := b; 
        b := b^.Block
    END;
    IF b = NIL THEN
        IF index # Overflow THEN need :=FreeList[index].Size END;
        IF (INTEGER (Free) < need) AND NOT GetBlock (need) THEN 
            IF Mode = 1 THEN RAISE(AllocateFailed) ELSE a := NIL; RETURN END;
        END;
        b := Base;
        INC (Base, need * INTEGER(AddressesPerUnit));
        DEC (Free, CARDINAL (need));
    ELSIF p = NIL THEN
        FreeList [index].Block := b^.Block;
        (* DEC (WordsDeallocated, CARDINAL (need)); *)
    ELSE p^.Block := b^.Block
    END;
    (* INC (WordsAllocated, need); *)
    a := ADDRESS (b) + AddressesPerUnit*TSIZE(INTEGER);
    IF index = Overflow THEN b^.Size := -need ELSE b^.Size := -index END
END ALLOCATE;


PROCEDURE DEALLOCATE (VAR a: ADDRESS; n: CARDINAL);
    VAR b: BLOCK;
        index, words: CARDINAL;
BEGIN
    IF a # NIL THEN
        b := BLOCK (a - AddressesPerUnit*TSIZE(WORD));
        WITH b^ DO
            IF Size < 0 THEN index := -Size ELSE RAISE(DeAllocateFailed) END;
            IF index < Overflow THEN 
                IF INTEGER (n) > FreeList [index].Size THEN
                    RAISE(DeAllocateFailed)
                END;
                Block := FreeList [index].Block;
                Size := FreeList [index].Size;
                FreeList [index].Block := b;
                words := Size;
            ELSE IF INTEGER (n) > -Size THEN RAISE(DeAllocateFailed) END;
                Block := FreeList [Overflow].Block;
                FreeList [Overflow].Block := b;
                words := -Size
            END;
        END;
        (* INC (WordsDeallocated, words); *)
        (* DEC (WordsAllocated, words); *)
        a := NIL;
    END;
END DEALLOCATE;


PROCEDURE GetBlock (n: CARDINAL): BOOLEAN;
    VAR want, lack: CARDINAL;
BEGIN
    REPEAT
        (* ensure that any free space lies above Floor *)
        IF Base < Floor THEN
            lack := (Floor - Base) DIV AddressesPerUnit;
            IF lack > Free THEN lack := Free END;
            Base := Base + lack * AddressesPerUnit;
            DEC (Free, CARDINAL (lack));
            lack := (Floor - Base) DIV AddressesPerUnit;
        ELSE lack := 0;
        END;
        (* ensure that there is enough Free space *)
        want := n  + lack;
        IF Free < want THEN
            IF NOT HeapAllocate (want, Base, Free) THEN
                RETURN FALSE;
            END;
        END;
        (* but of course Base may have changed, so we may need to loop *)
    UNTIL (Base >= Floor) AND (Free >= want);
    RETURN TRUE;
END GetBlock;

PROCEDURE Set5 (index, s1, s2, s3, s4, s5: CARDINAL);
    VAR i: CARDINAL;
BEGIN
    FOR i := index TO index + 4 DO FreeList[i].Block := NIL END;
    FreeList[index].Size := s1*TSIZE(WORD);
    FreeList[index+1].Size := s2*TSIZE(WORD);
    FreeList[index+2].Size := s3*TSIZE(WORD);
    FreeList[index+3].Size := s4*TSIZE(WORD);
    FreeList[index+4].Size := s5*TSIZE(WORD);
END Set5;


BEGIN
(*    WordsAllocated := 0;  *)
(*    WordsDeallocated := 0; *)
    Mode := 1;
    Base := NIL;
    Free := 0;
    Floor := 0;
    Set5 (1,     2,    3,    4,    6,    8);
    Set5 (6,    10,   12,   15,   18,   22);
    Set5 (11,   26,   32,   39,   47,   57);
    Set5 (16,   68,   82,   98,  117,  140);
    Set5 (21,  168,  201,  241,  290,  348);
    Set5 (26,  418,  501,  601,  722,  866);
    Set5 (31, 1039, 1246, 1496, 1795, 2154);
    Set5 (36, 2585, 3102, 3722, 4466, 5360);
END Storage.
