(*
    Title: 	TextIO - Implementation
    LastEdit:	"Thu Nov 29 14:39:50 1984"
    Author: 	Mick Jordan
		Acorn Research Centre
*)

IMPLEMENTATION MODULE TextIO;

FROM SYSTEM IMPORT BITSPERWORD, MOSTSIGBIT;
FROM SystemTypes IMPORT String;
FROM Exceptions IMPORT RAISE;
FROM TypeToChars IMPORT
    INTEGERToChars, CARDINALToChars, BOOLEANToChars, REALToChars, 
    BasedCARDINALToChars, BITSETToChars;
FROM CharCodes IMPORT NewLineCh, SpaceCh, TabCh, FormFeedCh, CReturnCh;
IMPORT Streams; FROM Streams IMPORT Stream, ErrorCode, Put, Get, Backspace;
IMPORT Strings;

TYPE
    Digits = ARRAY [0..BITSPERWORD-1] OF CHAR;
    REALDigits = ARRAY [0..79] OF CHAR;
    BITSETDigits = ARRAY [0..2*BITSPERWORD+2] OF CHAR;

VAR
    TrueString, FalseString : ARRAY [0..6] OF CHAR;

(* --------------------- Output Procedures -------------------------------- *)

PROCEDURE WriteCHAR(s: Stream; ch: CHAR);
BEGIN
    Put(s, ch);
END WriteCHAR;

PROCEDURE NewLine(s: Stream);
BEGIN
    Put(s, NewLineCh);
END NewLine;

PROCEDURE WriteString(s: Stream; string: String);
VAR
    i: CARDINAL;
BEGIN
    i := 0;
    (* $T- *)
    WHILE string^[i] # 0C DO
        Put(s, string^[i]);
        INC(i);
    END (* while *);
    (* $T= *)
END WriteString;

PROCEDURE WriteChars(s: Stream; str: ARRAY OF CHAR);
VAR
    i: CARDINAL;
BEGIN
    i := 0;
    (* $T- *)
    WHILE (i <= HIGH(str)) AND (str[i] # 0C) DO
        Put(s, str[i]);
        INC(i);
    END (* while *);
    (* $T= *)
END WriteChars;

PROCEDURE WriteINTEGER(s : Stream;  i : INTEGER);
VAR  digits : Digits;
    void: CARDINAL;
BEGIN
    void := INTEGERToChars(i, digits);
    WriteChars(s, digits);
END WriteINTEGER;

PROCEDURE WriteCARDINAL(s : Stream;  c : CARDINAL);
VAR  digits : Digits;
    void: CARDINAL;
BEGIN
    void := CARDINALToChars(c, digits);
    WriteChars(s, digits);
END WriteCARDINAL;

PROCEDURE WriteBOOLEAN(s : Stream;  bool : BOOLEAN);
VAR
    digits: Digits;
    void: CARDINAL;
BEGIN
    void := BOOLEANToChars(bool, digits);
    WriteChars(s, digits)
END WriteBOOLEAN;

PROCEDURE WriteREAL(s : Stream;  real : REAL;  decplaces : CARDINAL);
VAR realdigits : REALDigits;
    void: CARDINAL;
BEGIN
    void := REALToChars(real, decplaces, realdigits);
    WriteChars(s, realdigits);
END WriteREAL;

PROCEDURE WriteBasedCARDINAL(s: Stream; n: CARDINAL; base: NumberBase);
VAR
    digits: Digits;
    void: CARDINAL;
BEGIN
    void := BasedCARDINALToChars(n, base, digits);
    WriteChars(s, digits);
END WriteBasedCARDINAL;

PROCEDURE WriteBITSET(s: Stream; b: BITSET; setSize: CARDINAL);
VAR
    digits: BITSETDigits;
    void: CARDINAL;
BEGIN
    void := BITSETToChars(b, setSize, digits);
    WriteChars(s, digits);
END WriteBITSET;

(* --------------------- Input Procedures -------------------------------- *)

PROCEDURE ReadCHAR(s: Stream): CHAR;
BEGIN
    RETURN CHAR( Get(s) );
END ReadCHAR;

TYPE
    SKIPRESULT = (First, Second, Third, Neither);
    (* returned by SkipToOneOf *)

PROCEDURE SkipToOneOf(s: Stream; ch1, ch2, ch3: CHAR): SKIPRESULT;
(*  return 'First' if 'ch1' encountered,
   'Second' if 'ch2' encountered,
   'Third' if 'ch3' encountered, 'Neither' otherwise 
*)
VAR ch: CHAR;
BEGIN
    ch := CHAR( Get(s) );
    LOOP
        CASE ch OF
          SpaceCh, TabCh, NewLineCh, FormFeedCh:
        ELSE
            IF    ch = ch1 THEN RETURN First
            ELSIF ch = ch2 THEN RETURN Second
            ELSIF ch = ch3 THEN RETURN Third
            ELSE  SafeBackspace(s); RETURN Neither END;
        END;
        ch := CHAR( Get(s) );
    END;
END SkipToOneOf;

PROCEDURE SkipWhiteSpace(s: Stream): CHAR;
VAR
    ch: CHAR;
BEGIN
    REPEAT
        ch := CHAR( Get(s) );
    UNTIL (ch#SpaceCh) AND (ch#NewLineCh) AND
          (ch#FormFeedCh) AND (ch#TabCh);
    RETURN ch;
END SkipWhiteSpace;

PROCEDURE ReadINTEGER(s: Stream): INTEGER;
VAR i : INTEGER;  negative: BOOLEAN;  ch: CHAR;
BEGIN
    negative := SkipToOneOf(s, '-', '+', 0C) = First;
    i := INTEGER(ReadBasedCARDINAL(s, 10));
    IF negative THEN RETURN -i ELSE RETURN i END;
END ReadINTEGER;

PROCEDURE ReadCARDINAL(s: Stream): CARDINAL;
BEGIN
    RETURN ReadBasedCARDINAL(s, 10);
END ReadCARDINAL;

PROCEDURE ReadBOOLEAN(s: Stream): BOOLEAN;

    PROCEDURE CheckTrueOrFalse(chars: ARRAY OF CHAR): BOOLEAN;
    VAR  ch: CHAR;  i: CARDINAL;
    BEGIN
        i := 1;   (* first char already swallowed *)
        WHILE (i <= HIGH(chars)) AND (chars[i] # 0C) DO
            ch := CHAR( Get(s) );
            IF ch # chars[i] THEN SafeBackspace(s);  RETURN FALSE END;
            INC(i);
        END (* while *);
        RETURN TRUE;
    END CheckTrueOrFalse;

VAR  skip: SKIPRESULT;
    void: BOOLEAN;
BEGIN
    skip := SkipToOneOf(s, 'T', 'F', 0C);
    CASE skip OF
      First:  RETURN CheckTrueOrFalse(TrueString);
    | Second: void := CheckTrueOrFalse(FalseString);
    | Neither:
    END (* case *);
    RETURN FALSE;
END ReadBOOLEAN;

PROCEDURE ReadBasedCARDINAL(s: Stream; base: NumberBase): CARDINAL;
VAR ch: CHAR;  sum, c: CARDINAL;
BEGIN
    ch := SkipWhiteSpace(s);
    sum := 0;
    LOOP
        IF    (ch >= '0') AND (ch <= '9') THEN c := ORD(ch) - ORD('0');
        ELSIF (ch >= 'A') AND (ch <= 'Z') THEN c := ORD(ch) - ORD('A') + 10;
        ELSIF (ch >= 'a') AND (ch <= 'z') THEN c := ORD(ch) - ORD('a') + 10;
        ELSE c := 100 END;
        IF c >= base THEN EXIT END;
        sum := sum*base + c;
        ch := CHAR( Get(s) );
    END;
    SafeBackspace(s);
    RETURN sum;
END ReadBasedCARDINAL;

PROCEDURE ReadREAL(s: Stream) : REAL;
VAR        real, factor, frac : REAL;
    negative, fractionalpart : BOOLEAN;
    ch : CHAR;
    exponent : INTEGER;
BEGIN
    real :=  0.0;  fractionalpart := FALSE;
    negative := SkipToOneOf(s, '-', '+', 0C) = First;
    LOOP
        ch := CHAR( Get(s) );
        CASE ch OF
          '0' .. '9':
            real := real * 10.0 + FLOAT (ORD (ch) - ORD ('0'));
        |
            '.':   fractionalpart := TRUE;
            EXIT;
        ELSE
            EXIT;
        END;
    END;
    IF fractionalpart THEN
        factor := 1.0; frac := 0.0;
        LOOP
            ch := CHAR( Get(s) );
            CASE ch OF
              '0' .. '9':
                factor := factor / 10.0;
              frac := frac + factor * FLOAT(ORD(ch) - ORD('0'));
            ELSE
                EXIT;
            END;
        END;
        real := real + frac;
    END;
    IF (ch = 'E') OR (ch = 'e') THEN
        exponent := ReadINTEGER (s);
        WHILE exponent > 0 DO  real := real * 10.0;  DEC (exponent) END;
        WHILE exponent < 0 DO  real := real / 10.0;  INC (exponent) END;
    ELSE SafeBackspace(s) END;
    IF negative THEN RETURN -real ELSE RETURN real END;
END ReadREAL;

PROCEDURE ReadBITSET(s: Stream): BITSET;
VAR
    skip: SKIPRESULT;
    bitset: BITSET;
    n, m: CARDINAL;
    ch: CHAR;
BEGIN
    bitset := {};
    skip := SkipToOneOf(s, '{', 0C, 0C);
    IF skip = First THEN
        REPEAT
            ch := SkipWhiteSpace(s);
            SafeBackspace(s);
            IF (ch >= '0') AND (ch <= '9') THEN
                n := ReadCARDINAL(s);
                INCL(bitset, n);
            END (* if *);
            skip := SkipToOneOf(s, ',', '.', '}');
            IF skip = Second THEN    (* n..m *)
                ch := CHAR( Get(s) );
                m := ReadCARDINAL(s);
                WHILE n <= m DO
                    INCL(bitset, n);
                    INC(n);
                END (* while *);
            END (* if *);
        UNTIL NOT ((skip = First) OR (skip = Second));
    END (* if *);
    RETURN bitset;
END ReadBITSET;

PROCEDURE WhiteSpace(ch: CHAR): BOOLEAN;
BEGIN
    RETURN (ch = SpaceCh) OR (ch = TabCh) OR (ch = NewLineCh) OR
           (ch = CReturnCh) OR (ch = FormFeedCh);
END WhiteSpace;

PROCEDURE EndOfLine(ch: CHAR): BOOLEAN;
BEGIN
    RETURN (ch = NewLineCh) OR
           (ch = CReturnCh) OR (ch = FormFeedCh)
END EndOfLine;

PROCEDURE ReadChars(s: Stream; VAR chars: ARRAY OF CHAR; T: TermProc);
VAR
    i: CARDINAL;
    ch: CHAR;
BEGIN
    i := 0;
    LOOP
        ch := CHAR( Get(s) );
        IF (s^.status # Success) OR T(ch) THEN
            EXIT
        ELSE
            chars[i] := ch;
            INC(i);
        END (* if *);
    END (* loop *);
    IF i <= HIGH(chars) THEN
        chars[i] := 0C;
    END (* if *);
    SafeBackspace(s);
END ReadChars;

PROCEDURE ReadString(s: Stream; T: TermProc): String;
VAR
    array: ARRAY [0..255] OF CHAR;
BEGIN
    ReadChars(s, array, T);
    RETURN Strings.CopyCS(array);
END ReadString;

PROCEDURE SafeBackspace(s: Stream);
  BEGIN
    IF s^.status = Success THEN
      Backspace(s);
    END;
  END SafeBackspace;

BEGIN
    TrueString   := "TRUE";
    FalseString  := "FALSE";
END TextIO.
