(*
    Title: 	UnixArchive - Implementation
    LastEdit:	"Tue Sep 25 13:33:57 1984"
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory
    
    VAX/Unix portable format
*)

(* $T-, $R- *)

IMPLEMENTATION MODULE UnixArchive;

FROM SYSTEM IMPORT WORD, TSIZE, BYTESPERWORD, ADR, ADDRESS, MAXINT;
FROM SystemTypes IMPORT LongCARDINAL, LongINTEGER;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM Streams IMPORT 
    Stream, SeekMode, ErrorCode, Create, StreamProcs, OptionalFunction;
FROM StreamRep IMPORT DefaultProcs, Procs;
FROM CharCodes IMPORT NewLineCh, SpaceCh;
FROM UnixIO IMPORT OpenForRead, Open, Close, Read, Lseek;
IMPORT Strings;

TYPE 
     MemberId = INTEGER;        (* seek offset to header *)

TYPE ARCHEADER = RECORD
(*  Modula-2 arrays are word aligned *)
        name: ARRAY [0 .. 15] OF CHAR;
        data: ARRAY [0 .. 11] OF CHAR;
        uidgid: ARRAY [0 .. 11] OF CHAR;
        mode: ARRAY [0 .. 7] OF CHAR;
        sizefmag: ARRAY [0 .. 11] OF CHAR;
    END (* record *);

TYPE TABLEENTRY = RECORD
        name: ARRAY [0 .. 15] OF CHAR;
        offset: CARDINAL;
    END (* record *);

TYPE TABLEOFCONTENTS = POINTER TO ARRAY [0 .. 500] OF TABLEENTRY;

TYPE ARCHIVE = RECORD
       fildes: INTEGER;    (* Unix file descriptor *)
        tableofcontents: TABLEOFCONTENTS;
        usetable: BOOLEAN;
        nentries: CARDINAL;     (* number of entries in tableofcontents *)
        centry: CARDINAL;       (* current entry during FindNextMember *)
        database: CARDINAL;     (* offset of actual archive data *)
    END (* record *);

TYPE ArchiveId = POINTER TO ARCHIVE;

TYPE CHARS8 = ARRAY [0..7] OF CHAR;

VAR ArMag, CTable: CHARS8;
    header: ARCHEADER;

CONST
    BEGINNING = 0;  (* Unix lseek constants *)
    CURRENT = 1;

TYPE MEMBERDATA = POINTER TO ARRAY [0 .. MAXINT DIV 4] OF CHAR;

TYPE MEMBER = POINTER TO RECORD
        fildes: INTEGER;
        memberdata: MEMBERDATA;
        memberdatasize: CARDINAL;
        wordsize: CARDINAL;
        index: CARDINAL;
    END (* record *);

PROCEDURE ReadHeader(fildes: INTEGER): INTEGER;
BEGIN
    RETURN Read(fildes, header,
      (TSIZE(ARCHEADER) DIV TSIZE(WORD))*BYTESPERWORD);
END ReadHeader;

PROCEDURE CardSize(charsize: ARRAY OF CHAR): CARDINAL;
VAR
    i: CARDINAL;
    size: CARDINAL;
    ch: CHAR;
BEGIN
    size := 0;
    i := 0;
    REPEAT
        ch := charsize[i];
        IF (ch >= '0') AND (ch <= '9') THEN
            size := size*10 + CARDINAL(ch) - CARDINAL('0');
        END (* if *);
        INC(i);
    UNTIL (ch = SpaceCh) OR (i>=10);
    RETURN size;
END CardSize;

PROCEDURE SkipFile(fildes: INTEGER; VAR endseek: INTEGER): INTEGER;
VAR
    msize: CARDINAL;
BEGIN
    msize := CardSize(header.sizefmag);
    IF ODD(msize) THEN
        INC(msize);
    END (* if *);
    endseek := INTEGER(msize);
    RETURN Lseek(fildes, endseek, CURRENT);
END SkipFile;

PROCEDURE OpenArchive(name: ARRAY OF CHAR; VAR arcid: ArchiveId): BOOLEAN;
VAR
    r, seeko, f: INTEGER;
    i: CARDINAL;
    charbuf: CHARS8;
    offsetbuf: ARRAY [0 .. 11] OF CHAR;
    tableheader: RECORD
        padding: ARRAY [0 .. 7] OF CHAR;
        entrycount: ARRAY [0 .. 11] OF CHAR;
    END (* record *);
    unixName: ARRAY [0..255] OF CHAR;
BEGIN
    Strings.CopyCC(name, unixName);   (* appends a 0C *)
    f := Open(unixName, OpenForRead, {});
    IF f<0 THEN
        RETURN FALSE;
    ELSE
        r := Read(f, charbuf, 8);
        IF (r<=0) OR (NOT Strings.EqualCC(charbuf, ArMag)) THEN
            RETURN FALSE;
        END (* if *);

        NEW(arcid);
        WITH arcid^ DO
            (* look for table of contents *)
            fildes := f;
            r := Read(fildes, charbuf, 8);
            IF Strings.EqualCC(charbuf, CTable) THEN
                usetable := TRUE;
                r := Read(fildes, tableheader, 20);
                nentries := CardSize(tableheader.entrycount);
                centry := 0;
                ALLOCATE(tableofcontents, nentries*TSIZE(TABLEENTRY));
                FOR i := 0 TO nentries-1 DO
                    r := Read(fildes, tableofcontents^[i].name, 16);
                    MakeM2Chars(tableofcontents^[i].name);
                    r := Read(fildes, offsetbuf, 12);
                    tableofcontents^[i].offset := CardSize(offsetbuf);
                    seeko := 0;
                    r := Lseek(fildes, seeko, CURRENT);
                    database := CARDINAL(seeko);
                END (* for *);
            ELSE    (*  seek back to start of archive data *)
                seeko := -8;
                r := Lseek(fildes, seeko, CURRENT);
                usetable := FALSE;
            END (* if *);
        END (* with *);
    END (* if *);
    RETURN TRUE;
END OpenArchive;

PROCEDURE CloseArchive(a: ArchiveId);
VAR
    r: INTEGER;
BEGIN
    WITH a^ DO
        r := Close(fildes);
        IF usetable THEN
            DEALLOCATE(tableofcontents, nentries*TSIZE(TABLEENTRY));
        END (* if *);
    END (* with *);
    DISPOSE(a);
END CloseArchive;

PROCEDURE MakeM2Chars(VAR chars: ARRAY OF CHAR);
(*  put 'chars' into standard form, i.e pad with 0C if necessary *)
VAR
    i: CARDINAL;
BEGIN
    FOR i := 0 TO 15 DO
        IF chars[i] = SpaceCh THEN
            chars[i] := 0C;
            RETURN
        END (* if *);
    END (* for *);
END MakeM2Chars;

PROCEDURE FindMember(arcid: ArchiveId; name: ARRAY OF CHAR;
                     VAR memberid: MemberId): BOOLEAN;
VAR
    r, i: INTEGER;
    seek: INTEGER;

BEGIN
    WITH arcid^ DO
        IF usetable THEN
            FOR i := 0 TO nentries-1 DO
                IF Strings.EqualCC(name, tableofcontents^[i].name) THEN
                    memberid := database + tableofcontents^[i].offset;
                    RETURN TRUE;
                END (* if *);
            END (* for *);
        ELSE
            seek := 8;
            r := Lseek(fildes, seek, BEGINNING);
            memberid := seek;
            WHILE (r>=0) AND (ReadHeader(fildes) > 0) DO
                MakeM2Chars(header.name);
                IF Strings.EqualCC(name, header.name) THEN
                    RETURN TRUE;
                END (* if *);
                r := SkipFile(fildes, memberid);
            END (* while *);
        END (* if *);
    END (* with *);
    RETURN FALSE;
END FindMember;

PROCEDURE FindNextMember(arcid: ArchiveId; VAR name: ARRAY OF CHAR;
                            VAR memberid: MemberId): BOOLEAN;
    VAR
        r: INTEGER;
    BEGIN
      WITH arcid^ DO
          IF usetable THEN
              IF centry >= nentries THEN
                RETURN FALSE
              ELSE
                  Strings.CopyCC(tableofcontents^[centry].name, name);
                  memberid := database + tableofcontents^[centry].offset;
                  INC(centry);
                  RETURN TRUE;
              END (* if *);
          ELSE
            memberid := 0;
            r := Lseek(fildes, memberid, CURRENT);
            IF ReadHeader(fildes) <= 0 THEN
                RETURN FALSE;
            END (* if *);
            MakeM2Chars(header.name);
            Strings.CopyCC(header.name, name);
            r := SkipFile(fildes, r);
            RETURN TRUE;
          END (* if *);
      END (* with *);
    END FindNextMember;

PROCEDURE CreateInput(a: ArchiveId; name: ARRAY OF CHAR): Stream;
VAR
    memberid: MemberId;
    r: INTEGER;
    s: Stream;
BEGIN
    s := Create(StreamProcs(procs), 0);
    IF FindMember(a, name, memberid) THEN
        WITH a^ DO
            IF usetable THEN
                r := Lseek(fildes, memberid, BEGINNING);
                r := ReadHeader(fildes);
            END (* if *);
            Initialise(s, fildes);
        END (* with *);
    ELSE
        s^.status := StreamError;
    END (* if *);
    RETURN s;
END CreateInput;

PROCEDURE CreateFastInput(a: ArchiveId; memberid: MemberId): Stream;
VAR
    s: Stream;
BEGIN
    s := Create(StreamProcs(procs), 0);
    WITH a^ DO
        IF (Lseek(fildes, memberid, BEGINNING) >= 0) AND
              (ReadHeader(fildes) > 0) THEN
            Initialise(s, fildes);
        ELSE
            s^.status := StreamError;
        END (* if *);
        RETURN s;
    END (* with *);
    RETURN s;
END CreateFastInput;

PROCEDURE Initialise(s: Stream; xfildes: CARDINAL);
(*
    The header for the member is currently stored in 'header'
    and the file is positioned at the first byte of the actual data.
    We block read the whole member into the heap.
*)
VAR
    member: MEMBER;
    r: INTEGER;
BEGIN
    NEW(member);
    s^.instanceData := WORD(member);
    WITH member^ DO
        fildes := xfildes;
        memberdatasize := CardSize(header.sizefmag);
        wordsize := memberdatasize;
        IF (wordsize MOD 4) # 0 THEN
            INC(wordsize, 4- (wordsize MOD 4));
        END (* if *);
        wordsize := wordsize DIV 4;
        ALLOCATE(memberdata, wordsize*TSIZE(WORD));
        index := 0;
        r := Read(fildes, memberdata^, memberdatasize);
        IF r<0 THEN
            DEALLOCATE(memberdata, wordsize*TSIZE(WORD));
            s^.status := StreamError;
        END (* if *);
    END (* with *);
END Initialise;

PROCEDURE Get(s: Stream): WORD;
VAR
    member: MEMBER;
BEGIN
    member := MEMBER(s^.instanceData);
    WITH member^ DO
        IF index >= memberdatasize THEN
            s^.status := EndOfStream;
            RETURN WORD(0C);
        ELSE
            INC(index);
            RETURN WORD(memberdata^[index-1]);
        END (* if *);
    END (* with *);
END Get;

PROCEDURE GetN(s: Stream; a: ADDRESS; n: CARDINAL): CARDINAL;
VAR
    member: MEMBER;
    getBuf: MEMBERDATA;
    i, toGet: CARDINAL;
BEGIN
    member := MEMBER(s^.instanceData);
    getBuf := MEMBERDATA(a);
    IF member^.memberdatasize - member^.index >= n THEN
        toGet := n;
    ELSE
        toGet := member^.memberdatasize - member^.index;
        IF toGet = 0 THEN
            s^.status := EndOfStream;
        END (* if *);
    END (* if *);
    IF toGet > 0 THEN
        FOR i := 0 TO toGet-1 DO
            getBuf^[i] := member^.memberdata^[i+member^.index];
        END (* for *);
        INC(member^.index, toGet);
    END (* if *);
    RETURN toGet;
END GetN;

PROCEDURE Backspace(s: Stream);
VAR
    member: MEMBER;
BEGIN
    member := MEMBER(s^.instanceData);
    IF member^.index > 0 THEN
        DEC(member^.index);
    END (* if *);
END Backspace;

PROCEDURE Tell(s: Stream; VAR p: LongCARDINAL);
VAR
    member: MEMBER;
BEGIN
    member := MEMBER(s^.instanceData);
    p := member^.index;
END Tell;

PROCEDURE Seek(s: Stream; seekMode: SeekMode; offset: LongINTEGER);
VAR
    member: MEMBER;
BEGIN
    member := MEMBER(s^.instanceData);
    WITH member^ DO
        CASE seekMode OF
          Beginning:
            index := CARDINAL(offset);
        | Current:
            index := CARDINAL(INTEGER(index)+offset);
        | End:
            index := CARDINAL(INTEGER(memberdatasize)+offset);
        END (* case *);
    END (* with *);
END Seek;

PROCEDURE Delete(s: Stream);
VAR
    member: MEMBER;
BEGIN
    member := MEMBER(s^.instanceData);
    DEALLOCATE(member^.memberdata, member^.wordsize*TSIZE(WORD));
    DISPOSE(member);
END Delete;

PROCEDURE Implements(s: Stream; f: OptionalFunction): BOOLEAN;
BEGIN
    RETURN TRUE;
END Implements;

PROCEDURE InitProcs(): Procs;
VAR
    p: Procs;
BEGIN
    p := DefaultProcs(8);
    p^.Get := Get;
    p^.GetN := GetN;
    p^.Delete := Delete;
    p^.Seek := Seek;
    p^.Tell := Tell;
    p^.Backspace := Backspace;
    RETURN p;
END InitProcs;
            
VAR procs: Procs;

BEGIN
    procs := InitProcs();
    ArMag := "!<arch>0";
    ArMag[7] := NewLineCh;
    CTable := "__CTABLE";
END UnixArchive.
