(*
    Title: 	WriteF - Implementation
    LastEdit:	"Thu Nov  1 16:07:01 1984"
    Author: 	Mick Jordan
		Acorn Research Centre
*)

IMPLEMENTATION MODULE WriteF;

FROM SYSTEM IMPORT WORD, BITSPERWORD, MAXINT, ADR;
FROM SystemTypes IMPORT String;
FROM Exceptions IMPORT RAISE;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM Streams IMPORT Stream, Get, Put, Backspace, ErrorCode;
FROM CharCodes IMPORT NewLineCh, TabCh, SpaceCh, FormFeedCh, CReturnCh, 
  CapitalCh;
FROM TypeToChars IMPORT
    INTEGERToChars, CARDINALToChars, BOOLEANToChars, REALToChars,
    NumberBase, BasedCARDINALToChars, BITSETToChars (*, ArrayOverflow *);

CONST  EscapeCh = '\';

TYPE 
    ArgArray = ARRAY [0..3] OF WORD;

    KeyType = (StringKey, RealKey);
    KeyIndex = [0..15];
    
    FormatKey = KeyIndex;
    
    Key = RECORD
        CASE keytype: KeyType OF
          StringKey: 
            upb: CARDINAL;
            string: String;
        | RealKey:
            realValue: REAL;
        END (* case *);
    END (* record *);

VAR Keys: ARRAY KeyIndex OF RECORD  (* needs locking if processes *)
        inUse: BOOLEAN;
        key: Key;
    END (* record *);

(* $T-  *) (* all self-checked or intrinsically safe after here *)

(*
    The purpose of H and S is to record the HIGH value of the string
    and its address and give the user a unique key so that
    the order of evaluation of arguments does not need to correspond
    to the order of %S occurrences in the format string.
    A similar argument applies to %R and R(real)
*)

(* $T-  *)
(* all self-checked or intrinsically safe after here *)

PROCEDURE H(VAR xstring: ARRAY OF CHAR): FormatKey
          (* RAISES {NoMoreKeys} *);
VAR keyIndex: KeyIndex;
BEGIN
    keyIndex := FreeKey();
    WITH Keys[keyIndex].key DO
        keytype := StringKey;
        upb := HIGH(xstring);
        string := String(ADR(xstring));
    END (* with *);
    RETURN keyIndex;
END H;

PROCEDURE S(p: String): FormatKey
          (* RAISES {NoMoreKeys} *);
VAR keyIndex: KeyIndex;
BEGIN
    keyIndex := FreeKey();
    WITH Keys[keyIndex].key DO
        keytype := StringKey;
        upb := MAXINT;   (* a large string *)
        string := p;
    END (* with *);
    RETURN keyIndex;
END S;

PROCEDURE R(r: REAL): FormatKey
          (* RAISES {NoMoreKeys} *);
VAR keyIndex: KeyIndex;
BEGIN
    keyIndex := FreeKey();
    WITH Keys[keyIndex].key DO
        keytype := RealKey;
        realValue := r;
    END (* with *);
    RETURN keyIndex;
END R;

PROCEDURE InitKeys;
VAR
    i: KeyIndex;
BEGIN
    FOR i := 0 TO HIGH(Keys) DO
        Keys[i].inUse := FALSE;
    END (* for *);
END InitKeys;

PROCEDURE FreeKey(): KeyIndex;
VAR
    i: KeyIndex;
BEGIN
    FOR i := 0 TO HIGH(Keys) DO
        IF NOT Keys[i].inUse THEN
            Keys[i].inUse := TRUE;
            RETURN i
        END (* if *);
    END (* for *);
    RAISE("NoMoreKeys");
END FreeKey;

PROCEDURE WriteFVAR(s : Stream; WriteFargs : ArgArray;
                         format : ARRAY OF CHAR; argCount : CARDINAL);

VAR formatIndex, argsUsed: CARDINAL;
    type, ch: CHAR;
    arg: WORD;
    width, realFracwidth, setSize: CARDINAL;
    realKey: FormatKey;
    digits: ARRAY [0..79] OF CHAR;
    argWidth: CARDINAL;

    PROCEDURE WriteFS(keyIndex: KeyIndex; width: CARDINAL);
    VAR  i, actualL: CARDINAL;
    BEGIN
        WITH Keys[keyIndex].key DO
            actualL := 0;
            WHILE (actualL <= upb) AND (string^[actualL] # 0C) DO
                INC(actualL);
            END;
            i := actualL;
            WHILE i < width DO
                Put(s, SpaceCh); INC(i);
            END;
            i:= 0;
            WHILE (i < actualL) DO
                Put(s, string^[i]);  INC(i);
            END;
        END (* with *);
        Keys[keyIndex].inUse := FALSE;
    END WriteFS;

    PROCEDURE NextIsDigit(hex: BOOLEAN; VAR c: CARDINAL): BOOLEAN;
    VAR  ch: CHAR;
    BEGIN
      IF formatIndex < HIGH(format) THEN
        ch:= CapitalCh(format[formatIndex+1]);
        IF (ch>='0') AND (ch<='9') THEN
          INC(formatIndex);  c:= ORD(ch) - ORD('0');
          RETURN TRUE
        ELSIF hex AND ((ch>='A') AND (ch<='Z')) THEN
          INC(formatIndex); c := ORD(ch) - ORD('A') + 10;
          RETURN TRUE;
        END (* if *);
      END (* if *);
      RETURN FALSE;
    END NextIsDigit;

    PROCEDURE FieldWidth(): CARDINAL;
    (*  Zero, one or two digits *)
    VAR n, n2: CARDINAL;
    BEGIN
      n:= 0;
      IF NextIsDigit(FALSE, n) THEN
        IF NextIsDigit(FALSE, n2) THEN n:= n*10 + n2 END;
      END (* if *);
      RETURN n
    END FieldWidth;
    
    PROCEDURE TransEscape(ch: CHAR): CHAR;
      VAR
        n, n2: CARDINAL;
      BEGIN
        CASE ch OF
          'N','n':    ch := NewLineCh;
        | 'P','p':    ch := FormFeedCh;
        | 'S','s':    ch := SpaceCh;
        | 'T','t':    ch := TabCh;
        | 'C','c':    ch := CReturnCh;
        | 'X','x':
          n := 0;
          IF NextIsDigit(TRUE, n) THEN
            IF NextIsDigit(TRUE, n2) THEN n := n*16 + n2; END;
          END;
          ch := CHAR(n);
        ELSE (* no translation *) 
        END;
        RETURN ch;
      END TransEscape;

BEGIN
  argsUsed:= 0; formatIndex:= 0;
  WHILE (formatIndex <= HIGH(format)) AND (format[formatIndex] # 0C) DO
    ch:= format[formatIndex];
    IF ch = '%' THEN
      INC(formatIndex);
      type:= format[formatIndex];  arg:= WriteFargs[argsUsed];
      CASE type OF
        'C', 'c', 'U', 'u', 'I', 'i', 'S', 's', 'R', 'r', 'T', 't',
          'B', 'b', 'X', 'x', 'O', 'o':
          width:= FieldWidth()
      ELSE (* no width *)
      END (* case *);
      CASE type OF
        '$':      (* just skip the argument *)
      | 'C','c':  
          WHILE width > 1 DO
            Put(s, SpaceCh);  DEC(width);
          END (* while *);
          Put(s, CHAR(arg));
      | 'U','u':  
          argWidth:= BasedCARDINALToChars(CARDINAL(arg), 10, digits);
          WriteNChars(s, digits, argWidth, width);
      | 'I','i':
          argWidth:= INTEGERToChars(INTEGER(arg), digits);
          WriteNChars(s, digits, argWidth, width);
      | 'B','b':
          argWidth:= BOOLEANToChars(BOOLEAN(arg), digits);
          WriteNChars(s, digits, argWidth, width);
      | 'S','s':
          WriteFS(KeyIndex(arg), width);
      | 'O','o':  
          argWidth:= BasedCARDINALToChars(CARDINAL(arg), 8, digits);
          WriteNChars(s, digits, argWidth, width);
      | 'X','x':
          argWidth:= BasedCARDINALToChars(CARDINAL(arg), 16, digits);  
          WriteNChars(s, digits, argWidth, width);
      | 'R','r':  
          realFracwidth:= 0;
          IF (formatIndex<HIGH(format)) AND (format[formatIndex+1]='.') THEN
             INC(formatIndex);  realFracwidth:= FieldWidth();
          END;
          realKey:= FormatKey(arg);
          argWidth:= REALToChars(Keys[realKey].key.realValue, realFracwidth, digits);
          Keys[realKey].inUse := FALSE;
          WriteNChars(s, digits, argWidth, width);
      | 'T','t':
          setSize:= BITSPERWORD;
          IF (formatIndex<HIGH(format)) AND (format[formatIndex+1]='.') THEN
             INC(formatIndex);  setSize:= FieldWidth();
          END;
          argWidth:= BITSETToChars(BITSET(arg), setSize, digits);
          WriteNChars(s, digits, argWidth, width);
      ELSE
          Put(s, type);
      END;
      INC(argsUsed);
    ELSE
      IF ch = EscapeCh THEN
        INC(formatIndex);  ch:= TransEscape(format[formatIndex]);
      END;
      Put(s, ch);
    END;
    INC(formatIndex);
  END;
END WriteFVAR;

PROCEDURE WriteF0(s: Stream; format: ARRAY OF CHAR);
    VAR
        writeFargs: ArgArray;
    BEGIN
        WriteFVAR(s, writeFargs, format, 0);
    END WriteF0;

PROCEDURE WriteF1(s: Stream; format: ARRAY OF CHAR; arg1: WORD);
VAR writeFargs: ArgArray;
BEGIN
    writeFargs[0]:= arg1;
    WriteFVAR(s, writeFargs, format, 1);
END WriteF1;

PROCEDURE WriteF2(s: Stream; format: ARRAY OF CHAR; arg1: WORD; arg2: WORD);
VAR writeFargs: ArgArray;
BEGIN
    writeFargs[0]:= arg1;
    writeFargs[1]:= arg2;
    WriteFVAR(s, writeFargs, format, 2);
END WriteF2;

PROCEDURE WriteF3(s: Stream; format: ARRAY OF CHAR; arg1: WORD; arg2: WORD;
                                         arg3: WORD);
VAR writeFargs: ArgArray;
BEGIN
    writeFargs[0]:= arg1;
    writeFargs[1]:= arg2;
    writeFargs[2]:= arg3;
    WriteFVAR(s, writeFargs, format, 3);
END WriteF3;

PROCEDURE WriteF4(s: Stream; format: ARRAY OF CHAR; arg1: WORD; arg2: WORD;
                                         arg3: WORD; arg4: WORD);
VAR writeFargs: ArgArray;
BEGIN
    writeFargs[0]:= arg1;
    writeFargs[1]:= arg2;
    writeFargs[2]:= arg3;
    writeFargs[3]:= arg4;
    WriteFVAR(s, writeFargs, format, 4);
END WriteF4;

PROCEDURE WriteNChars(s: Stream; str: ARRAY OF CHAR; 
                                  actualWidth, width: CARDINAL);
VAR  i: CARDINAL;
BEGIN
  i:= actualWidth;
  WHILE i < width DO
    Put(s, SpaceCh); INC(i);
  END;
  i:= 0;
  WHILE i < actualWidth DO
    Put(s, str[i]);  INC(i);
  END;
END WriteNChars;


BEGIN
    InitKeys;
END WriteF .
