IMPLEMENTATION MODULE Context;

(*
    Title:      The abstract Context object for binding strings to values
    LastEdit:	"Wed Sep  4 18:00:46 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM CharCodes IMPORT  SpaceCh, NewLineCh;
FROM Arithmetic IMPORT IMax, IMin, CMax, CMin;
FROM Ctrl IMPORT CtrlPut;
FROM Strings IMPORT LengthC;
FROM Streams IMPORT Stream;

CONST EndStreamCh = 0C;

MODULE context;

    (* A Context is an N-ary tree. Each node defines a prefix string
       which is common to each of its children. This Name string may be
       empty in the case of a non-terminal one of whose sub-components
       is a substring of another. The root name is also empty.
       The actual command string corresponding to a bound procedure is
       therefore the concatenation of its Name to the Names of its tree
       ancestors.
    *)


    IMPORT
        ALLOCATE, DEALLOCATE, IMin, EnumerateProc, 
        CtrlPut, Stream, ContextProc;
    EXPORT
        Context, NullCh,
        NewContext, DisposeContext, CopyContext,
        NewChild, NewLevel, NewSibling, Collapse,
        DefineTerminal, SameName, Synonyms, 
        BoundName, PrintName, EnumerateContext;

    CONST
        NodeNameHigh = 9;   (* any value will do *)
        NullCh       = 0C;  (* used to represent the null string *)
    

    TYPE
        NAME = ARRAY [0..NodeNameHigh] OF CHAR;

        Context = POINTER TO ContextDATA;

        ContextDATA = RECORD
            Parent, Sibling:            Context;
            Name:                       NAME;
            High:                       INTEGER; (* of Name: -1 means empty *)
            CASE Terminal:              BOOLEAN OF
                TRUE:           Proc:   ContextProc;
            |   FALSE:          Child:  Context;
            END;
        END;


    PROCEDURE NewContext (VAR c: Context);
    BEGIN
        NEW (c);
        WITH c^ DO
            Parent := NIL;
            Child := NIL;
            Sibling := NIL;
            High := -1;
            Terminal := FALSE;
        END;
    END NewContext;


    PROCEDURE DisposeContext (VAR root: Context);
        VAR c: Context;
    BEGIN
        IF root # NIL THEN
            IF NOT root^.Terminal THEN
                WHILE root^.Child # NIL DO
                    c := root^.Child;
                    DisposeContext (c);
                END;
            END;
            IF root^.Parent # NIL THEN
                IF root^.Parent^.Child = root THEN
                    root^.Parent^.Child := root^.Sibling;
                ELSE c := root^.Parent^.Child;
                    WHILE c^.Sibling # root DO c := c^.Sibling END;
                    c^.Sibling := root^.Sibling;
                END;
            END;
            DISPOSE (root);
            root := NIL;
        END;
    END DisposeContext;


    PROCEDURE CopyContext (from: Context; VAR to: Context);
    BEGIN
        DoCopyContext (from, to, NIL);
    END CopyContext;
    
    PROCEDURE DoCopyContext (from: Context; VAR to: Context; parent: Context);
        BEGIN
            IF from # NIL THEN
                NEW (to);
                to^ := from^;
                to^.Parent := parent;
                IF NOT from^.Terminal THEN
                    DoCopyContext (from^.Child, to^.Child, to);
                END (* if *);
                DoCopyContext (from^.Sibling, to^.Sibling, parent);
            END;
        END DoCopyContext;


    PROCEDURE NewChild (parent: Context): Context;
        (* make a new first child of the parent *)
        VAR child: Context;
    BEGIN
        NEW (child);
        child^ := parent^;
        WITH child^ DO
            Parent := parent;
            IF parent^.Terminal THEN 
                Sibling := NIL;
            ELSE Sibling := Parent^.Child;
            END;
        END;
        WITH parent^ DO
            Terminal := FALSE;
            Child := child;
        END;
        RETURN child;
    END NewChild;

    
    PROCEDURE NewLevel (parent: Context): Context;
        (* Add a new tree level between parent and its children *)
        VAR level: Context;
    BEGIN
        NEW (level);
        level^ := parent^;
        WITH level^ DO
            Parent := parent;
            Sibling := NIL;
            IF NOT Terminal THEN (* parent's children are now level's *)
                WHILE parent^.Child # NIL DO
                    parent^.Child^.Parent := level;
                    parent^.Child := Parent^.Child^.Sibling;
                END;
            END;
        END;
        WITH parent^ DO
            Terminal := FALSE;
            Child := level;
        END;
        RETURN level;
    END NewLevel;


    PROCEDURE NewSibling (elder: Context): Context;
        VAR sibling: Context;
    BEGIN
        NEW (sibling);
        WITH sibling^ DO
            Parent := elder^.Parent;
            Sibling := elder^.Sibling;
            Child := NIL;
        END;
        elder^.Sibling := sibling;
        RETURN sibling;
    END NewSibling;

    
    PROCEDURE Collapse (child: Context);
        VAR i: INTEGER;
    BEGIN
        WITH child^ DO
            IF (Sibling = NIL) AND (Parent # NIL) AND (Parent^.Child = child)
            AND (High+1 + Parent^.High <= NodeNameHigh) THEN
                FOR i := 0 TO High DO
                    Parent^.Name [Parent^.High + i + 1] := Name [i]
                END;
                INC (Parent^.High, High + 1);
                Parent^.Terminal := Terminal;
                IF Terminal THEN
                    Parent^.Proc := Proc;
                ELSE Parent^.Child := Child;
                    WHILE Child # NIL DO (* enumerate child^.Children *)
                        Child^.Parent := Parent;
                        Child := Child^.Sibling;
                    END;
                END;
                DISPOSE (child);
            END;
        END;
    END Collapse;


    PROCEDURE DefineTerminal (c: Context; name: ARRAY OF CHAR;
        low, high: INTEGER; proc: ContextProc);
        (* Can create subnodes for c if name is too long *)
        VAR child: Context;
        VAR i: INTEGER;
    BEGIN
        WITH c^ DO
            Terminal := TRUE;
            High := IMin (NodeNameHigh, high - low);
            FOR i := 0 TO High DO
                Name [i] := name [low + i];
            END;
            INC (low, High + 1);
            IF high >= low THEN
                child := NewChild (c);
                Terminal := FALSE;
                DefineTerminal (child, name, low, high, proc);
            ELSE Proc := proc;
            END;
        END;
    END DefineTerminal;


    PROCEDURE SameName (c1, c2: Context): BOOLEAN;
        VAR i: INTEGER;
    BEGIN
        IF (c1 = NIL) OR (c2 = NIL) OR (c1^.Parent # c2^.Parent) THEN
            RETURN FALSE
        ELSIF c1^.High # c2^.High THEN
            RETURN FALSE
        ELSE FOR i := 0 TO c1^.High DO
                IF c1^.Name[i] # c2^.Name[i] THEN
                    RETURN FALSE
                END;
            END;
            RETURN TRUE;
        END;
    END SameName;


    PROCEDURE Synonyms (c: Context): BOOLEAN;
        VAR pred: Context;
    BEGIN
        IF c^.Sibling = NIL THEN
            RETURN TRUE
        ELSE WHILE c^.Sibling # NIL DO
                pred := c;
                c := c^.Sibling;
                IF NOT SameName (c, pred) THEN RETURN FALSE END;
            END;
            RETURN TRUE;
        END;
    END Synonyms;
    
    
    PROCEDURE ContextName (VAR l: CARDINAL; c: Context; VAR s: ARRAY OF CHAR);
        VAR i: INTEGER;
    BEGIN
        IF c # NIL THEN
            WITH c^ DO
                ContextName (l, Parent, s);
                FOR i := 0 TO High DO
                    IF l <= HIGH (s) THEN
                        s [l] := Name [i];
                        INC (l);
                    END;
                END;
            END;
        END;
    END ContextName;


    PROCEDURE BoundName (c: Context; VAR s: ARRAY OF CHAR);
        VAR l: CARDINAL;
    BEGIN
        l := 0;
        ContextName (l, c, s);
        IF l <= HIGH (s) THEN s[l] := 0C END;
    END BoundName;
    
    
    PROCEDURE PrintName (in, out: Stream; c: Context);
        VAR a: ARRAY [0..50] OF CHAR;
        VAR i: CARDINAL;
    BEGIN
        BoundName (c, a);
        FOR i := 0 TO HIGH (a) DO
            IF a[i] = 0C THEN 
                RETURN
            ELSE CtrlPut (out, a[i]);
            END
        END;
    END PrintName;
    
    PROCEDURE EnumerateContext (in, out: Stream;c: Context; p: EnumerateProc);
    BEGIN
        IF c # NIL THEN
            IF c^.Terminal THEN
                p (in, out, c);
            ELSE c := c^.Child;
                WHILE c # NIL DO 
                    EnumerateContext (in, out, c, p);
                    c := c^.Sibling;
                END;
            END;
        END;
    END EnumerateContext;



END context;
(**********)


MODULE cursor;
(************)

    (* A Cursor is used to record a position in a Context.
       The cursor can be moved by Advancing past a matched character,
       or by Retreating to the last matched character. The Context can also
       be Split at a cursor postion to add a new level to the Context tree,
       and a new node in the Context can be Inserted after a given position.
    *)

    IMPORT Context, Cursor, NewLevel, NewChild, NewSibling, Synonyms,
        EndStreamCh, NullCh, ContextProc;
    
    EXPORT 
        NewCursor, EndOf, Binding, Bound, Advance, Retreat,
        Insert, Split, Suffix;

    
    PROCEDURE NewCursor (VAR cursor: Cursor; c: Context);
    BEGIN
        WITH cursor DO
            context := c;
            next := 0;
        END;
    END NewCursor;


    PROCEDURE EndOf (c: Cursor; terminal: BOOLEAN): BOOLEAN;
    BEGIN
        WITH c DO
            IF next > context^.High THEN
                RETURN context^.Terminal = terminal
            ELSE RETURN FALSE
            END;
        END;
    END EndOf;

    
    PROCEDURE Binding (c: Cursor): Context;
    BEGIN
        WITH c DO
            IF EndOf (c, TRUE) THEN
                RETURN context;
            ELSIF EndOf (c,FALSE) 
            AND (context^.Child # NIL) AND (context^.Child^.High < 0) THEN
                RETURN context^.Child;
            ELSE RETURN NIL
            END;
        END;
    END Binding;


    PROCEDURE Bound (c: Cursor; VAR proc: ContextProc): BOOLEAN;
        VAR cont: Context;
    BEGIN
        cont := Binding (c);
        IF cont # NIL THEN 
            proc := cont^.Proc;
            RETURN TRUE
        ELSE RETURN FALSE
        END;
    END Bound;


    PROCEDURE Advance (VAR cursor: Cursor; c: CHAR): BOOLEAN;
        VAR cont: Context;
    BEGIN
        WITH cursor DO
            IF next <= context^.High THEN
                IF c = context^.Name [next] THEN
                    INC (next);
                    RETURN TRUE;
                ELSE RETURN FALSE
                END;
            ELSIF context^.Terminal THEN
                RETURN FALSE;
            ELSE cont := context^.Child;
                WHILE cont # NIL DO
                    IF (cont^.High >= 0) AND (c = cont^.Name [0]) THEN
                        context := cont;
                        next := 1;
                        RETURN TRUE;
                    ELSE cont := cont^.Sibling;
                    END;
                END;
                RETURN FALSE;
            END;
        END;
    END Advance;


    PROCEDURE Suffix (VAR c: Cursor): CHAR;
    BEGIN
        WITH c.context^ DO
            IF c.next <= High THEN
                INC (c.next);
                RETURN Name [c.next-1];
            ELSIF Terminal THEN
                RETURN EndStreamCh
            ELSIF (Child # NIL) AND Synonyms (Child) THEN
                c.context := Child;
                c.next := 0;
                RETURN Suffix (c);
            ELSE RETURN EndStreamCh
            END;
        END;
    END Suffix;


    PROCEDURE Retreat (VAR cursor: Cursor): BOOLEAN;
    BEGIN
        WITH cursor DO
            DEC (next);
            WHILE next <= 0 DO
                context := context^.Parent;
                IF context = NIL THEN RETURN FALSE END;
                next := context^.High + 1;
            END;
        END;
        RETURN TRUE;
    END Retreat;
    
    
    PROCEDURE Split (cursor: Cursor);
        VAR level: Context;
        VAR i: INTEGER;
    BEGIN
        WITH cursor DO
            level := NewLevel (context);
            WITH level^ DO
                IF next <= context^.High THEN
                    High := context^.High - next;
                    FOR i := 0 TO High DO 
                        Name [i] := context^.Name [next+i];
                    END;
                ELSE High := -1;
                END;
            END;
            WITH context^ DO
                IF next <= High THEN High := next - 1 END;
            END;
        END;
    END Split;

    
    PROCEDURE Insert (cursor: Cursor; char: CHAR): Context;
        VAR c, new: Context;
    BEGIN
        (* Assumes no child of cursor Context has name starting with char!
           char = NullCh puts new child at head of list
        *)
        c := cursor.context^.Child;
        IF (c = NIL) 
        OR (char = NullCh)
        OR ((c^.High >= 0) AND (char < c^.Name[0])) THEN
            new := NewChild (cursor.context);
        ELSE WHILE (c^.Sibling # NIL)
            AND ((c^.Sibling^.High < 0) OR (char > c^.Sibling^.Name [0])) DO
                c := c^.Sibling;
            END;
            new := NewSibling (c);
        END;
        RETURN new;
    END Insert;


END cursor;
(*********)



PROCEDURE Bind (c: Context; name: ARRAY OF CHAR; proc: ContextProc);
    VAR i, length: CARDINAL;
    VAR cursor: Cursor;
    VAR new: Context;
    VAR char: CHAR;
BEGIN
    length := CMax (LengthC (name), 1);
    IF (c # NIL) THEN
        NewCursor (cursor, c);
        i := 0;
        LOOP
            char := name [i];
            IF NOT Advance (cursor, char) THEN EXIT END;
            INC (i);
            IF i >= length THEN char := NullCh; EXIT END;
        END;
        IF NOT EndOf (cursor, FALSE) THEN Split (cursor) END;
        new := Insert (cursor, char);
        DefineTerminal (new, name, i, length - 1, proc);
    END;
END Bind;


PROCEDURE BindChar (c: Context; char: CHAR; proc: ContextProc);
    VAR name: ARRAY [0..0] OF CHAR;
BEGIN
    name [0] := char;
    Bind (c, name, proc);
END BindChar;


PROCEDURE Find (VAR cursor: Cursor; name: ARRAY OF CHAR): BOOLEAN;
    VAR i: INTEGER;
BEGIN
    FOR i := 0 TO IMax (LengthC (name), 1) - 1 DO
        IF NOT Advance (cursor, name [i]) THEN RETURN FALSE END;
    END;
    RETURN TRUE;
END Find;


PROCEDURE Lookup (c: Context; name: ARRAY OF CHAR;
    abbreviation: BOOLEAN; VAR proc: ContextProc): BOOLEAN;
    VAR cursor: Cursor;
BEGIN
    NewCursor (cursor, c);
    IF Find (cursor, name) THEN
        IF abbreviation THEN
            WHILE Suffix (cursor) # EndStreamCh DO END;
        END;
        RETURN Bound (cursor, proc)
    ELSE RETURN FALSE
    END;
END Lookup;


PROCEDURE Unbind (c: Context; name: ARRAY OF CHAR);
    (* Prune the tree at the highest level possible *)
    VAR binding, parent: Context;
    VAR cursor: Cursor;
BEGIN
    NewCursor (cursor, c);
    IF Find (cursor, name) THEN
        binding  := Binding (cursor);
        IF binding # NIL THEN
            WHILE (binding^.Sibling = NIL) AND (binding^.Parent^.Parent # NIL)
            AND (binding^.Parent^.Child = binding) DO
                binding := binding^.Parent;
            END;
            parent := binding^.Parent;
            DisposeContext (binding);
            IF parent^.Child # NIL THEN
                Collapse (parent^.Child)
            ELSE Collapse (parent);
            END;
        END;
    END;
END Unbind;


PROCEDURE UnbindChar (c: Context; name: CHAR);
    VAR s: ARRAY [0..0] OF CHAR;
BEGIN
    s [0] := name;
    Unbind (c, s);
END UnbindChar;


END Context.
