IMPLEMENTATION MODULE Dispatch;

(*
    Title: 	A General Command Dispatcher
    LastEdit:	"Thu Oct 22 18:19:44 1987" by Tony Mann
    Author: 	jd
		Cambridge University Computer Laboratory
*)

FROM SYSTEM IMPORT WORD;
FROM SystemTypes IMPORT LongINTEGER, LongCARDINAL, LongWORD;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM StreamRep IMPORT 
    Procs, PutProc, GetProc, BackspaceProc, FlushProc, DeleteProc, 
    SeekProc, TellProc, DefaultProcs, SetStreamError;
FROM Streams IMPORT
    Stream, StreamProcs, ErrorCode, SeekMode, Create, Put, Get, 
    Backspace, Delete, Seek, Tell;
FROM SysStreams IMPORT sysIn, sysOut;
FROM WriteF IMPORT WriteF0, WriteF1, S;
FROM CharCodes IMPORT CapitalCh, NewLineCh, SpaceCh;
FROM SCTermStream IMPORT CreateInOutput;
FROM Context IMPORT
    Context, NewContext, Bind, BindChar, SameName, PrintName, ContextProc, 
    EnumerateContext, Cursor, NewCursor, Advance, Retreat, Bound, Suffix,
    EndCh;
FROM Filter IMPORT CreateFilterInput;
FROM Arithmetic IMPORT IMax, IMin, CMax, CMin;
FROM Strings IMPORT String, Dispose, CopyCS;

CONST
    DeleteWordCh    =  35C;
    DeleteCh        = 177C;
    BellCh          =   7C;
    TabCh           =  11C;

    LineLength = 260;           (* max 260 chars in a command line *)


MODULE stream;
(************)

    (* In this module, we build an IO stream on a command line. It is much
       like BStringIO, except that both reading and writing take place on the
       same buffer, and the function USERFUNCTION has the effect of emptying
       the line buffer.
    *)
    
    IMPORT
        Procs, PutProc, GetProc, BackspaceProc, FlushProc, DeleteProc, 
        SeekProc, TellProc, DefaultProcs, SetStreamError, 
        Stream, StreamProcs, ErrorCode, SeekMode, Create, 
        Put, Get, Backspace, Delete, Seek, Tell, EndCh, 
        WORD, ALLOCATE, DEALLOCATE, LongWORD, LongINTEGER, 
        LongCARDINAL, LineLength;

    EXPORT CreateLineIO, SetLineLength;

    TYPE LINEDATA = RECORD
        pos:     INTEGER;                     (* next offset to read/write  *)
        hwm:     INTEGER;                     (* the highest offset written *)
        Line:    ARRAY [1..LineLength] OF CHAR; (* the line buffer          *)
    END;
    
    LINE = POINTER TO LINEDATA;


    PROCEDURE LinePutProc (s: Stream; a: WORD);
    VAR line: LINE;
    BEGIN
        line := LINE (s^.instanceData);
        WITH line^ DO
            IF (pos > LineLength) OR (pos < 1) THEN
                s^.status := StreamError;
            ELSE Line [pos] := CHAR (a);
                IF hwm < pos THEN hwm := pos END;
                INC (pos);
            END;
        END;
    END LinePutProc;
    
    PROCEDURE LineGetProc (s: Stream): WORD;
    VAR line: LINE;
    BEGIN
        line := LINE (s^.instanceData);
        WITH line^ DO
            IF (pos > hwm) OR (pos < 1) THEN 
                s^.status := EndOfStream;
                RETURN WORD (0);
            ELSE INC (pos);
                RETURN WORD(Line [pos-1]);
            END;
        END;
    END LineGetProc;
    
    PROCEDURE LineBackspaceProc (s: Stream);
    BEGIN
        LineSeekProc (s, Current, -1);
    END LineBackspaceProc;
    
    PROCEDURE LineDeleteProc (s: Stream);
    VAR line: LINE;
    BEGIN
        line := LINE (s^.instanceData);
        DISPOSE (line);
    END LineDeleteProc;
    
    PROCEDURE LineSeekProc (s: Stream; mode: SeekMode; to: LongINTEGER);
    VAR line: LINE;
        seeko: INTEGER;
    BEGIN
        line := LINE (s^.instanceData);
        seeko := INTEGER (to);
        WITH line^ DO
            CASE mode OF
                Beginning:   pos := 1 + seeko;
            |   Current:     INC (pos, seeko);
            |   End:         pos := hwm + seeko;
            END;
            IF pos < 1 THEN pos :=1 END;
        END;
    END LineSeekProc;
    
    PROCEDURE LineTellProc (s: Stream; VAR tell: LongCARDINAL);
    VAR line: LINE;
    BEGIN
        line := LINE (s^.instanceData);
        tell := LongCARDINAL (line^.pos - 1);
    END LineTellProc;


    PROCEDURE SetLineLength (s: Stream; c: CARDINAL);
        VAR line: LINE;
    BEGIN
        line := LINE (s^.instanceData);
        line^.hwm := INTEGER (c) -1;
    END SetLineLength;
    
    PROCEDURE CreateLineIO (): Stream;
    VAR line: LINE;
        procs: Procs;
    BEGIN
        NEW (line);
        WITH line^ DO
            pos := 1; hwm := 0;
        END;
        procs := DefaultProcs (8);
        WITH procs^ DO
            Put := LinePutProc;
            Get := LineGetProc;
            Backspace := LineBackspaceProc;
            Delete := LineDeleteProc;
            Seek := LineSeekProc;
            Tell := LineTellProc;
        END;
        RETURN Create (StreamProcs (procs), line);
    END CreateLineIO;

END stream;
(*********)

(* State variables record the state of the world on entry to
   Dispatch so that recursive calls can discover it
*)

TYPE DISPATCHSTATE = 
    RECORD
        Immediate,
        Delayed:        Context;
        InputStream,
        OutputStream:   Stream;
        Prompt:         String;
    END;

VAR Top: DISPATCHSTATE;  (* set by most recent dispatcher invocation *)


PROCEDURE ChangePrompt (to: ARRAY OF CHAR);
BEGIN
    Dispose (Top.Prompt);
    Top.Prompt := CopyCS (to);
END ChangePrompt;

PROCEDURE FindContexts (VAR immediate, delayed: Context);
BEGIN
    IF level > 0 THEN
        immediate := Top.Immediate;
        delayed := Top.Delayed;
    END;
END FindContexts;


PROCEDURE ImmediateContext (): Context;
    VAR c: Context;
BEGIN
    NewContext (c);
    BindChar (c, EndCh, Quit);
    RETURN c;
END ImmediateContext;


PROCEDURE DelayedContext (): Context;
    VAR c: Context;
BEGIN
    NewContext (c);
    BindChar (c, "?", Describe);
    Bind     (c, "quit", Quit);
    RETURN c;
END DelayedContext;

(*                  Predefined Actions                                   *)

VAR pred: Context;
VAR first: BOOLEAN;


PROCEDURE Leaf (in, out: Stream; c: Context);
BEGIN
    IF NOT SameName (c, pred) THEN
        IF NOT first THEN WriteF0 (out, ", ") END;
        first := FALSE;
        PrintName (in, out, c);
        pred := c;
    END;
END Leaf;


PROCEDURE Describe (in, out: Stream);
BEGIN
    IF level > 0 THEN
        first := TRUE;
        pred := Top.Delayed;
        EnumerateContext (in, out, Top.Delayed, Leaf);
        Put (out, NewLineCh);
    END;
END Describe;


PROCEDURE DescribeAll (in, out: Stream);
BEGIN
    IF level > 0 THEN
        first := TRUE;
        pred := Top.Immediate;
        EnumerateContext (in, out, Top.Immediate, Leaf);
        EnumerateContext (in, out, Top.Delayed, Leaf);
        Put (out, NewLineCh);
    END;
END DescribeAll;


PROCEDURE Quit (in, out: Stream);
BEGIN
    DEC (level);
END Quit;

(***        End of Predefined Actions               ***)


PROCEDURE FindRealInput (): Stream;
BEGIN
    IF level > 0 THEN
        RETURN Top.InputStream; 
 (* ELSE intentional abort here! *)
    END;
END FindRealInput;


CONST backspace = 10C;

PROCEDURE Dispatch (instream, outstream: Stream; 
                    imm, del: Context; prompt: ARRAY OF CHAR);
    VAR Old: DISPATCHSTATE;
    VAR reflect, sc, filter, line: Stream;
    VAR cursor: Cursor;
    VAR proc: ContextProc;
    VAR ch, suffixch: CHAR;
    VAR newch, bound: BOOLEAN;
    VAR lev, layoutlength, length, matchedlength, argpos: CARDINAL;
    VAR w: WORD;
    VAR term: BOOLEAN;
    VAR size: ARRAY [1 .. LineLength] OF INTEGER;
    TYPE STATE = (Layout, Command, Args);
    VAR state: STATE;

    PROCEDURE Reflect (ch: CHAR);
    BEGIN
        INC (length);
        IF length > LineLength THEN
            Put (reflect, WORD (BellCh));
        ELSE
            IF (ch < SpaceCh) AND (ch # NewLineCh) THEN
                size [length] := 2;
                Put (reflect, WORD ('^'));
                Put (reflect, WORD (ORD (ch) + 100B));
            ELSE
                IF ch = SpaceCh THEN
                    size [length] := -1; 
                ELSE size [length] := 1
                END;
                Put (reflect, WORD (ch));
            END;
        END (* if *);
        Put (line, WORD (ch));
    END Reflect;

    PROCEDURE DeleteChar;
        VAR i: CARDINAL;
    BEGIN
        IF length > 0 THEN
            IF length <= LineLength THEN
                FOR i := 1 TO ABS (size [length]) DO
                    Put (reflect, WORD (backspace)); 
                    Put (reflect, WORD (SpaceCh)); 
                    Put (reflect, WORD (backspace));
                END;
            END (* if *);
            Seek (line, Current, -1);
            SetLineLength (line, length);
            DEC (length);
            IF length <= layoutlength THEN 
                state := Layout;
                bound := FALSE
            ELSIF length < matchedlength THEN
                IF Retreat (cursor) THEN DEC (matchedlength) END;
            ELSIF length = matchedlength THEN
                state := Command
            END;
        END;
    END DeleteChar;

    PROCEDURE DeleteWord;
    BEGIN
        WHILE (length > 0) AND ((length > LineLength) OR (size [length] < 0))
        DO DeleteChar END;
        WHILE (length > 0) AND (size [length] > 0) DO DeleteChar END;
    END DeleteWord;

BEGIN
    INC (level);
    lev := level;
    term := instream = NIL;
    reflect := outstream;
    Old := Top;         (* save current dispatcher variables *)
    IF term THEN
        sc := CreateInOutput ();
        IF sc^.status # Success THEN
            Delete (sc); DISPOSE (sc);
            term := FALSE;
            instream := sysIn;
            reflect := sysOut;
        ELSE
            instream := sc;
            reflect := sc;
        END (* if *);
    END;
    IF reflect = NIL THEN reflect := sysOut END;
    WITH Top DO     (* define new dispatcher variables *)
        InputStream := instream;
        OutputStream := reflect;
        Immediate := imm;
        Delayed := del;
        Prompt := CopyCS (prompt);
    END;
    filter := CreateFilterInput (Top.InputStream, imm);
    line := CreateLineIO ();

    LOOP
        length := 0;
        bound := FALSE;
        newch := TRUE;
        SetLineLength (line, 0);
        Seek (line, Beginning, 0);
        WriteF1 (reflect, "%S", S(Top.Prompt));
        state := Layout;
        LOOP
            IF newch THEN
                ch := CHAR (Get (filter)); 
                IF level < lev THEN Backspace (filter); EXIT END; 
                IF ch = TabCh THEN ch := SpaceCh END;
            END;
            newch := TRUE;
            CASE state OF
                Layout:
                    IF ch = NewLineCh THEN 
                        Reflect (ch);
                        EXIT;
                    ELSIF ch = SpaceCh THEN
                        Reflect (ch);
                    ELSIF ch = DeleteCh THEN
                        DeleteChar;
                    ELSIF ch = DeleteWordCh THEN
                        DeleteWord;
                    ELSE NewCursor (cursor, del);
                        layoutlength := length;
                        matchedlength := length;
                        state := Command; newch := FALSE;
                    END;
            |   Command:
                    IF (length = matchedlength) 
                    AND Advance (cursor, ch) THEN
                        INC (matchedlength);
                        Reflect (ch);
                    ELSIF (ch = NewLineCh) OR (ch = SpaceCh) THEN
                        IF length = matchedlength THEN
                            LOOP
                                bound := Bound (cursor, proc);
                                IF bound THEN
                                    newch := FALSE; state := Args; EXIT
                                END;
                                suffixch := Suffix (cursor);
                                IF suffixch = EndCh THEN
                                    Put (reflect, WORD (BellCh)); 
                                    newch := TRUE; EXIT
                                END;
                                Reflect (suffixch);
                                INC (matchedlength);
                            END;
                        ELSE bound := FALSE; newch := FALSE; state := Args
                        END;
                        argpos := length;
                    ELSIF ch = DeleteCh THEN
                        DeleteChar;
                    ELSIF ch = DeleteWordCh THEN
                        DeleteWord;
                    ELSE IF length = matchedlength THEN 
                            Put (reflect, WORD (BellCh)) 
                        END;
                        Reflect (ch);
                    END;
            |   Args:
                    IF ch = NewLineCh THEN
                        IF NOT bound THEN WriteF0 (reflect, "  ??") END;
                        Reflect (ch);
                        EXIT
                    ELSIF ch = DeleteCh THEN
                        DeleteChar;
                    ELSIF ch = DeleteWordCh THEN
                        DeleteWord;
                    ELSE Reflect (ch)
                    END;
            END;
        END;
        IF level < lev THEN 
            EXIT
        ELSIF bound THEN
            ch := CHAR (Get (line)); (* just in case a Backspace is pending *)
            Seek (line, Beginning, LongCARDINAL (argpos));
            proc (line, reflect);
            IF level < lev THEN EXIT END;
        END;
    END;
    Delete (filter); DISPOSE (filter);
    IF term THEN Delete (sc); DISPOSE (sc) END;
    Delete (line); DISPOSE (line);
    Dispose (Top.Prompt);
    Top := Old;     (* restore previous dispatcher variables *)
END Dispatch;


VAR i: CARDINAL;

BEGIN
    level := 0;

END Dispatch.
