IMPLEMENTATION MODULE GKSClip;

(*
    Title: 	GKS Shape Clipping
    LastEdit:	"Fri Aug  1 10:10:46 1986"
    Author: 	jd
		Cambridge University Computer Laboratory
*)

(*  These algorithms are taken from "Principles of Interactive
    Computer Graphics", Newmann &  Sproull, McGraw Hill, 2nd edition.
    See also Foley & Van Dam "Fundamentals of Interactive Computer Graphics",
    Addison & Wesley, for a better exposition of the polygon clipping
    algorithm of Sutherland & Hodgson.
*)

(*  $T-, $R- *)

FROM Vector IMPORT RVEC, AXIS, RBOX, BOUND, REqual;

FROM GKSTrans IMPORT
    ValidBox;

FROM Arithmetic IMPORT CMin, RMin, RMax, SquareRoot;


TYPE
    EDGE   = (Left, Right, Below, Above);
    SIDE   = (Inside, Outside);

    EDGES  = SET OF EDGE;
    SIDES  = SET OF SIDE;

VAR Visible: EDGES;


PROCEDURE AreaCode (VAR p: RVEC; VAR r: RBOX): EDGES;
    VAR c: EDGES;
BEGIN
    c := Visible;
    IF p [X] < r [Low, X] THEN
        INCL (c, Left)
    ELSIF p [X] > r [High, X] THEN
        INCL (c, Right)
    END;
    IF p [Y] < r [Low, Y] THEN
         INCL (c, Below)
    ELSIF p [Y] > r [High, Y] THEN
         INCL (c, Above)
    END;
    RETURN c;
END AreaCode;



PROCEDURE LineSide (VAR p: RVEC; VAR r: RBOX; edge: EDGE): SIDE;
    (* is the point inside or outside one of the infinite
       bounding lines of a box
    *)
    VAR d: REAL;
BEGIN
    CASE edge OF
        Left:     d := p [X] - r [Low, X];
      | Right:    d := r [High, X] - p [X];
      | Above:    d := r [High, Y] - p [Y];
      | Below:    d := p [Y] - r [Low, Y];
    ELSE END;
    IF d >= 0.0 THEN
        RETURN Inside
    ELSE RETURN Outside
    END;
END LineSide;


(* Intersection returns TRUE only if start and end lie on different
   sides of the box edge. It returns FALSE if they lie on the same
   side
*)    

PROCEDURE Intersection (VAR start, end: RVEC; VAR int: RVEC;
    VAR r: RBOX; edge: EDGE): BOOLEAN;
    VAR Dx, Dy: REAL;
    VAR startside, endside: SIDE;
BEGIN
    startside := LineSide (start, r, edge);
    endside := LineSide (end, r, edge);
    IF startside = endside THEN
        RETURN FALSE;
    ELSE Dx := end [X] - start [X];
        Dy := end [Y] - start [Y];
        CASE edge OF
            Left:
                int [Y] := start [Y] + (r [Low, X] - start [X]) * Dy / Dx;
                int [X] := r [Low, X];
          | Right:  
                int [Y] := start [Y] + (r [High, X] - start [X]) * Dy / Dx;
                int [X] := r [High, X];
          | Below:
                int [X] := start [X] + (r [Low, Y] - start [Y]) * Dx / Dy;
                int [Y] := r [Low, Y];
          | Above:
                int [X] := start [X] + (r [High, Y] - start [Y]) * Dx / Dy;
                int [Y] := r [High, Y];
        ELSE END;
        RETURN TRUE;
    END;
END Intersection;


PROCEDURE InPolygon (VAR p: RVEC; high: INTEGER;
    VAR poly: ARRAY OF RVEC): BOOLEAN;
    VAR r: RBOX;
        q, int: RVEC;
        i, crossings: INTEGER;
    (* Count the number of intersections of the edges of the polygon with
       the semi-infinite line x = p [X], y >= p [Y]. The point is inside
       the polygon iff this number is odd.
    *)
BEGIN
    crossings := 0;
    r [Low] := p;
    r [High] := p;
    q := poly [high];
    FOR i := 0 TO high DO
        IF (poly [i, X] # p [X]) (* otherwise we would double count *)
        AND Intersection (q, poly[i], int, r, Left) 
        AND (int [Y] >= p [Y]) THEN
            INC (crossings)
        END;
        q := poly [i];
    END;
    RETURN ODD (crossings);
END InPolygon;


PROCEDURE ClipLine (VAR start, end: RVEC; VAR box: RBOX): BOOLEAN;
    (* Explicit calculation of intersections from Newman & Sproull *)
    VAR c: ARRAY BOOLEAN OF EDGES;
    VAR p: ARRAY BOOLEAN OF RVEC;
    VAR Dx, Dy: REAL;
        b: BOOLEAN;
BEGIN
    c [FALSE] := AreaCode (start, box);
    c [TRUE] := AreaCode (end, box);
    IF c [FALSE] + c [TRUE] = Visible THEN
        RETURN TRUE
    ELSIF c [FALSE] * c [TRUE] # Visible THEN
        RETURN FALSE
    ELSE p [FALSE] := start;
        p [TRUE] := end;
        Dx := end [X] - start [X];
        Dy := end [Y] - start [Y];
        FOR b := FALSE TO TRUE DO
            IF Left IN c [b] THEN
                p [b, X] := box [Low, X];
                p [b, Y] := start [Y] + (box [Low, X] - start [X]) * Dy / Dx;
            ELSIF Right IN c [b] THEN
                p [b, X] := box [High, X];
                p [b, Y] := start [Y] + (box [High, X] - start [X]) * Dy / Dx;
            END;
            c [b] := AreaCode (p [b], box);
            IF Below IN c [b] THEN
                p [b, X] := start [X] + (box [Low, Y] - start [Y]) * Dx / Dy;
                p [b, Y] := box [Low, Y];
            ELSIF Above IN c [b] THEN
                p [b, X] := start [X] + (box [High, Y] - start [Y]) * Dx / Dy;
                p [b, Y] := box [High, Y];
            END;
        END;
        start := p [FALSE];
        end := p [TRUE];
        RETURN TRUE;
    END;
END ClipLine;


PROCEDURE ClipBox (VAR r: RBOX; VAR box: RBOX): BOOLEAN;
    VAR llcode, urcode: EDGES;
BEGIN
    llcode := AreaCode (r [Low], box);
    urcode := AreaCode (r [High], box);
    IF (llcode = Visible) AND (urcode = Visible) THEN
        RETURN TRUE
    ELSIF llcode * urcode # Visible THEN
        RETURN FALSE
    ELSE
        IF Left IN llcode THEN
            r [Low, X] := box [Low, X]
        END;
        IF Below IN llcode THEN
            r [Low, Y] := box [Low, Y]
        END;
        IF Right IN urcode THEN
            r [High, X] := box [High, X]
        END;
        IF Above IN urcode THEN
            r [High, Y] := box [High, Y]
        END;
        RETURN TRUE;
    END;
END ClipBox;


PROCEDURE ClipPoints (fromhigh: INTEGER; VAR from: ARRAY OF RVEC;
    VAR tohigh: INTEGER; VAR to: ARRAY OF RVEC;
    VAR box: RBOX): BOOLEAN;
    VAR i, tonext: INTEGER;
BEGIN
    tonext := 0;
    i := 0;
    WHILE (i <= fromhigh) AND (tonext <= tohigh) DO
        IF AreaCode (from [i], box) = Visible THEN
            to [tonext] := from [i];
            INC (tonext);
        END;
        INC (i);
    END;
    tohigh := tonext - 1;
    RETURN tonext > 0;
END ClipPoints;


PROCEDURE ClipPolyLine (VAR fromlow, fromhigh: INTEGER;
    VAR from: ARRAY OF RVEC; VAR tohigh: INTEGER;
    VAR to: ARRAY OF RVEC; VAR box: RBOX): BOOLEAN;
    VAR p: RVEC;
BEGIN
    IF HIGH (to) < 1 THEN RETURN FALSE END;
    (* find a visible segment *)
    REPEAT
        IF fromlow >= fromhigh THEN RETURN FALSE END;
        to [0] := from [fromlow];
        INC (fromlow);
        to [1] := from [fromlow];
    UNTIL ClipLine (to [0], to [1], box);
    tohigh := 1;
    (* continue the visible segment until we go outside the box *)
    WHILE (fromlow < fromhigh) AND (tohigh < INTEGER (HIGH (to))) 
    AND REqual (to [tohigh], from [fromlow], 0.0) DO
        INC (fromlow);
        INC (tohigh);
        to [tohigh] := from [fromlow];
        IF AreaCode (to [tohigh], box) # Visible THEN
            IF ClipLine (to [tohigh-1], to [tohigh], box) THEN
                RETURN TRUE
            ELSE HALT
            END;
        END;
    END;
    RETURN TRUE;
END ClipPolyLine;


(* To clip a polygon against a box, proceed in three stages:
    1) check degenerate cases of entire polygon visible or invisible.
    2) use Sutherland/Hodgson algorithm to clip polygon by clipping
       it successively against each edge of the box. This is
       implemented recursively by the procedures VertexClip and Close.
    3) if no crossings of the box boundary were found in 2), 
       pick one vertex of the box, and find whether it is
       inside or outside the box. Return the box or nothing
       as the clipped polygon respectively
*)

PROCEDURE ClipPolygon (fromhigh: INTEGER; VAR from: ARRAY OF RVEC;
    VAR tohigh: INTEGER; VAR to: ARRAY OF RVEC;
    VAR r: RBOX): BOOLEAN;

    VAR i: INTEGER;

    PROCEDURE Output (VAR p: RVEC; edge: EDGE);
    BEGIN
        IF edge = Above THEN
            IF tohigh < INTEGER (HIGH (to)) THEN
                IF (tohigh < 0) 
                OR (to [tohigh, X] # p [X]) OR (to [tohigh, Y] # p [Y]) THEN
                    INC (tohigh); 
                    to [tohigh] := p; 
                END;
            END
        ELSE INC (edge);
            VertexClip (p, edge);
        END;
    END Output;

    VAR F, S: ARRAY EDGE OF RVEC; (* first and predecessor points *)
    VAR FirstPoint: ARRAY EDGE OF BOOLEAN; (* TRUE on first entry *)


    PROCEDURE VertexClip (VAR P: RVEC; edge: EDGE);
        VAR I: RVEC;
    BEGIN
        IF FirstPoint [edge] THEN
            F [edge] := P;
            FirstPoint [edge] := FALSE;
        ELSIF Intersection (S [edge], P, I, r, edge) THEN
            Output (I, edge);
        END;
        S [edge] := P;
        IF LineSide (P, r, edge) # Outside THEN
            Output (P, edge);
        END;
    END VertexClip;

    PROCEDURE Close (edge: EDGE);
        VAR I: RVEC;
    BEGIN
        IF Intersection (S [edge], F [edge], I, r, edge) THEN
            Output (I, edge);
        END;
        IF edge # Above THEN INC (edge); Close (edge) END;
    END Close;

    VAR c, union, intersection: EDGES;
        linecode: SIDES;
        e: EDGE;

BEGIN
    union := EDGES {};
    intersection := EDGES {Left, Right, Above, Below};
    FOR i := 0 TO CMin (fromhigh, HIGH (to)) DO
        to [i] := from [i];  (* assume it's all visible *)
        c := AreaCode (from [i], r);
        union := union + c;
        intersection := intersection * c;
    END;
    IF union = Visible THEN
        tohigh := CMin (fromhigh, HIGH (to));
        RETURN TRUE
    ELSIF intersection # Visible THEN
        RETURN FALSE;
    ELSE tohigh := -1;
        FOR e := Left TO Above DO FirstPoint [e] := TRUE END;
        FOR i := 0 TO fromhigh DO
            VertexClip (from [i], Left);
        END;
        Close (Left);
        IF tohigh >= 0 THEN
            RETURN TRUE;
        ELSE to [0] := r [Low];
            IF InPolygon (to [0], fromhigh, from) THEN
                to [1, Y] := to [0, Y]; to [1, X] := r [High, X];
                to [2] := r [High];
                to [3, Y] := to [2, Y]; to [3, X] := r [Low, X];
                tohigh := 3;
                RETURN TRUE;
            ELSE RETURN FALSE;
            END;
        END;
    END;
END ClipPolygon;


BEGIN
    Visible := EDGES {};
END GKSClip.
