IMPLEMENTATION MODULE GKSTrans;

(*
    Title: 	GKS Transformation Module
    LastEdit:	"Fri May 17 16:45:47 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)

(* $T-,$R- *)

(* A transformation is a 3 * 3 matrix T whose third column is [0, 0, 1].
   To transform a point (x, y) under T, form the row vector v = [x, y, 1]
   and compute the row vector  [x', y', 1] = v T.

   If T does not rotate then the non-zero entries of T are:
        T[1,1] = x scale factor,   T[3,1] = x translation  
        T[2,2] = y scale factor,   T[3,2] = y translation
   (See Newman and Sproull, "Principles of Interactive Computer Graphics")

   A transformation each of whose entries is within epsilon of the identity
   transformation's entries is considered to be the identity transformation.

   To compose transformations T1 and T2, we do ordinary matrix multiplication.
*)

FROM Vector IMPORT RVEC, AXIS, RSEG, RBOX, BOUND, RSize, RVec, RBox, RBoxSeg;

FROM GKSTypes IMPORT WS, TRANSFORM, SEGMENT;

FROM MathLib IMPORT Radians, Sin, Cos;

FROM Storage IMPORT ALLOCATE;

FROM Arithmetic IMPORT RMin, RMax, SquareRoot;

CONST epsilon = 0.001;


PROCEDURE ValidBox (VAR r: RBOX): BOOLEAN;
    VAR t: RBOX;
        temp: REAL;
        z: AXIS;
BEGIN
    FOR z := X TO Y DO
        IF r [Low, z] > r [High, z] THEN
            temp := r [Low, z];
            r [Low, z] := r [High, z];
            r [High, z] := temp;
        ELSIF r [Low, z] = r [High, z] THEN
            RETURN FALSE
        END;
    END;
    RETURN TRUE
END ValidBox;



PROCEDURE ConstructTransform (VAR window, viewport: RBOX; 
    EqualScale: BOOLEAN; VAR T: TRANSFORM);
    VAR dx, dy: REAL;
        scale: RVEC;
        vs, ws: RSEG;
        z: AXIS;
        T1, T2, S: TRANSFORM;
BEGIN
    TranslationTransform (-window[Low, X], -window[Low, Y], T1);
    FOR z := X TO Y DO
        RBoxSeg (viewport, z, vs);
        RBoxSeg (window, z, ws);
        scale [z] := RSize (vs) / RSize (ws);
    END;
    IF EqualScale THEN
        scale [X] := RMin (scale [X], scale [Y]);
        scale [Y] := scale [X];
    END;
    ScalingTransform (scale [X], scale [Y], S);
    TranslationTransform (viewport[Low, X], viewport[Low, Y], T2);
    ComposeTransforms (T1, S, T);
    ComposeTransforms (T, T2, T)
END ConstructTransform;


PROCEDURE TranslationTransform (Dx, Dy: REAL; VAR T: TRANSFORM);
BEGIN
    T := IdentityTransform;
    WITH T DO
        Matrix [3,1] := Dx;
        Matrix [3,2] := Dy;
        Identity := (ABS (Dx) < epsilon) AND (ABS (Dy) < epsilon);
    END;
END TranslationTransform;


PROCEDURE ScalingTransform (Sx, Sy: REAL; VAR T: TRANSFORM);
BEGIN
    T := IdentityTransform;
    WITH T DO
        Matrix [1,1] := Sx;
        Matrix [2,2] := Sy;
        Identity := (ABS (Sx-1.0) < epsilon) AND (ABS (Sy-1.0) < epsilon);
    END;
END ScalingTransform;


PROCEDURE RotationTransform (degrees: REAL; VAR T: TRANSFORM);
    (* NB Counter-clockwise rotation- NOT as in Newman & Sproull *)
    VAR r: REAL;
BEGIN
    r := Radians (degrees);
    T := IdentityTransform;
    WITH T DO
        Matrix [1,1] := Cos (r);
        Matrix [1,2] := Sin (r);
        Matrix [2,1] := 0.0 - Sin (r);
        Matrix [2,2] := Cos (r);
        CheckIdentity (T);
    END;
END RotationTransform;


PROCEDURE ComposeTransforms (t1, t2: TRANSFORM; VAR T: TRANSFORM);
(*
    **************** CODE GENERATOR BUG *****************
    If ever t1 or t2 are the same as T, then updating T
    will also update the other. Hence these must be copied.
*)
VAR i, j, k: INTEGER;
    T1, T2: TRANSFORM;
BEGIN
    T1 := t1;
    T2 := t2;
    IF T1.Identity THEN
        T := T2;
    ELSIF T2.Identity THEN
        T := T1;
    ELSE WITH T DO
            FOR i := 1 TO 3 DO
                FOR j := 1 TO 3 DO
                    Matrix [i,j] := 0.0;
                    FOR k := 1 TO 3 DO
                         Matrix [i,j] := Matrix [i,j] + 
                            T1.Matrix[i, k] * T2.Matrix[k, j];
                    END;
                END;
            END;
        END;
        CheckIdentity (T);
    END;
END ComposeTransforms;


PROCEDURE CheckIdentity (VAR T: TRANSFORM);
    VAR i, j: INTEGER;
BEGIN
    WITH T DO
        Identity := FALSE;
        FOR i := 1 TO 3 DO
            FOR j := 1 TO 3 DO
                IF i = j THEN
                    IF ABS (Matrix [i,j] - 1.0) > epsilon THEN RETURN END;
                ELSE IF ABS (Matrix [i,j]) > epsilon THEN RETURN END;
                END;
            END;
        END;
        Identity := TRUE;
    END;
END CheckIdentity;


PROCEDURE TransformPoint (VAR T: TRANSFORM; VAR p: RVEC);
    VAR temp: RVEC;
BEGIN
    WITH T DO
        IF NOT Identity THEN
            temp[X] := Matrix [1,1] * p[X] 
                + Matrix [2,1] * p[Y] + Matrix [3,1];
            temp[Y] := Matrix [1,2] * p[X]
                + Matrix [2,2] * p[Y] + Matrix [3,2];
            p := temp;
        END;
    END;
END TransformPoint;


PROCEDURE UnTransformPoint (VAR T: TRANSFORM; VAR p: RVEC): BOOLEAN;
BEGIN
    WITH T DO
        IF Identity THEN
            RETURN TRUE
        ELSIF (Matrix [1,2] = 0.0) AND (Matrix [2,1] = 0.0) 
        AND (Matrix [1,1] # 0.0) AND (Matrix [2,2] # 0.0) THEN
            p [X] := (p [X] - Matrix [3,1]) / Matrix [1,1];
            p [Y] := (p [Y] - Matrix [3,2]) / Matrix [2,2];
            RETURN TRUE;
        END (* if *);
        RETURN FALSE;
    END (* with *);
END UnTransformPoint;


PROCEDURE TransformVector (VAR T: TRANSFORM; VAR v: RVEC);
    VAR origin: RVEC;
BEGIN
    RVec (origin, 0.0, 0.0);
    TransformPoint (T, origin);
    TransformPoint (T, v);
    v[X] := v[X] - origin[X]; 
    v[Y] := v[Y] - origin[Y];
END TransformVector;


PROCEDURE TransformBox (VAR T: TRANSFORM; VAR r: RBOX;
    VAR parallelogram: ARRAY OF RVEC);
    VAR i: INTEGER;
BEGIN
    parallelogram [0] := r [Low];
    parallelogram [1] := parallelogram [0];
    parallelogram [1, X] := r [High, X];
    parallelogram [2] := r [High];
    parallelogram [3] := parallelogram [2];
    parallelogram [3, X] := r[Low, X];
    FOR i := 0 TO 3 DO
        TransformPoint (T, parallelogram [i]);
    END;
END TransformBox;


PROCEDURE TransformPolyLine (VAR T: TRANSFORM;
    high: INTEGER; VAR points: ARRAY OF RVEC);
    VAR i: INTEGER;
BEGIN
    IF NOT T.Identity THEN
        FOR i := 0 TO high DO
            TransformPoint (T, points [i]);
        END;
    END;
END TransformPolyLine;


VAR i, j: INTEGER;

BEGIN
    RBox (IdentityBox, 0.0, 0.0, 1.0, 1.0);
    WITH IdentityTransform DO
        FOR i := 1 TO 3 DO
            FOR j := 1 TO 3 DO
                IF i = j THEN
                    Matrix [i,j] := 1.0
                ELSE Matrix [i,j] := 0.0;
                END;
            END;
        END;
        Identity := TRUE;
    END;
END GKSTrans.
