(*
    Title: 	User interface to RPC
    LastEdit:	"Mon Feb  1 13:34:47 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE RPC;


IMPORT
  Iterator, Marshal, NetMember, PktBufs, RPCLocal, RPCNets, 
  RPCRemote, RPCTable;

FROM RPCBasics   IMPORT CallEnd, RC, MemberRC;
FROM SYSTEM      IMPORT ADDRESS, WORD, ADR, TSIZE, BYTESPERWORD;
FROM SystemTypes IMPORT String;


PROCEDURE ClientArguments(
  target:  Interface; 
  proc, 
  version: CARDINAL;
  protocol:CallSemantics
 ):Marshal.T;
  VAR 
    p:PktBuf;
    m:Marshal.T;
  BEGIN
    p := PktBufs.New();
    p^.location[ServerEnd] := target.location;
    p^.serviceId := target.service;
    p^.protocolHdr.protocol := protocol;
    m := Marshal.New(p);
    m^.marshal := TRUE;
    Marshal.AWord(proc, m);
    Marshal.AWord(version, m);
    RETURN m
  END ClientArguments;


PROCEDURE Call(m:Marshal.T):RC;
  VAR 
    p: PktBuf;
    rc:RC;
  BEGIN
    p := m^.iPktBuf;
    rc := CallNoDiscard(m);
    PktBufs.Discard(p);
    RETURN rc
  END Call;

    
PROCEDURE CallNoDiscard(m:Marshal.T):RC;
  VAR 
    c:RC;
    p:PktBuf;
  BEGIN
    IF SameLocation(RPCNets.thisLocation,
                    m^.iPktBuf^.location[ServerEnd]) THEN 
      c := RPCLocal.Call(m^.iPktBuf)
    ELSE
      c := RPCRemote.Call(m^.iPktBuf)
    END;
    IF c = OK THEN
      m^.marshal := FALSE;
      m^.cPktBuf := m^.iPktBuf;
      Marshal.Start(m);
      Marshal.AWord(c, m);
      IF c # OK THEN PktBufs.Discard(m^.iPktBuf) END
    END;
    IF c # OK THEN Marshal.Discard(m) END;
    RETURN c
  END CallNoDiscard;


PROCEDURE EndOfResults(VAR m:Marshal.T (* InOut *)):RC;
  VAR
    b:BOOLEAN;
  BEGIN
    b := Marshal.EndOfArgs(m);
    PktBufs.Discard(m^.iPktBuf);
    Marshal.Discard(m);
    IF NOT b THEN RETURN ResultError END;
    RETURN OK
  END EndOfResults;


PROCEDURE ServerProcedure(
     p:      PktBuf;
 VAR m:      Marshal.T;
 VAR version:CARDINAL
 ):CARDINAL;
  (* Initialises Marshal.T for unmarshalling *)
  VAR
    proc:CARDINAL;
  BEGIN
    m := Marshal.New(p);
    m^.marshal := FALSE;
    Marshal.AWord(proc, m);
    Marshal.AWord(version, m);
    RETURN proc
  END ServerProcedure;


PROCEDURE ServerProcedureVCheck(
     p:      PktBuf;
 VAR m:      Marshal.T; (* Out *)
 VAR proc:   CARDINAL;  (* Out *)
     version:CARDINAL
 ):RC;
  VAR
    callVersion:CARDINAL;
  BEGIN
    proc := ServerProcedure(p, m, callVersion);
    IF version # callVersion THEN RETURN VersionError END;
    RETURN OK
  END ServerProcedureVCheck;


PROCEDURE EndOfArgs(m:Marshal.T):RC;
  VAR
    b:BOOLEAN;
  BEGIN
    b := Marshal.EndOfArgs(m);
    WITH m^ DO
      iPktBuf := PktBufs.New();
      cPktBuf := iPktBuf;
      marshal := TRUE;
      swapEndian := FALSE
    END;
    Marshal.Start(m);
    IF b THEN
      Marshal.ValueWord(OK, m);
      RETURN OK
    ELSE
      Marshal.ValueWord(ArgError, m);
      RETURN ArgError
    END;
  END EndOfArgs;


PROCEDURE SetReturnCode(
  rc:RC;
  m: Marshal.T
 );
  (* Must work regardless of whether EndOfArgs has happened. *)
  VAR
    rcDummy : RC;
  BEGIN
    IF NOT m^.marshal THEN rcDummy := EndOfArgs(m) END;
    m^.iPktBuf^.data[0] := WORD(rc)
  END SetReturnCode;


PROCEDURE EndOfCall(m:Marshal.T):PktBuf;
  VAR
    p:PktBuf;
  BEGIN
    p := m^.iPktBuf;
    Marshal.Discard(m);
    RETURN p
  END EndOfCall;


PROCEDURE Find(
      name: String;
      d:    Distance; 
  VAR place:Place    (* Out *)
 ):MemberRC;
  BEGIN
    IF RPCTable.Find(name, place.service) THEN
      place.location := RPCNets.thisLocation;
      RETURN BindOK
    END;
    IF d > ThisLocality THEN 
      RETURN NetMember.Find(name, d, place)
    ELSE
      RETURN NotFound 
    END;
  END Find;


PROCEDURE Offer(
      name:String;
      d:   Distance;
      p:   ReceiveProc;
      i:   Instance;
  VAR svId:ServiceId
 ):MemberRC;
  VAR
    place:Place;
  BEGIN
    IF RPCTable.Offer(name, d, p, i, svId) THEN
      IF d > ThisLocality THEN 
        place.location := RPCNets.thisLocation;
        place.service := svId;
        RETURN NetMember.Offer(name, d, place)
      ELSE
        RETURN BindOK
      END
    END;
    RETURN Exists
  END Offer;


PROCEDURE SuggestId(): ServiceId;
  BEGIN
    RETURN RPCTable.SuggestId()
  END SuggestId;


PROCEDURE OfferId(
  id:ServiceId; 
  p: ReceiveProc;
  i: Instance
 ):BOOLEAN;
  BEGIN
    RETURN RPCTable.OfferId(id, p, i) 
  END OfferId;


PROCEDURE ReOffer(
  id:ServiceId;
  p: ReceiveProc;
  i: Instance
 ):BOOLEAN;
  BEGIN
    RETURN RPCTable.ReOffer(id, p, i)
  END ReOffer;(*RAISES BadId*)


PROCEDURE Retract(service: ServiceId);
  BEGIN
    RPCTable.Retract(service)
  END Retract;


PROCEDURE SamePlace(place1, place2:Place):BOOLEAN;
  BEGIN
    RETURN (place1.service = place2.service)
       AND (SameLocation(place1.location, place2.location))
  END SamePlace;


PROCEDURE SameLocation(location1, location2:Location):BOOLEAN;
  VAR
    i:INTEGER;
  BEGIN
    FOR i := 0 TO 2 DO
      IF location1.word[i] # location2.word[i] THEN
        RETURN FALSE
      END
    END;
    RETURN TRUE
  END SameLocation;


PROCEDURE MarshalServiceId(VAR svId:ServiceId; (* InOut *) m:Marshal.T);
  BEGIN
    Marshal.Chars(ADR(svId), BYTESPERWORD, m)
  END MarshalServiceId;


PROCEDURE MarshalPlace(VAR place:Place; (* InOut *) m:Marshal.T);
  BEGIN
    Marshal.Chars(ADR(place), TSIZE(Place) * BYTESPERWORD DIV TSIZE(WORD), m)
  END MarshalPlace;


PROCEDURE MarshalLocation(VAR loc:Location; (* InOut *) m:Marshal.T);
  BEGIN
    Marshal.Chars(ADR(loc), TSIZE(Location) * BYTESPERWORD DIV TSIZE(WORD), m)
  END MarshalLocation;


END RPC.
