(*
    Title: 	Client code
    LastEdit:	"Thu Feb 25 18:07:08 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE RPCRemote;


IMPORT
  Arithmetic, Endian, MCNetLib, PktBufs, RPC, RPCBasics, RPCNaff, 
  RPCNets, RPCReceive, RPCStates, SYSTEM, Threads, ThreadsExtra, Timers;


FROM RPCBasics IMPORT
  CallEnd, CallId, RC, ConverseId, PacketType, ProtocolType, 
  ServiceId;


CONST
  FirstAliveTime   = 3; 
  LastAliveTime    = 32;
  AlivenessRetries = 8;
  CallRetries      = 16; (* Number of retries on EO or ALO *)

  CongestionDelay  = 2;


TYPE
  RecStats =
    RECORD
      total,
      normal,
      except,
      broken,
      active:CARDINAL
    END;


VAR
  bootNum,
  nextId:  CARDINAL;
  idLock:  Threads.Mutex;
  recStats:RecStats;


PROCEDURE Call(VAR p:PktBufs.T):RPCBasics.RC;
  VAR
    success:   BOOLEAN;
    state:     RPCStates.T;
    aloretries:CARDINAL;
    cr:        RC;
  BEGIN
    INC(recStats.total);
    IF (p^.next # NIL) THEN
      INC(recStats.broken);
      RETURN CallTooBig
    END;
    INC(recStats.active);
    state := RPCStates.New();
    WITH p^ DO     
      location[ClientEnd] := RPCNets.thisLocation;
      thread := CARDINAL(ThreadsExtra.currentThread);
      callId := NextCallId();
      protocolHdr.packet := CallPkt;
    END;
    WITH state^ DO
      RPCStates.CopyOutOfPkt(p, state);

      sArguments := p;
      sStatus := RPCStates.ClientWaitReply;
      success := FALSE;
      sResults := NIL;

      Threads.Acquire(sLock);
        RPCReceive.Add(ClientEnd, state);
        cr := RPCNets.Send(ServerEnd, state^.sArguments);
        IF cr = OK THEN 
          IF sProtocol = ExactlyOnce THEN
            cr := AwaitEOReply(state);
            IF cr = OK THEN RPCStates.SendTerminator(state) END;
          ELSE
            aloretries := CallRetries;
            LOOP
              cr := AwaitAOReply(state);
              IF (cr = OK)
              OR (sProtocol = AtMostOnce) 
              OR (aloretries = 0) THEN 
                EXIT
              END;
              DEC(aloretries);
              cr := RPCNets.Send(ServerEnd, state^.sArguments);
              IF cr # OK THEN EXIT END
            END 
          END
        END;
        p := sResults;
      Threads.Release(sLock);
      IF p # NIL THEN INC(recStats.normal) END;
      DEC(recStats.active);
      RPCReceive.Remove(state);
      RPCStates.Dispose(state)
    END;
    RETURN cr
  END Call;


PROCEDURE AwaitEOReply(state:RPCStates.T):RC;
  VAR
    retry,
    timeout,
    atimeout,
    aliveness:CARDINAL;
    cr:RC;

  PROCEDURE EOReplyUnknown(VAR cr:RC):BOOLEAN;
    VAR
      reason:RC;
    BEGIN
      cr := OK;
      WITH state^ DO
        CASE sUnknown^.protocolHdr.packet OF
          ReplyPkt:
            sResults := sUnknown;
            sUnknown := NIL;
            RETURN TRUE
        | DeadPkt:
            reason := RC(PktBufs.FetchDataWORD(sUnknown, 0));
            PktBufs.Discard(sUnknown);
            CASE reason OF
              Congested, NoSuchCall:
                DEC(retry);
                IF retry = 0 THEN
                  cr := reason; 
                  RETURN TRUE
                END;
                Timers.WaitSecs(CongestionDelay, Timers.FromNow);
                cr := RPCNets.Send(ServerEnd, state^.sArguments) 
            ELSE
                cr := reason;
                RETURN TRUE
            END
        | AliveAckPkt:
            state^.sAlive := TRUE;
            PktBufs.Discard(sUnknown)
        ELSE
            PktBufs.Discard(sUnknown)
        END
      END;
      RETURN FALSE
    END EOReplyUnknown;

  BEGIN
    timeout := FirstAliveTime;
    atimeout := FirstAliveTime;
    retry := CallRetries;
    aliveness := 0;
    WITH state^ DO
      IF (sUnknown # NIL) AND EOReplyUnknown(cr) THEN RETURN cr END;
      LOOP (* Waiting for a reply *)
        RPCNaff.WaitTimeOut(sLock, sCondition, timeout);
        IF sUnknown = NIL THEN
          IF aliveness = 0 THEN 
            aliveness := AlivenessRetries;
            sAlive := FALSE
          END
        ELSIF EOReplyUnknown(cr) THEN
          RETURN cr
        END;
        IF aliveness # 0 THEN
          IF sAlive THEN
            aliveness := 0;
            atimeout := Arithmetic.CMin(atimeout * 2, LastAliveTime);
            timeout := atimeout
          ELSE
            DEC(aliveness);
            IF aliveness = 0 THEN RETURN ServerCrash END;
            timeout := FirstAliveTime;
            RPCStates.SendAliveOne(ServerEnd, state)
          END
        END
      END (* loopy *)
    END
  END AwaitEOReply;


PROCEDURE AwaitAOReply(state:RPCStates.T):RC;
  VAR
    reason:RC;
  BEGIN
    WITH state^ DO
      RPCNaff.WaitTimeOut(sLock, sCondition, FirstAliveTime);
      IF sUnknown # NIL THEN
        CASE sUnknown^.protocolHdr.packet OF
          ReplyPkt:
            sResults := sUnknown;
            sUnknown := NIL;
            RETURN OK
        | DeadPkt:
            reason := RC(PktBufs.FetchDataWORD(sUnknown, 0));
            PktBufs.Discard(sUnknown);
            RETURN reason
        ELSE
        END
      END
    END;
    RETURN NoReply
  END AwaitAOReply;


PROCEDURE ReceivedPkt(
  p:    PktBufs.T;
  state:RPCStates.T
 );
  BEGIN
    Threads.Acquire(state^.sLock);
      CASE state^.sStatus OF
        RPCStates.ClientWaitReply:
          ActiveCall(p, state)
      ELSE
          PktBufs.Discard(p);
      END;
    Threads.Release(state^.sLock)
  END ReceivedPkt;


PROCEDURE ActiveCall(
  pkt:  PktBufs.T;
  state:RPCStates.T
 );
  BEGIN
    CASE pkt^.protocolHdr.packet OF
      ReplyPkt,
      DeadPkt:
        RPCStates.SignalPacket(state, pkt)
    | AliveAckPkt:
        state^.sAlive := TRUE;
        PktBufs.Discard(pkt)
    ELSE
        PktBufs.Discard(pkt)
    END
  END ActiveCall;


PROCEDURE NextCallId():CallId;
  VAR
    n:CARDINAL;
  BEGIN
    Threads.Acquire(idLock);
      n := bootNum + nextId;
      nextId := (nextId+1) MOD 01000000h;
    Threads.Release(idLock);
    RETURN n
  END NextCallId;


PROCEDURE Calls(
  VAR total: CARDINAL; (* Out *)
  VAR normal:CARDINAL; (* Out *)
  VAR except:CARDINAL; (* Out *)
  VAR broken:CARDINAL; (* Out *)
  VAR active:CARDINAL; (* Out *)
      reset: BOOLEAN
 );
  BEGIN
    total := recStats.total;
    normal := recStats.normal;
    except := recStats.except;
    broken := recStats.broken;
    active := recStats.active;
    IF reset THEN ResetCalls() END
  END Calls;


PROCEDURE ResetCalls();
  BEGIN
    recStats.total := 0;
    recStats.normal := 0;
    recStats.except := 0;
    recStats.broken := 0
  END ResetCalls;


PROCEDURE Init();
  VAR
    timeVal:Timers.Value;
  BEGIN
    ResetCalls();
    recStats.active := 0;
    Timers.Read(timeVal);
    bootNum := (timeVal.usecs MOD 256) * 01000000H;
    nextId := 1;
    Threads.InitMutex(idLock)
  END Init;


END RPCRemote.
