(*
    Title: 	Table of services in this address space
    LastEdit:	"Mon Feb  1 14:03:04 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE RPCTable;


IMPORT
  Exceptions, Iterator, MCLib, NetMember, PktBufs, RPCBasics, RPCNets,
  Strings, Threads, Timers;

FROM RPCBasics   IMPORT ServiceId, Distance, Instance, ThisLocality,
                        BinderInfo;
FROM Storage     IMPORT ALLOCATE, DEALLOCATE;
FROM SYSTEM      IMPORT ADR;
FROM SystemTypes IMPORT String;


CONST
  HashShift  = 8000000H;
  MaxElement = 32;
  BigPrime   = 134217689;
  PrimRoot   = 51815410; (*Good mathematical numbers!*)


TYPE 
  BinderTablePointer = POINTER TO BinderTableElement;
  BinderTableElement = 
    RECORD
      ptr:    BinderTablePointer; (* Must be 1st field *)
      service:ServiceId;
      name:   String;
      at:     PktBufs.ReceiveProc;
      i:      Instance;
      dist:   Distance
    END;
  BinderTableIndex   = [0..MaxElement-1];


VAR 
  mutex:      Threads.Mutex;
  currId:     CARDINAL;
  generation: Iterator.Generation;
  binderTable:ARRAY BinderTableIndex OF BinderTablePointer;


PROCEDURE Find(name:String; VAR svId:ServiceId):BOOLEAN;
  VAR
    j:CARDINAL;
    b:BinderTablePointer;
    s:ServiceId;
  BEGIN
    IF name = NIL THEN Exceptions.RAISE("BadName") END;
    Threads.Acquire(mutex);
      b := binderTable[MCLib.HashBytes(name, -1, MaxElement)];
      WHILE b # NIL DO
        IF (b^.name # NIL) AND Strings.EqualSS(name, b^.name) THEN
          s := b^.service;
          svId := s;
          Threads.Release(mutex);
          RETURN TRUE
        END;
        b := b^.ptr
      END;
    Threads.Release(mutex);
    RETURN FALSE
  END Find;


PROCEDURE FindId(
      id:    ServiceId;
  VAR at:    PktBufs.ReceiveProc;
  VAR i:     Instance
 ):BOOLEAN;
  VAR 
    b,c:BinderTablePointer;
  BEGIN
    Threads.Acquire(mutex);
      b := FindEntry(id, c);
      IF b # NIL THEN
        at := b^.at;
        i := b^.i
      END;
    Threads.Release(mutex);
    RETURN (b # NIL)
  END FindId;


PROCEDURE Offer(
      name:  String;
      d:     Distance;
      p:     PktBufs.ReceiveProc;
      i:     Instance;
  VAR svId:ServiceId
 ):BOOLEAN;
  VAR    
    ind: BinderTableIndex;
    id:  ServiceId;
    b,c: BinderTablePointer;
    proc:PROC;
  BEGIN
    IF name = NIL THEN RETURN FALSE END;
    Threads.Acquire(mutex);
      ind := MCLib.HashBytes(name, -1, MaxElement);
      b := binderTable[ind];
      c := ADR(binderTable[ind]);
      id := ServiceId(BITSET(NextId()) + BITSET(ind * HashShift));
      WHILE b # NIL DO
        IF b^.service = id THEN (* id already used, so get another one *)
          id := ServiceId(BITSET(NextId()) + BITSET(ind * HashShift));
          b := binderTable[ind]; (* and back to base of chain *)
          c := ADR(binderTable[ind])
        ELSIF (b^.name # NIL) AND Strings.EqualSS(name, b^.name) THEN
          Threads.Release(mutex);
          RETURN FALSE
        ELSE
          c := b;
          b := b^.ptr
        END
      END;
      c^.ptr := AddBinderTableEntry(id, p, i, d);
      b := c^.ptr;
      b^.name := Strings.CopySS(name);
    Threads.Release(mutex);
    svId := id;
    RETURN TRUE
  END Offer;


PROCEDURE SuggestId(): ServiceId;
  VAR
    id:ServiceId;
    ok:BOOLEAN;
    b: BinderTablePointer;
  BEGIN
    Threads.Acquire(mutex);
      REPEAT
        id := NextId();
        ok := TRUE;
        b := binderTable[0];
        WHILE ok AND (b # NIL) DO
          IF b^.service = id THEN
            ok := FALSE;
          ELSE
            b := b^.ptr;
          END
        END
      UNTIL ok;
    (* Note this hashes onto element 0 *)
    Threads.Release(mutex);
    RETURN id
  END SuggestId;


PROCEDURE OfferId(
  id:    ServiceId;
  p:     PktBufs.ReceiveProc;
  i:     Instance
 ):BOOLEAN;
  VAR
    b,c: BinderTablePointer;
    proc:PROC;
  BEGIN
    IF (id DIV HashShift) # 0 THEN RETURN FALSE END;
    Threads.Acquire(mutex);
      b := binderTable[0];
      c := ADR(binderTable[0]);
      WHILE b # NIL DO
        IF b^.service = id THEN
          Threads.Release(mutex);
          RETURN FALSE
        END;
        c := b;
        b := b^.ptr
      END;
      c^.ptr := AddNILBinderTableEntry(id, p, i);
    Threads.Release(mutex);
    RETURN TRUE
  END OfferId;


PROCEDURE ReOffer(
  id:ServiceId;
  p: PktBufs.ReceiveProc;
  i: Instance
 ):BOOLEAN;
  VAR 
    b,c: BinderTablePointer;
    proc:PROC;
  BEGIN
    IF (id DIV HashShift) # 0 THEN Exceptions.RAISE("BadId") END;
    Threads.Acquire(mutex);
      b := FindEntry(id, c);
      IF b # NIL THEN
        b^.at := p;
        b^.i := i
      END;
    Threads.Release(mutex);
    RETURN (b # NIL)
  END ReOffer;


PROCEDURE FindEntry(
      id:  ServiceId;
  VAR last:BinderTablePointer
 ):BinderTablePointer;
  VAR
    b:BinderTablePointer;
    h:CARDINAL;
  BEGIN
    h := HashId(id);
    b := binderTable[h];
    last := ADR(binderTable[h]);
    WHILE b # NIL DO
      IF (b^.service = id) THEN RETURN b END;
      last := b;
      b := b^.ptr
    END;
    RETURN NIL
  END FindEntry;


PROCEDURE AddNILBinderTableEntry(svId: ServiceId;
                                    p: PktBufs.ReceiveProc;
                                 inst: Instance): BinderTablePointer;
VAR b: BinderTablePointer;

  BEGIN
    b := AddBinderTableEntry(svId, p, inst, Distance(ORD(ThisLocality)-1));
    b^.name := NIL;
    RETURN b;
  END AddNILBinderTableEntry;


PROCEDURE AddBinderTableEntry(svId: ServiceId;
                                 p: PktBufs.ReceiveProc;
                              inst: Instance;
                                 d: Distance): BinderTablePointer;
  VAR
    b:BinderTablePointer;
  BEGIN
    INC(generation);(*Iterators going out of date*)
    NEW(b);
    WITH b^ DO
      ptr := NIL;
      service := svId;
      at := p;
      i := inst;
      dist := d
    END;
    RETURN b
  END AddBinderTableEntry;


PROCEDURE Info(
  VAR s:    Iterator.State;     (* InOut *)
  VAR count:CARDINAL;           (* Out   *)
  VAR ids:  ARRAY OF ServiceId; (* Out   *)
      bra:  Distance;
      ket:  Distance
 );
  VAR 
    i,j,k:CARDINAL;
    b:    BinderTablePointer;
  BEGIN
    count := 0;
    i := Iterator.Check(s, generation, 1);
    IF Iterator.Exhausted(s) OR Iterator.Finished(s) OR Iterator.Broken(s) THEN
      RETURN (* Give up if broken, exhausted or finished *)
    END;
    j := i DIV 10000H;
    i := i MOD 10000H;
    IF (j > (MaxElement-1)) OR (i = 0) THEN
      Iterator.Finish(s);
      Iterator.Break(s);
      RETURN
    END; (* if *)
    DEC(i); (* Counting from 1 *)
    Threads.Acquire(mutex);
      b := binderTable[j];
      k := 0;
      WHILE (i # 0) AND (b # NIL) DO
        b := b^.ptr;
        DEC(i);
        INC(k)
      END;
      LOOP
        WHILE (b # NIL) DO
          IF (b^.name # NIL)
          AND (b^.dist >= bra) AND (b^.dist <= ket) THEN
            ids[count] := b^.service;
            INC(count);
            IF count > HIGH(ids) THEN
              s.pos := j * 10000H + k + 2; (*Next element*)
              EXIT
            END
          END;
          b := b^.ptr;
          INC(k)
        END;
        INC(j);
        IF j > (MaxElement-1) THEN
          s.pos := 0;
          IF count = 0 THEN
            Iterator.Exhaust(s);
          END; (* if *)
          EXIT 
        END;
        k := 0;
        b := binderTable[j]
      END;
    Threads.Release(mutex)
  END Info;


PROCEDURE InfoById(
      id:  ServiceId;
  VAR b:   BinderInfo;   (* Out *)
  VAR name:ARRAY OF CHAR (* Out *)
 ):BOOLEAN;
  VAR 
    bptr,
    cptr:BinderTablePointer;
  BEGIN
    Threads.Acquire(mutex);
      bptr := FindEntry(id, cptr);
      IF bptr # NIL THEN
        b.id := bptr^.service;
        b.i := bptr^.i;
        b.d := bptr^.dist;
        IF  (bptr^.name # NIL)
        AND (HIGH(name) < Strings.LengthS(bptr^.name)) THEN
          Exceptions.RAISE("Bad Parameter")
        END;
        IF (bptr^.name # NIL) THEN
          Strings.CopySC(bptr^.name, name)
        ELSE
          name[0] := 0c
        END
      END;
    Threads.Release(mutex);
    RETURN (bptr # NIL)
  END InfoById;


PROCEDURE Retract(service: ServiceId);
  VAR
    i:   BinderTableIndex;
    b,c: BinderTablePointer;
    proc:PROC;
    id:  ServiceId;
  BEGIN
    Threads.Acquire(mutex);
      b := FindEntry(service, c);
      IF b # NIL THEN
        i := HashId(service);
        c^.ptr := b^.ptr;
        b^.ptr := NIL;
        INC(generation); (* Just in case someone is iterating *)
      END;
    Threads.Release(mutex);
    ThrowAway(b)
  END Retract;


PROCEDURE HashId(id: ServiceId): BinderTableIndex;
  (* Look up a service by id in the binder table *)
  BEGIN
    RETURN BinderTableIndex(id DIV HashShift)
  END HashId;


PROCEDURE NextId(): ServiceId;
(*Produces a new service Id from the last one*)
  BEGIN
    REPEAT
      currId := (((currId*32) MOD BigPrime) * 31) MOD BigPrime;
      (* Note 32*31 = 992 is a primitive root *)
    UNTIL currId > 100;
    RETURN (currId MOD HashShift)
  END NextId;


PROCEDURE ThrowAway(dead:BinderTablePointer);
  VAR
    b:    BinderTablePointer;
    place:RPCBasics.Place;
  BEGIN
    place.location := RPCNets.thisLocation;
    WHILE dead # NIL DO
      IF dead^.dist > ThisLocality THEN 
        place.service := dead^.service;
        NetMember.Retract(dead^.name, place, dead^.dist) 
      END;
      IF dead^.name # NIL THEN Strings.Dispose(dead^.name) END;
      b := dead;
      dead := dead^.ptr;
      DISPOSE(b)
    END
  END ThrowAway;


PROCEDURE Init();
  VAR
    i:CARDINAL;
    t:Timers.Value;
  BEGIN
    Threads.InitMutex(mutex);
    FOR i := 0 TO MaxElement-1 DO
      binderTable[i] := NIL;
    END;
    Timers.Read(t);
    currId := t.usecs;
    generation := 0
  END Init;


END RPCTable.
