IMPLEMENTATION MODULE Timers;


IMPORT
  Exceptions, SYSTEM, Threads, ThreadsExtra, UnixCall, UnixSignals;

IMPORT
  SysStreams, WriteF;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;


TYPE
  Entry = POINTER TO EntryR;
  EntryR =
    RECORD
      next:    Entry;
      flag:    Threads.Condition;
      absTime: CARDINAL;
      callBack:CallBackProc;
      callArg: CARDINAL
    END;


VAR
  timeLock:    Threads.Mutex;
  timeQueue,
  spareEntries:Entry;
  checkTime:   BOOLEAN;
  timeEntries: INTEGER;
  

  
PROCEDURE Read(VAR timeVal:Value);
  VAR
    timeZone:ARRAY [0..1] OF CARDINAL;
  BEGIN
    SYSTEM.CALLEXTPROC(UnixCall.gettimeofday, 
                       SYSTEM.ADR(timeVal), SYSTEM.ADR(timeZone))
  END Read;
    

PROCEDURE Wait(period:Value; type:Type);
  BEGIN
    IF period.usecs > 5000000 THEN INC(period.secs) END;
    CallBackSecs(period.secs, type, CallBackProc(0), 0)
  END Wait;


PROCEDURE WaitSecs(secs:CARDINAL; type:Type);
  BEGIN
    CallBackSecs(secs, type, CallBackProc(0), 0)
  END WaitSecs;


PROCEDURE CallBack(
      period:Value; 
      type:  Type; 
      proc:  CallBackProc; 
      arg:   CARDINAL
 );
  BEGIN
    IF period.usecs > 5000000 THEN INC(period.secs) END;
    CallBackSecs(period.secs, type, proc, arg)
  END CallBack;


PROCEDURE CallBackSecs(
      secs:CARDINAL; 
      type:Type; 
      proc:CallBackProc;
      arg: CARDINAL
 );
  VAR
    entry:  Entry;
    curTime:Value;
  BEGIN
    IF secs = 0 THEN RETURN END;
    Read(curTime); 
    IF type = FromNow THEN INC(secs, curTime.secs) END;
    IF secs > 5 + curTime.secs THEN
      WriteF.WriteF1(SysStreams.sysOut, 
                     "Time wait %i\n", secs - curTime.secs);
      BreakPt()
    END;
    Threads.Acquire(timeLock);
      entry := NewEntry();
      WITH entry^ DO
        absTime := secs;
        callBack := proc;
        callArg := arg;
        Insert(entry, curTime.secs);
        IF CARDINAL(proc) = 0 THEN
          Threads.Wait(timeLock, flag);
          DisposeEntry(entry)
        END;
      END;
    Threads.Release(timeLock)
  END CallBackSecs;


PROCEDURE BreakPt();
  BEGIN
  END BreakPt;


PROCEDURE Cancel(
      proc:CallBackProc;
      arg: CARDINAL
 );
  VAR
    entry,
    pentry:Entry;
  BEGIN 
    Threads.Acquire(timeLock);
      pentry := SYSTEM.ADR(timeQueue);
      LOOP
        entry := pentry^.next;
        IF entry = NIL THEN BreakPt(); EXIT END;
        WITH entry^ DO
          IF  (arg = callArg)
          AND (CARDINAL(proc) = CARDINAL(callBack)) THEN
            pentry^.next := next;
            DisposeEntry(entry);
            EXIT
          END
        END;
        pentry := entry
      END;
    Threads.Release(timeLock)
  END Cancel;
 

PROCEDURE Check();
  VAR
    entry:  Entry;
    curTime:Value;
  BEGIN
    IF (timeQueue = NIL) OR (NOT checkTime) THEN RETURN END;
    Read(curTime);
    WHILE (timeQueue # NIL) AND (timeQueue^.absTime <= curTime.secs) DO
   (* WriteF.WriteF2(SysStreams.sysErr, 
                     "timeQueue = 0x%x, entries = %i\n", 
                     timeQueue, timeEntries); *)
      WITH timeQueue^ DO
        IF CARDINAL(callBack) = 0 THEN
          Threads.Signal(flag);
          timeQueue := next
        ELSE
          callBack(callArg);
          entry := timeQueue;
          timeQueue := next;
          DisposeEntry(entry)        
        END;
      END
    END;
    checkTime := FALSE;
    IF timeQueue # NIL THEN
      SYSTEM.CALLEXTPROC(UnixCall.alarm,
                         timeQueue^.absTime - curTime.secs);
    END;
  END Check;


PROCEDURE NextInterval(VAR timeVal:Value);
  VAR
    curTime:Value;
  BEGIN
    Read(curTime);
    IF timeQueue = NIL THEN
      timeVal.secs := 300
    ELSIF curTime.secs >= timeQueue^.absTime THEN
      timeVal.secs := 0;
      checkTime := TRUE
    ELSE
      timeVal.secs := timeQueue^.absTime - curTime.secs;
    END;  
    timeVal.usecs := 0
  END NextInterval;


PROCEDURE Insert(entry:Entry; secs:CARDINAL);
  VAR
    time:CARDINAL;
    prev,queue:Entry;
  BEGIN
    queue := timeQueue;
    prev := SYSTEM.ADR(timeQueue);
    WHILE (queue # NIL) AND (queue^.absTime < entry^.absTime)  DO
      prev := queue;
      queue := queue^.next
    END;
    prev^.next := entry;
    entry^.next := queue;
    time := timeQueue^.absTime - secs;
    IF time = 0 THEN time := 1 END;
    SYSTEM.CALLEXTPROC(UnixCall.alarm, time) 
  END Insert;


PROCEDURE NewEntry():Entry;
  VAR
    entry:Entry;
  BEGIN
    IF spareEntries = NIL THEN
      NEW(entry);
      Threads.InitCondition(entry^.flag)
    ELSE
      entry := spareEntries;
      spareEntries := spareEntries^.next
    END;
    entry^.next := NIL;
    INC(timeEntries);
    RETURN entry
  END NewEntry;


PROCEDURE DisposeEntry(entry:Entry);
  BEGIN
    IF entry # NIL THEN 
      DEC(timeEntries);
      entry^.next := spareEntries;
      spareEntries := entry
    ELSE
      BreakPt()
    END
  END DisposeEntry;


PROCEDURE Handler(sig, code:SYSTEM.WORD; scp:SYSTEM.ADDRESS);
  BEGIN
    IF timeQueue # NIL THEN checkTime := TRUE END;
    ThreadsExtra.preEmpt := TRUE 
  END Handler;


PROCEDURE Init();
  VAR
    oldHand:SYSTEM.WORD;
  BEGIN
    checkTime := FALSE;
    Threads.InitMutex(timeLock);
    timeQueue := NIL;
    spareEntries := NIL;
    timeEntries := 0;
    oldHand := UnixSignals.Signal(UnixSignals.SIGALARM, Handler)
  END Init;


BEGIN
  Init()
END Timers.
