(*
    Title: 	UNIX Synchronous IO
    LastEdit:	"Thu Jan 28 10:10:37 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE UnixSelect;


IMPORT 
  Exceptions, Threads, SYSTEM, Timers, UnixCall, UnixMCI;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;

IMPORT
  SysStreams, WriteF;

CONST
  FDWords = 1;
  PossibleFD = (FDWords * SYSTEM.BITSPERWORD);


TYPE
  FDStatus = POINTER TO FDStatusR;
  FDStatusR =
    RECORD
      flag:   Threads.Condition;
      waiters:INTEGER;
      ready:  BOOLEAN
    END;
  FileDesc = ARRAY [0..FDWords-1] OF BITSET;
  FileStat = ARRAY [0..PossibleFD-1] OF FDStatus;
  IOType = (ReadIO, WritIO);
  RWFileDesc = ARRAY IOType OF FileDesc;
  RWFileStat = ARRAY IOType OF FileStat;
  TimeValP = POINTER TO TimeVal;


VAR
  highFD:  INTEGER;
  nullFs:  FileDesc;
  fileDesc:RWFileDesc;
  fileStat:RWFileStat;
  descLock:Threads.Mutex;
  nullTime:TimeVal;
  infTime: TimeValP;


PROCEDURE WaitForRead(fd:INTEGER);
  VAR
    n:INTEGER;
    readFs:FileDesc;
  BEGIN 
    IF (fd < 0) OR (fd > highFD) THEN CheckFD(fd) END;
    readFs[0] := {};
    INCL(readFs[0], fd);
    n := Select(PossibleFD, readFs, nullFs, nullFs, nullTime);
    IF n <= 0 THEN WaitForIO(fd, ReadIO) END
  END WaitForRead;


PROCEDURE WaitForWrite(fd:INTEGER);
  VAR
    n:INTEGER;
    writeFs:FileDesc;
  BEGIN 
    IF (fd < 0) OR (fd > highFD) THEN CheckFD(fd) END;
    writeFs[0] := {};
    INCL(writeFs[0], fd);
    n := Select(PossibleFD, nullFs, writeFs, nullFs, nullTime);
    IF n <= 0 THEN WaitForIO(fd, WritIO) END
  END WaitForWrite;

  
PROCEDURE LastThread();
  VAR
    n: INTEGER;
    fs:RWFileDesc;
  BEGIN 
    REPEAT
      CopyFs(fs); 
      n := Select(PossibleFD, fs[ReadIO], fs[WritIO], nullFs, infTime^);
    UNTIL n # 0;
    IF n > 0 THEN KickMarkedFD(n, fs) END;
  END LastThread;


PROCEDURE Poll();
  VAR
    n: INTEGER;
    fs:RWFileDesc;
  BEGIN 
    Timers.Check();
    CopyFs(fs); 
    n := Select(PossibleFD, fs[ReadIO], fs[WritIO], nullFs, nullTime);
    IF n > 0 THEN KickMarkedFD(n, fs) END;
  END Poll;


PROCEDURE WaitForIO(fd:INTEGER; ioType:IOType);
  VAR
    n: INTEGER;
    fs:RWFileDesc;
  BEGIN 
    Threads.Acquire(descLock);
      WITH fileStat[ioType, fd]^ DO
        INC(waiters); INCL(fileDesc[ioType, 0], fd);
        Poll();
        WHILE NOT ready DO Threads.Wait(descLock, flag) END;
        ready := FALSE;
        DEC(waiters); IF waiters = 0 THEN EXCL(fileDesc[ioType, 0], fd) END
      END;
    Threads.Release(descLock)
  END WaitForIO;


PROCEDURE CopyFs(VAR fs:RWFileDesc);
  BEGIN
    fs[ReadIO, 0] := fileDesc[ReadIO, 0];
    fs[WritIO, 0] := fileDesc[WritIO, 0]
  END CopyFs;
  

PROCEDURE KickMarkedFD(n:INTEGER; VAR fs:RWFileDesc);
  VAR
    cfd:INTEGER;
    ioType:IOType;
  BEGIN
    cfd := 0;
    WHILE n > 0 DO
      FOR ioType := ReadIO TO WritIO DO
        IF CARDINAL(cfd) IN fs[ioType, 0] THEN 
          WITH fileStat[ioType, cfd]^ DO
            ready := TRUE;
            Threads.Signal(flag);
          END;
          DEC(n)
        END;
      END;
      INC(cfd)
    END
  END KickMarkedFD;

  
PROCEDURE Select(
      nfds:INTEGER; 
  VAR readFs, writeFs, exceptFs:ARRAY OF BITSET;
  VAR timeout:TimeVal
 ):INTEGER;
  VAR
    n:INTEGER;
  BEGIN
    n :=
      INTEGER(
        SYSTEM.CALLEXTFUNC(UnixCall.select, 
                           nfds,
                           SYSTEM.ADR(readFs), 
                           SYSTEM.ADR(writeFs), 
                           SYSTEM.ADR(exceptFs), 
                           SYSTEM.ADR(timeout)));
 (* IF n < 0 THEN 
      WriteF.WriteF1(SysStreams.sysErr, 
                     "Select fail: %d\n", UnixMCI.errnoPtr^)
    END; *)
    RETURN n
  END Select;


PROCEDURE CheckFD(fd:INTEGER);
  VAR
    i:INTEGER;
    ioType:IOType;
  BEGIN
    IF (fd < 0) OR (fd > SYSTEM.BITSPERWORD) THEN 
      Exceptions.RAISE("fd too large") 
    END;
    FOR i := highFD+1 TO fd DO
      FOR ioType := ReadIO TO WritIO DO
        NEW(fileStat[ioType, i]);
        WITH fileStat[ioType,i]^ DO
          Threads.InitCondition(flag);
          waiters := 0;
          ready := FALSE
        END
      END
    END;
    highFD := fd;
  END CheckFD;


PROCEDURE Init();
  VAR
    i:INTEGER;
    ioType:IOType;
  BEGIN
    highFD := -1;
    nullFs[0] := {};
    FOR ioType := ReadIO TO WritIO DO
      fileDesc[ioType, 0] := {}
    END;
    Threads.InitMutex(descLock);
    nullTime.secs := 0;
    nullTime.usecs := 0;
    infTime := TimeValP(0)
  END Init;


BEGIN
  IF ValidFD > 32 THEN
    Exceptions.RAISE("Select implemented with 32 FD")
  END;
  Init()
END UnixSelect.
