IMPLEMENTATION MODULE GenOp;

FROM items IMPORT ITEM, MODE, REGISTER, Operator, MonOperator,
                  equal, immediate, InDataReg, NullRegItem, PackItem;
FROM PsStk IMPORT ResultItem;
FROM AuxGen1 IMPORT AddAndOr, Subtract, ExclusiveOr;
FROM AuxGen2 IMPORT shift;
FROM AuxGen3 IMPORT Multiply, DivMod;
FROM MonGen IMPORT AbsoluteValue, SimpleMonadicOps, ShiftMonadicOps;
FROM Relations IMPORT CompareAndStoreResult;
FROM GenProcs IMPORT GenMoveImm;
FROM UtilError IMPORT ReportError;


(* The procedure 'generate' is called by the procedure 'pop'
   in module PsStk. Its purpose is to generate the code to
   do some operation (specified by the 'Operator' argument)
   on the arguments 'Opd1' and 'Opd2' and to put the result
   in the place given by 'Dest'.
   There are many cases when the places where the values of
   Opd1 and Opd2 are held and the place where the result has
   to go influence the code generated very substantially.
   e.g. suppose Opd1 is held in the same place as the result
   has to go to and Opd2 is a register. the code required if
   the operation is addition is just:
         ADD.L Opd2,Opd1
   In contrast if Opd1 and Opd2 are in two different places
   in store and Dest is another place in store then an
   intermediate register, say R, must be used :
         MOVE.L Opd1, R
	 ADD.L  Opd2, R
	 MOVE.L R, Dest
   Note that in this case R now holds the value that is also
   in the location in store given by Dest and this fact should
   be noted in the register slave.
   A special case is when Dest is given the value 'NullRegItem'.
   This means that the result has to go into a register but
   it is not important which. This is useful to know if one of
   the operands - say Opd1 - is already in a register because
   the result can then be put in Opd1 and the fact that the
   result is in that register noted :
         ADD.L Opd2,Opd1
	
   The operands can be dealt with in special ways if:
    i)    one of them is immediate data.
    ii)   one or both of them is held in a register.
    iii)  the value of one or both of them is slaved -
          i.e. the value of the operand is held in a
	  register due to some previous calculation
	  though the operand itself specifies a location
	  in store.
	
   Often an intermediate register is required (as in one of the
   examples above) and obviously when the procedure FreeReg is
   called to free a register any register holding information
   about Opd1 or Opd2 (see ii and iii above) must not be 'freed'
   as it holds vital information. Because of this before entry
   to 'generate' - in the procedure 'pop' - the use bit of any
   register holding the value of an operand is set so it cannot
   be affected by free reg. On returning to 'pop' these bits are
   unset again.
   Note also that if the value of say Opd1 is slaved in a register
   R, and the value of this register is changed during the
   operation e.g.
         ADD.L Opd2,R
	 MOVE.L R,Dest
   then the register slave must be informed that R is no longer
   slaving the value of Opd1 (and is, in this case, now slaving
   the value of Dest).
*)

PROCEDURE generate (Op : Operator; VAR Opd1, Opd2, Dest : ITEM);
    BEGIN
        IF immediate (Opd1) AND immediate (Opd2) THEN
            DoOp (Op, Opd1.DispOrData, Opd2.DispOrData, Dest);
            RETURN;
        END (* if *);
        CASE Op OF
          Uadd, Add,
          And, Or    : AddAndOr (Op, Opd1, Opd2, Dest) |
          Usub, Sub  : Subtract (Op, Opd1, Opd2, Dest) |
          Umul, Mul  : Multiply (Op, Opd1, Opd2, Dest) |
          Udiv, Div,
          Umod, Mod  : DivMod (Op, Opd1, Opd2, Dest)   |
          Xor        : ExclusiveOr (Opd1, Opd2, Dest)  |
  	  Shl, Shr,
  	  Ashl, Ashr,
          Ror        : shift   (Op, Opd1, Opd2, Dest, FALSE)  |
          In, Eql, Neq, Ulss, Lss, Uleq, Leq, Ugtr, Gtr,
          Ugeq, Geq  : CompareAndStoreResult
                              (Op, Opd1, Opd2, Dest)
        END (* case *);
    END generate;

PROCEDURE GenerateMonadicOp (Op : MonOperator; VAR Opd1, Dest : ITEM);
    BEGIN
        IF immediate (Opd1) THEN
            DoMonadicOp (Op, Opd1.DispOrData, Dest);
            RETURN;
        END (* if *);
        CASE Op OF

          Abs       :   AbsoluteValue (Opd1, Dest)          |
          Neg, Com,
          Not       :   SimpleMonadicOps (Op, Opd1, Dest)   |
          Msk, Bit  :   ShiftMonadicOps (Op, Opd1, Dest)

        END (* case *);
    END GenerateMonadicOp;

PROCEDURE DoOp (Op : Operator; val1, val2 : INTEGER; VAR Dest : ITEM);
    VAR ResultVal : INTEGER;
    BEGIN
        CASE Op OF

          Uadd  :   ResultVal := CARDINAL (val1) + CARDINAL (val2)       |
          Add   :   ResultVal := val1 + val2                             |
          And   :   ResultVal := INTEGER (BITSET (val1) * BITSET (val2)) |
          Or    :   ResultVal := INTEGER (BITSET (val1) + BITSET (val2)) |
          Usub  :   ResultVal := CARDINAL (val2) - CARDINAL (val1)       |
          Sub   :   ResultVal := val2 - val1                             |
          Umul  :   ResultVal := CARDINAL (val1) * CARDINAL (val2)       |
          Mul   :   ResultVal := val1 * val2                             |
          Udiv  :   ResultVal := CARDINAL (val2) DIV CARDINAL (val1)     |
          Div   :   ResultVal := val2 DIV val1                           |
          Umod  :   ResultVal := CARDINAL (val2) MOD CARDINAL (val1)     |
          Mod   :   ResultVal := val2 MOD val1                           |
          Xor   :   ResultVal := INTEGER (BITSET (val2) / BITSET (val1)) |
          Shl   :   ResultVal := DoShiftLeft (val2, val1)                |
          Shr   :   ResultVal := DoShiftRight (val2, val1)
        ELSE
            ReportError (TRUE,
                         "Trying to rotate or compare two immediate values*N",
                         0, 0);
        END (* case *);
        IF equal (Dest, NullRegItem) THEN
            PackItem (ResultItem, Immediate, NoReg, ResultVal, NoReg);
        ELSE
            GenMoveImm (ResultVal, Dest);
        END (* if *);
    END DoOp;

PROCEDURE DoShiftLeft (shifted, shifter : CARDINAL) : CARDINAL;
    VAR res, count : CARDINAL;
    BEGIN
        res := shifted;
        FOR count := 1 TO shifter DO
            res := res * 2;
        END (* for *);
        RETURN res;
    END DoShiftLeft;

PROCEDURE DoShiftRight (shifted, shifter : CARDINAL) : CARDINAL;
    VAR res, count : CARDINAL;
    BEGIN
        res := shifted;
        FOR count := 1 TO shifter DO
            res := res DIV 2;
        END (* for *);
        RETURN res;
    END DoShiftRight;

PROCEDURE DoMonadicOp (Op : MonOperator; val : INTEGER; VAR Dest : ITEM);
    VAR ResultVal : INTEGER;
        NullSet : BITSET;
    BEGIN
        NullSet := BITSET {};
        CASE Op OF

            Abs     :   ResultVal := ABS (val);
        |   Neg     :   ResultVal := - val;
        |   Com     :   ResultVal := (- val) - 1;
        |   Not     :   IF val = 0 THEN
                            ResultVal := 1;
                        ELSE
                            ResultVal := 0;
                        END (* if *);
        |   Bit     :   INCL (NullSet, CARDINAL (val));
                        ResultVal := INTEGER (NullSet);
        |   Msk     :   INCL (NullSet, CARDINAL (val));
                        ResultVal := (-INTEGER (NullSet)) - 1;

        END (* case *);
        IF equal (Dest, NullRegItem) THEN
            PackItem (ResultItem, Immediate, NoReg, ResultVal, NoReg);
        ELSE
            GenMoveImm (ResultVal, Dest);
        END (* if *);
    END DoMonadicOp;

PROCEDURE SetResultItem (VAR result : ITEM);
    BEGIN
        ResultItem := result
    END SetResultItem;

END GenOp.
