IMPLEMENTATION MODULE RegSlv;

FROM items IMPORT ITEM, MODE, REGISTER, PackItem,
                  InDataReg, slaveable, equal, 
		  NullRegItem, UsesSlavedAddrReg;
FROM PsStk IMPORT PutInScratchSpace;
FROM AddrReg IMPORT DecUseCount, PurgeAddrRegSlave,
                    PurgeVariablesInAddrRegSlave, ClearFromAddrRegSlave;
FROM UtilError IMPORT ReportError;


TYPE RegisterState = (Free, Slaving, OnStack);
     (* These states indicate that a register is being
	used as follows :

	  Free : The register is not holding any useful
		 information.
	  
	  Slaving :
	      The register holds a value which is also
	      the value of a word in store. In the 
	      register slave the address of this word
	      is stored along with the fact that the 
	      register is 'Slaving' (in the PSItem field
	      of a RegSlaveItem - see below). If the 
	      value of this word is ever needed in the
	      program it can be seen from the register
	      slave that the value is available in a
	      register and need not be fetched from store.

	  OnStack :
	      The register holds a value which is on
	      the expression stack i.e. the pseudo
	      stack has an entry which says that this
	      register is holding a value which will
	      be required when working out the 
	      expression currently being processed.
	      The value in the register is therefore 
	      valuable as it is not stored anywhere 
	      else; so before such a register can be 
	      freed for some other purpose this value 
	      must be saved in store.
     *)
     RegSlaveItem  = RECORD
                        State  : RegisterState;
                        Item   : ITEM;
                        Age    : CARDINAL;
			InUse  : BOOLEAN
                     END;
     (* The use bit indicates whether the register is 
        actually in use at a given time. It is set so
	that if you need to get hold of another register
	but want to preserve the contents of some other
	ones you can set the use bits on these so that
	when FreeReg is called they will not be affected *)

CONST Youngest = 0;


VAR RegSlave         : ARRAY [D0..D7] OF RegSlaveItem;
    AnyUseBitsSet    : BOOLEAN;
    reg              : REGISTER;
    RegSlaveIsPurged : BOOLEAN;

PROCEDURE RegCanBeFreedSimply () : BOOLEAN;
    BEGIN
        FOR reg := D0 TO D7 DO
            WITH RegSlave [reg] DO
                IF NOT (InUse OR (State = OnStack)) THEN
                    RETURN TRUE;
                END (* if *);
            END (* with *);
        END (* for *);
        RETURN FALSE;
    END RegCanBeFreedSimply;

(* This procedure takes a PseudStkItem which should contain
   the address of some object in store. The register slave is
   searched to see if this address is slaved by some register.
   If it is the appropriate register is freed. If the item 
   given as the argument is not an address or it is not being
   slaved then no action is taken.
*)
PROCEDURE StopSlaving (VAR item : ITEM);
   VAR Reg : REGISTER;
   BEGIN
      IF NOT slaveable (item) THEN RETURN 
      END;
      ClearFromAddrRegSlave (item);
      Reg := D0;
      LOOP
         WITH RegSlave [Reg] DO
	    IF (State = Slaving)
	    AND equal (item, Item)
	    THEN State := Free; EXIT
	    END
         END;
	 INC (Reg); 
	 IF Reg > D7 THEN RETURN END
      END;
      DecrementAges (Reg)
   END StopSlaving;


(* This procedure takes a register as its argument. If that
   register is marked in the register slave as slaving some
   value it is freed.
*)
PROCEDURE DeleteFromSlave (Reg : REGISTER);
   BEGIN
      WITH RegSlave [Reg] DO
         IF State = Slaving THEN State := Free
	 END
      END;
      DecrementAges (Reg)
   END DeleteFromSlave;

(* This procedure takes two arguments : 'reg' and
   'item' (which may represent some address in store). 
   If this address is already slaved then the register 
   slaving it is freed.
   If the address can be slaved and 'reg' does in fact
   represent a register item then the address is marked 
   as being slaved by 'reg'.
   If 'reg' has changed ('regchanged' indicates this) but
   'item' is not slaveable then it is marked as 'Free'.
*)
PROCEDURE PutIntoSlave (VAR reg, item : ITEM; regchanged: BOOLEAN);
   VAR Reg : REGISTER;
BEGIN
    StopSlaving (item);
    IF InDataReg (reg) THEN
        IF slaveable (item) THEN 
            RegSlaveIsPurged := FALSE;
            WITH RegSlave [reg.Register] DO
                State := Slaving; Item := item; 
                Age := 0; InUse := FALSE;
            END;
            FOR Reg := D0 TO D7 DO
                WITH RegSlave [Reg] DO
                    IF State = Slaving THEN INC (Age) END
                END
            END
        ELSE
            IF regchanged THEN RegSlave[reg.Register].State := Free END;
        END
    END
END PutIntoSlave;


PROCEDURE DecrementAges (reg : REGISTER);
   VAR Reg : REGISTER;
   BEGIN
      FOR Reg := D0 TO D7 DO
	 WITH RegSlave [Reg] DO
	    IF (State = Slaving) 
	    AND (Age > RegSlave [reg].Age)
	    THEN DEC (Age) END
         END
      END
   END DecrementAges;

PROCEDURE FreeReg (VAR item : ITEM; UseBit : BOOLEAN);

   PROCEDURE Setitem (reg : REGISTER);
      BEGIN
         WITH RegSlave [reg] DO
	    State := Free; InUse := UseBit;
	 END;
	 PackItem (item, DataReg, reg, 0, NoReg)
      END Setitem;

   VAR Reg, OldestReg : REGISTER;
       MaxAge : CARDINAL;

   BEGIN
      IF NOT AnyUseBitsSet
      THEN AnyUseBitsSet := UseBit
      END;
      FOR Reg := D0 TO D7 DO
         WITH RegSlave [Reg] DO
            IF NOT InUse
	    AND (State = Free)
	    THEN Setitem (Reg); RETURN
            END
	 END
      END;
      OldestReg := NoReg;
      FOR Reg := D0 TO D7 DO
            WITH RegSlave [Reg] DO
            IF NOT InUse AND (State = Slaving)
            AND (Age > MaxAge)
            THEN MaxAge := Age;
                 OldestReg := Reg
            END
         END
      END;
      IF OldestReg <> NoReg
      THEN DecrementAges (OldestReg);
           Setitem (OldestReg); RETURN
      END;
      Reg := D0;
      WHILE (Reg <= D7) AND RegSlave [Reg].InUse DO
         INC (Reg)
      END;
      IF Reg <= D7
      THEN PutInScratchSpace (Reg);
           Setitem (Reg); RETURN
      ELSE ReportError (TRUE,
                        "no free registers - panic*N",
                        0, 0)
      END
   END FreeReg;
   
PROCEDURE IsFree (reg : REGISTER) : BOOLEAN;
    BEGIN
        WITH RegSlave [reg] DO
            RETURN (State = Free) AND (NOT InUse);
        END (* with *);
    END IsFree;

(* If the item given as an argument is
   slaved the item is set to the register
   which slaves it.
*)
PROCEDURE slave (VAR item : ITEM);
   VAR Reg : REGISTER;
   BEGIN
      FOR Reg := D0 TO D7 DO
         WITH RegSlave [Reg] DO
            IF (State = Slaving) AND equal (Item, item)
            THEN PackItem (item, DataReg, Reg, 0, NoReg);
                 RETURN
            END
         END
      END;
   END slave;


PROCEDURE Slave (VAR item, SlaveItem : ITEM);
   BEGIN
      SlaveItem := item;
      slave (SlaveItem)
   END Slave;


(* This procedure adjusts the register slave
   after the value of an address register has
   been changed. Registers slaving addresses
   using this address register are freed
*)
PROCEDURE AdjustRegSlave (AddrReg : REGISTER);
   VAR reg : REGISTER;
   BEGIN
      FOR reg := D0 TO D7 DO
         WITH RegSlave [reg] DO
	    IF (State = Slaving) 
	    AND (Item.Register = AddrReg)
	    THEN State := Free;
	    END
         END
      END
   END AdjustRegSlave;


PROCEDURE SetRegFree (reg : REGISTER);
   BEGIN
      RegSlave [reg].State := Free
   END SetRegFree;


PROCEDURE protect (VAR item : ITEM);
   BEGIN
      IF InDataReg (item) AND NOT equal (item, NullRegItem)
      THEN RegSlave [item.Register].InUse := TRUE;
           AnyUseBitsSet := TRUE
      END
   END protect;


PROCEDURE ResetUseBits ();
   VAR Reg : REGISTER;
   BEGIN 
      IF AnyUseBitsSet
      THEN AnyUseBitsSet := FALSE
      ELSE RETURN
      END;
      FOR Reg := D0 TO D7 DO 
         RegSlave [Reg].InUse := FALSE
      END
   END ResetUseBits;


PROCEDURE MarkOnStack (reg : REGISTER);
   BEGIN
      RegSlaveIsPurged := FALSE;
      RegSlave [reg].State := OnStack
   END MarkOnStack;


PROCEDURE PurgeRegSlaves (AfterProcCall : BOOLEAN);
   BEGIN
      PurgeDataRegSlave ();
      PurgeAddrRegSlave (AfterProcCall);
   END PurgeRegSlaves;


PROCEDURE PurgeDataRegSlave ();
   VAR Reg : REGISTER;
   BEGIN
      IF NOT RegSlaveIsPurged THEN
         FOR Reg := D0 TO D7 DO
            WITH RegSlave [Reg] DO
               State := Free; InUse := FALSE
            END
         END
      END;
      RegSlaveIsPurged := TRUE;
   END PurgeDataRegSlave;


PROCEDURE PurgeAfterStackStore (VAR item : ITEM);
   BEGIN
      PurgeDataRegSlave ();
      PurgeVariablesInAddrRegSlave ()
   END PurgeAfterStackStore;


BEGIN
   AnyUseBitsSet := FALSE;
   RegSlaveIsPurged := FALSE;
   PurgeRegSlaves (TRUE);
END RegSlv.
