(*
  Title:	GenGraph
  $Header: GenGraph.mod,v 1.2 84/12/04 12:39:22 mjj Exp $
  		Copyright (C) 1984 by Acorn Research Centre
*)

IMPLEMENTATION MODULE GenGraph;

IMPORT Strings;
IMPORT HashStrings;
IMPORT Slist;

FROM Storage IMPORT ALLOCATE;
FROM SysStreams IMPORT sysOut;
FROM WriteF IMPORT WriteF0, WriteF1, WriteF2, WriteF3, H, S;
FROM TextIO IMPORT WriteCHAR;
IMPORT StringF;
FROM UtilError IMPORT ReportError;

FROM Dir IMPORT
  NullDir, DirHandle, DirExists, 
  OpenDir, NextDirEntry, CloseDir;
FROM M2Lex IMPORT
  Symbol, LStream, OpenStream, CloseStream,
  NextSymbol, Accept, AcceptId, Skip, LineNumber;
FROM OSNames IMPORT IsModuleName, GenFileName, GenPathName;

CONST
  Spaces = 4;
  HighChar = 255;

VAR
  indentation:    INTEGER;            (* for pretty reporting *)


PROCEDURE Indent;
  VAR i: INTEGER;
  BEGIN
    FOR i := 1 TO indentation DO WriteCHAR (sysOut, " ") END;
  END Indent;

PROCEDURE AddModule(d: Directory; file, m2mod: ARRAY OF CHAR; c: Class);
  VAR
    m2: M2Module; is: ModuleInstance;
    h: HashStrings.HashId;
  BEGIN
    (* Checks whether a module of this name has already been encountered.
       If not, creates the 'M2Module' record.
       Then creates a 'ModuleInstance' record and appends it to the
       'instances' list.
       *)
    IF HashStrings.EnterChars(moduleTable, m2mod, h) THEN
      (* new module *)
      NEW (m2);
      WITH m2^ DO
        moduleName := ModuleName(h);
        importedBy := NIL;
        instances := NIL;
        (* record the 'M2Module' record in the hashtable *)
        HashStrings.Assoc(moduleTable, h, m2);
        Slist.AddE(moduleList, m2);
      END;
    ELSE
      m2 := M2Module(HashStrings.Retrieve(moduleTable, h));
    END;
    
    NEW(is);
    Slist.AddE(m2^.instances, is);
    WITH is^ DO
      fileName := Strings.CopyCS (file);
      directory := d;
      m2Module := m2;
      class := c;
      imports := NIL;
      ignore := FALSE;
      marked := FALSE;
    END;
  END AddModule;
  
PROCEDURE AddImport(s: LStream; is: ModuleInstance; name: ARRAY OF CHAR);
  VAR
    ip: Imports;
    ipBy: ImportedBy;
    ipM2Module: M2Module;
    h: HashStrings.HashId;
    errMess: ARRAY [0..HighChar] OF CHAR;
  BEGIN
    IF HashStrings.LookupChars(moduleTable, name, h) THEN
      ipM2Module := M2Module(HashStrings.Retrieve(moduleTable, h));
      (* check not already imported *)
      ip := is^.imports;
      WHILE ip # NIL DO
        IF ip^.m2Module = ipM2Module THEN
          RETURN
        ELSE
          ip := ip^.next;
        END;
      END;
      (* so its new *)
      NEW(ip); NEW(ipBy);
      (* add the imported module to the 'imports' list.
         Also add is^.m2Module to the 'importedBy' list of
         the imported module
      *)
      ip^.m2Module := ipM2Module;
      Slist.AddE(is^.imports, ip);
      ipBy^.importingModule := is^.m2Module;
      Slist.AddE(ipM2Module^.importedBy, ipBy);
    ELSE
    (* did'nt find the module, at least in source form.
    Warn the user, but watch for SYSTEM.
    *)
      IF NOT Strings.EqualCC("SYSTEM", name) THEN
        StringF.StringF1(errMess,
         "imported module %s not found (in source form)",
         H(name));
        ReportSyntaxError(is, LineNumber(s), errMess);
      END;
    END;
  END AddImport;

PROCEDURE ParseImports (is: ModuleInstance; s: LStream);
  VAR
    name: ARRAY [0..HighChar] OF CHAR;
  BEGIN
    LOOP
      CASE NextSymbol (s, name) OF
        Fromsym:
        IF AcceptId (s, name) THEN
          AddImport(s, is, name);
          Skip (s, Semicolonsym);
        END;
      | Importsym:
          REPEAT
            IF AcceptId (s, name) THEN AddImport(s, is, name) END;
          UNTIL NOT Accept (s, Commasym)
      | Semicolonsym:
      ELSE EXIT
      END;
    END;
  END ParseImports;

PROCEDURE ReportSyntaxError (is: ModuleInstance; l: CARDINAL; 
    msg: ARRAY OF CHAR);
  VAR
    path: ARRAY [0..HighChar] OF CHAR;
  BEGIN
    GenPathName (is^.directory^.pathName, is^.fileName, path);
    WriteF3 (sysOut, "%s, line %u:  %s\n", H(path), l, H(msg));
    is^.ignore := TRUE;
  END ReportSyntaxError;

PROCEDURE ParseInstance (is: ModuleInstance; s: LStream);
  VAR
    name: ARRAY [0..HighChar] OF CHAR; 
    quit: BOOLEAN;
    
    PROCEDURE SyntaxError(msg: ARRAY OF CHAR);
      BEGIN
        ReportSyntaxError(is, LineNumber(s), msg);
        quit := TRUE;
      END SyntaxError;
      
  BEGIN
    quit := FALSE;
    CASE NextSymbol (s, name) OF
      Implementationsym:
        IF is^.class = Def THEN
          SyntaxError("not an implementation module");
        ELSE
          IF NOT Accept (s, Modulesym) THEN
            SyntaxError ("incorrect module header");
          END;
        END;
    | Definitionsym:
        is^.class := Def;
        IF is^.class # Def THEN
          SyntaxError("not a definition module");
        ELSE
          IF NOT Accept (s, Modulesym) THEN
            SyntaxError ("incorrect module header");
          END;
        END;
    | Modulesym:
        is^.class := Prog;
        IF is^.class = Def THEN
          SyntaxError("not an implementation module");
        END;
    ELSE SyntaxError ("missing module header");
    END;
    
    IF quit THEN RETURN END;

    IF AcceptId (s, name) AND Accept (s, Semicolonsym) THEN
        IF is^.class = Impl THEN
        (* all implementations implicitly import their definitions *)
          AddImport(s, is, name);
        END;
        ParseImports (is, s);
    ELSE SyntaxError ("missing module name");
    END;
  END ParseInstance;

PROCEDURE ParseModules;
  VAR
    m2: M2Module; is: ModuleInstance;
    s: LStream; path: ARRAY [0..HighChar] OF CHAR;
  BEGIN
    m2 := moduleList;
    WHILE m2 # NIL DO
      is := m2^.instances;
      WHILE is # NIL DO
        IF is^.directory^.toParse THEN
          GenPathName (is^.directory^.pathName, is^.fileName, path);
          IF OpenStream (s, path) THEN
            IF verbose THEN
              Indent; WriteF1 (sysOut, "parsing %s\n", H(path))
            END;
            INC (indentation, Spaces);
            ParseInstance (is, s);
            CloseStream (s);
            DEC (indentation, Spaces);
          ELSE
            WriteF1 (sysOut, "unable to open %s\n", H(path));
          END
        END;
        is := is^.next;
      END;
      m2 := m2^.next;
    END;
  END ParseModules;

PROCEDURE FindDirectory (d: Directory; VAR pathName: ARRAY OF CHAR): BOOLEAN;
(* Checks whether 'name' exists, and then checks whether we
have it under another name (for systems with aliases).
Returns TRUE if found and sets 'd' to the Directory record.
*)
  VAR 
    xuniqId: INTEGER;
    xd: Directory;
  BEGIN
    Strings.CopySC(d^.pathName, pathName);
    IF DirExists(pathName, xuniqId) THEN
      d^.uniqId := xuniqId;
      xd := directoryList;
      WHILE xd # NIL DO
        IF (xd # d) AND (xd^.uniqId = xuniqId) THEN
          RETURN FALSE; (* so it appears not to exist *)
        END;
        xd := xd^.next
      END;
      RETURN TRUE;
    ELSE 
      DirectoryError (pathName, "can't find");
      RETURN FALSE
    END;
  END FindDirectory;


PROCEDURE DirectoryError (name, error: ARRAY OF CHAR);
  BEGIN
    ReportError(FALSE, "%s directory %s", H(error), H(name));
  END DirectoryError;


PROCEDURE ScanDirectory (d: Directory);
(* Scans the files in directory 'name', adding those
which are Modula-2 source files to the 'instances' list
of the 'M2Module' record.
Nothing is done if the directory
has already been scanned.
*)
  VAR 
    dir: DirHandle; 
    name, m2mod: ARRAY [0..HighChar] OF CHAR;
    class: Class;
  BEGIN
    IF FindDirectory (d, name) THEN
      IF NOT d^.scanned THEN
        d^.scanned := TRUE;
        dir := OpenDir (name);
        IF NullDir(dir) THEN
          DirectoryError (name, "can't open");        
        ELSE 
          IF verbose THEN Indent; WriteF1(sysOut,"scanning %s\n", H(name)) END;
          INC (indentation, Spaces);
          WHILE NextDirEntry (dir, name) DO
            IF IsModuleName (name, m2mod, class) THEN
                AddModule (d, name, m2mod, class)
            END
          END;
          CloseDir (dir);
          DEC (indentation, Spaces);
        END
      END
    END;
  END ScanDirectory;

PROCEDURE ScanDirectories;
  VAR
    d: Directory;
  BEGIN
    d := directoryList;
    WHILE d # NIL DO
      ScanDirectory(d);
      d := d^.next;
    END;
  END ScanDirectories;

PROCEDURE Build;
  BEGIN
    ScanDirectories;
    ParseModules;
  END Build;

BEGIN
  moduleList := NIL;
  directoryList := NIL;
  moduleTable := HashStrings.New(128);
  parseRemote := FALSE;
  verbose := FALSE;
END GenGraph.

(*
$Log:	GenGraph.mod,v $
Revision 1.2  84/12/04  12:39:22  mjj
Error messages in standard (emacs) form
Added message about missing modules.
Only add an imported module once to the 'imports' list

Revision 1.1  84/11/30  16:06:32  mjj
Initial revision

*)

