(*
    Title: 	Modula2 pretty printer
    LastEdit:	"Mon Oct 29 15:01:51 1984"
    Author: 	Steven Glassman
		Acorn Research Centre
*)

MODULE m2pretty;

IMPORT TextIO, HashStrings, Strings;
IMPORT Streams, SysStreams, UnixArgs, M2Files, UtilError;

FROM Streams IMPORT
    ErrorCode, Stream;
FROM Strings IMPORT
    String;
FROM SysStreams IMPORT
    sysErr, sysIn, sysOut;
FROM CharCodes IMPORT
    NewLineCh, SpaceCh, TabCh;

MODULE Arguments;
(* a nested module, which knows how to decode arguments and
open streams.  The module defines the variable 'inStream'
to be used by the outside world.
*)
  IMPORT Streams, SysStreams, UnixArgs, M2Files, UtilError;
  
  EXPORT HandleArgument, inStream;

  VAR inStream: Streams.Stream;   (* the input stream to be used *)
  
  PROCEDURE HandleArgument;
    VAR
      arg: ARRAY [0..255] OF CHAR;
      ext: M2Files.Extension;
    BEGIN
      IF UnixArgs.argCount > 1 THEN
        UnixArgs.GetArg(1, arg);
        IF arg[0] = '-' THEN
          IF arg[1] = 'd' THEN
            IF UnixArgs.argCount > 2 THEN
              UnixArgs.GetArg(2, arg);
              ext := "def";
            ELSE
              UtilError.ReportError(TRUE, "module name missing",0,0);
            END;
          ELSE
            UtilError.ReportError(TRUE, "bad option",0,0);
          END;
        ELSE
          ext := "mod";
        END;
        inStream := M2Files.CreateInput(arg, ext);
        IF inStream^.status # Streams.Success THEN
          UtilError.ReportError(TRUE, "failed to open source module",0,0);
        END;
      ELSE
        inStream := SysStreams.sysIn;
      END;
      
    END HandleArgument;
  BEGIN
    UtilError.SetPrompt("m2pretty");
  END Arguments;


(*  The following Modula keywords affect the nesting level of the program
   as shown by the actions they cause to get stacked up.  At the start of
   the next line, all of the stacked actions get applied.
   Note that a FROM will cancel an IMPORT.

   BEGIN                   futureUnindent indentNext
   CASE                    indentNext
   CONST                   futureUnindent indentNext unindentCurrent
   DEFINITION              don't indent PROCEDUREs
   DO                      endStatement
   ELSE                    futureUnindent indentNext unindentCurrent
   ELSIF                   futureUnindent indentNext unindentCurrent
   END                     unindentCurrent
   EXCEPT                  indentNext
   EXPORT                  futureUnindent indentNext unindentCurrent
   FINALLY                 futureUnindent indentNext unindentCurrent
   FOR                     indentNext
   FROM                    futureUnindent indentNext unindentCurrent from
   IF                      indentNext
   IMPORT                  futureUnindent indentNext unindendCurrent
   LOCK                    indentNext
   LOOP                    indentNext
   OF                      endStatement
   PROCEDURE               indentNext
   RAISING                 futureUnindent indentNest unindentCurrent
   RECORD                  indentNext
   REPEAT                  indentNext
   THEN                    endStatement
   TRY                     indentNext
   TYPE                    futureUnindent indentNext unindentCurrent
   TYPECASE                indentNext
   UNTIL                   unindentCurrent
   VAR                     futureUnindent indentNext unindentCurrent
   WHILE                   indentNext
   WITH                    indentNext
   (*                      indentNext
   *)                      unindentNext
   | (case delimiter)      indentNext unindentCurrent
   ;                       endStatement

   Tokens are delimeted by space, semicolon and end of line.

   The system will maintain a stack of nesting levels which will be kept
   current as tokens are found.  At the start of each line, the FIRST token
   found will determine whether to use the current or previous nesting level
   for the line.  At the END of each line the last new nesting
   change found will determine the action to occur on the next line.
   *)

TYPE
    (* The token actions are as follows:
       1: beginBlock (FOR, IF, LOOP, etc) starts a new block
       indent the next line
       2: declarationBlock (VAR, CONST, TYPE, etc) forms a block by itself
       unindent the current line, indent the next line, do a future
       unindent on the next block
       3: endBlock (END, UNTIL) ends the current block
       unindent the next line
       4: nestBlock (ELSE, ELSIF) middle of a block
       unindent current line, indent next line
       5: from block (FROM) a declaration block, but since it is a from it
       cancels the follwing IMPORT
       *)
    TokenAction = (BeginBlock, DeclarationBlock, EndBlock, NestBlock,
      FromBlock, CommentOpen, CommentClose, EndStatement, NoChange);

    ActionType = (UnindentCurrent, IndentNext, UnindentNext, NullAction);

    NestLevel = CARDINAL;
    
    TokenType = (TBegin, TCase, TConst, TDef, TDo, TElse, TEnd, TExcept, 
      TExport, TFinal, TFor, TFrom, TIf, TImport, TLock, TLoop, TMod, TOf, 
      TProc, TRaise, TRec, TRep, TThen, TTry, TType, TTypeCase, TUntil, TVar,
      TWhile, TWith, TOpen, TClose, TVBar, TSemi, TOther, TEOF);

    StackLevel = RECORD
        tT: TokenType;
        nest: NestLevel;
    END;

    NestAction = RECORD
        action: ActionType;
        tT: TokenType;
        nest: NestLevel;
    END;

    NestTokenAction = RECORD
        action: TokenAction;
        nest: NestLevel;
        END;

    ActionIndex = [0..20];

CONST
    TabWidth = 4;
    NoNest = 0;
    CommentNest = 3;
    NormalNest = 2;
    HalfNest = NormalNest DIV 2;
    CaseNest = NormalNest;
    DoubleCaseNest = 2 * CaseNest;
    MaxNest = 400;

    ChNull = 0C;
    
VAR
    (* levelStack and currentNest are used to form a stack of nest levels
       pushing a nest level means incrementing currentNest and storing the
       new nest level in levelStack[currentNest]
       popping a nest level means decrementing currentNest
       the current nest level is levelStack[currentNest]
       *)
    levelStack: ARRAY [0..MaxNest] OF StackLevel;
    currentNest: CARDINAL;

    columnCount: CARDINAL;
    chars: ARRAY [0..1] OF CHAR; (* special to make 1 char an array of char *)

    (* countNewLines and leading spaces are set up in GetToken
       and are used to determine where the line breaks occur and how
       many leadingSpaces there are for a given token.  At the start
       of a line, leadingSpaces is ignored and the current nesting level is
       used, but inside the line, leadingSpaces is used
       *)
    countNewLines: CARDINAL;
    countComments: CARDINAL;
    leadingSpaces: CARDINAL;
    fFirstToken: BOOLEAN;

    (* The hashTable holds the Modula2 keywords, each of which is associated
       with a tokenType which is retrieved from the hashTable.  The tokeType
       is used to index into the tokenToNestAction array in order to get the
       nesting change and action to perform for the given token
       *)
    hashTable: HashStrings.HashTable;
    hashID, procHashID: HashStrings.HashId;
    tokenToNestAction: ARRAY[TBegin..TOther] OF NestTokenAction;

    (* actionStack is used to keep track of the pending actions of the
       current line being parsed *)
    actionStack: ARRAY ActionIndex OF NestAction;
    actionHead, actionTail: ActionIndex;

    (* distinguish between the two kinds of future unindents.  Some like:
       VAR, CONST, EXPORT, etc. unindent of any kind of keyword.  Others
       like: ELSE, ELSIF, | only unindent on an END 
       Now: never mind.  END's always unindent *)
    fFutureUnindent: BOOLEAN;

(* Get the next token from the stream h.  Tokens are delimeted by the
   characters space, semicolon and end of line
   Also return the type of the token.
   *)
PROCEDURE GetToken(h: Stream; VAR s: String): TokenType;
VAR
    firstChar: BOOLEAN;
    ch: CHAR;
    tT: TokenType;
BEGIN
    countNewLines := 0;
    leadingSpaces := 0;

    (* Skip any leading white space, semicolons don't get eaten since they
       must get echoed out to the file *)
    REPEAT
        ch := CHAR(Streams.Get(h));
        IF ch = NewLineCh THEN
            leadingSpaces := 0;
            INC(countNewLines);
        ELSIF ch = TabCh THEN
            INC(leadingSpaces, columnCount MOD TabWidth);
        ELSIF ch = SpaceCh THEN
            INC(leadingSpaces);
        END;
        UNTIL ((ch # SpaceCh) AND (ch # NewLineCh) AND (ch # TabCh));

    (* Suck up the string until the next delimiter *)
    firstChar:= TRUE;
    s^[0] := ChNull;
    
    LOOP
        chars[0] := ch;
        IF h^.status = Streams.EndOfStream THEN
            IF firstChar THEN
                RETURN(TEOF);
            ELSE
                (* End the token here including the last character
                   and report tEOF next time thru *)
                s := Strings.ConcatSC(s, chars);
                EXIT
            END;
        ELSIF ((ch = SpaceCh) OR (ch = TabCh)) THEN
            (* Spaces and tabs are NEVER a token or part of one, so they
               always get put back and end the token.  NOTE: a space or
               tab can NOT be the first character, since the first loop
               checks for this
               *)
            Streams.Backspace(h);
            EXIT
        ELSIF ((ch = ";") OR (ch = NewLineCh)) THEN
            (* Starting the token with a semicolon or newline is just fine.
               This character IS the token.  If the semicolon or newline
               follows any other characters then they end the token and are
               put back for the next time around.
               *)
            IF NOT firstChar THEN
                Streams.Backspace(h);
            ELSE
                s := Strings.ConcatSC(s, chars);
            END;
            EXIT
        END;

        firstChar := FALSE;
        s := Strings.ConcatSC(s, chars);
        ch := CHAR(Streams.Get(h));
        END;

    (* Look up the string to see if it is one of our key words *)
    IF HashStrings.LookupString(hashTable, s, hashID) THEN
        tT := TokenType(HashStrings.Retrieve(hashTable, hashID));
    ELSE
        RETURN(TOther);
    END;

    RETURN(tT);
END GetToken;

PROCEDURE CurrentNest():NestLevel;
BEGIN
    RETURN(levelStack[currentNest].nest);
END CurrentNest;

PROCEDURE PopNest():TokenType;
BEGIN
    IF currentNest > 0 THEN 
        DEC(currentNest); 
        (* vague cheat - return the tT of nest level we just popped off *)
        RETURN(levelStack[currentNest+1].tT);
    ELSE
        RETURN(TOther);
    END;
END PopNest;

PROCEDURE PushNest(n: NestLevel; tT: TokenType);
BEGIN
    IF currentNest = MaxNest-1 THEN
        TextIO.WriteChars(sysErr, 'Too many levels, need more "END"s');
    ELSE
        (* store current nest level + new indent into
           nestlevels[currentNest + 1]
           *)
        INC(n, levelStack[currentNest].nest);
        INC(currentNest);
        levelStack[currentNest].nest := n;
        levelStack[currentNest].tT := tT;
    END;
END PushNest;

PROCEDURE IncActionIndex(i: ActionIndex): ActionIndex;
BEGIN
    INC(i);
    IF i = HIGH(actionStack) THEN
        i:= 0;
        END;
    RETURN(i);
END IncActionIndex;

PROCEDURE DecActionIndex(i: ActionIndex): ActionIndex;
BEGIN
    IF i = 0 THEN
        i := HIGH(actionStack);
        END;
    RETURN(i-1);
END DecActionIndex;

(* Tell if there were actions to pop *)
PROCEDURE PopAction(VAR a: ActionType; VAR t: TokenType;
  VAR n: NestLevel): BOOLEAN;
BEGIN
    IF actionTail # actionHead THEN
        WITH actionStack[actionTail] DO
            n := nest;
            a := action;
            t := tT;
            END;
        actionTail := IncActionIndex(actionTail);
        RETURN(TRUE);
    ELSE
        a := NullAction;
        t := TOther;
        RETURN(FALSE);
    END;
END PopAction;

PROCEDURE PushAction(a: ActionType; t: TokenType; n: NestLevel);
BEGIN
    IF IncActionIndex(actionHead) = DecActionIndex(actionTail) THEN
        TextIO.WriteChars(sysErr, 'Too much stuff on a line');
    ELSE
        WITH actionStack[actionHead] DO
            nest := n;
            action := a;
            tT := t;
        END;
        actionHead := IncActionIndex(actionHead);
        END;
END PushAction;

PROCEDURE CreateEntry(s: ARRAY OF CHAR; tT: TokenType; nestEntry: NestLevel;
actionEntry: TokenAction);
VAR
    f: BOOLEAN;
BEGIN
    f:= HashStrings.EnterChars(hashTable, s, hashID);
    HashStrings.Assoc(hashTable, hashID, tT);
    WITH tokenToNestAction[tT] DO
        nest := nestEntry;
        action := actionEntry;
        END;
END CreateEntry;

PROCEDURE InitPP;
VAR
    f: BOOLEAN;
BEGIN

    columnCount := 0;
    currentNest := 0;
    actionHead := 0;
    actionTail := 0;
    fFutureUnindent := FALSE;
    fFirstToken := TRUE;

    (* Allocate the hash table 25% larger than hash entries
       Enter each of the magic strings and initialize their array values *)
    hashTable := HashStrings.New(ORD(TOther)*5 DIV 4);
    
    CreateEntry("BEGIN",      TBegin, NormalNest, BeginBlock);
    CreateEntry("CASE",       TCase,  CaseNest,   BeginBlock);
    CreateEntry("CONST",      TConst, NormalNest, DeclarationBlock);
    CreateEntry("DO",         TDo,    NoNest,     EndStatement);
    CreateEntry("DEFINITION", TDef,   NoNest,     NoChange);
    CreateEntry("ELSE",       TElse,  NormalNest, NestBlock);
    CreateEntry("ELSIF",      TElse,  NormalNest, NestBlock);
    CreateEntry("END",        TEnd,   NormalNest, EndBlock);
    CreateEntry("EXCEPT",     TExcept,CaseNest,   BeginBlock);
    CreateEntry("EXPORT",     TExport,NormalNest, BeginBlock);
    CreateEntry("FINALLY",    TFinal, NormalNest, NestBlock);
    CreateEntry("FOR",        TFor,   NormalNest, BeginBlock);
    CreateEntry("FROM",       TFrom,  NormalNest, FromBlock);
    CreateEntry("IF",         TIf,    NormalNest, BeginBlock);
    CreateEntry("IMPORT",     TImport,NormalNest, DeclarationBlock);
    CreateEntry("LOCK",       TLock,  NormalNest, BeginBlock);
    CreateEntry("LOOP",       TLoop,  NormalNest, BeginBlock);
    CreateEntry("MODULE",     TMod,   NoNest,     BeginBlock);
    CreateEntry("OF",         TOf,    NoNest,     EndStatement);
    CreateEntry("RECORD",     TRec,   NormalNest, BeginBlock);
    CreateEntry("PROCEDURE",  TProc,  NormalNest, BeginBlock);
    CreateEntry("RAISING",    TRaise, NormalNest, NestBlock);
    CreateEntry("REPEAT",     TRep,   NormalNest, BeginBlock);
    CreateEntry("TRY",        TTry,   NormalNest, BeginBlock);
    CreateEntry("TYPE",       TType,  NormalNest, DeclarationBlock);
    CreateEntry("TYPECASE",   TTypeCase, CaseNest, BeginBlock);
    CreateEntry("THEN",       TThen,  NoNest,     EndStatement);
    CreateEntry("UNTIL",      TUntil, NormalNest, EndBlock);
    CreateEntry("VAR",        TVar,   NormalNest, DeclarationBlock);
    CreateEntry("WHILE",      TWhile, NormalNest, BeginBlock);
    CreateEntry("WITH",       TWith,  NormalNest, BeginBlock);
    CreateEntry("(*",         TOpen,  CommentNest,CommentOpen);
    CreateEntry("*)",         TClose, NoNest,     CommentClose);

    (* Bogus string for TOther, but action/nest is correct *)
    CreateEntry("1xyz",       TOther, NoNest,     NoChange);

    chars[0] := "|";
    CreateEntry(chars,        TVBar,  DoubleCaseNest, NestBlock);

    chars[0] := ";";
    CreateEntry(chars,        TSemi,  NoNest,     EndStatement);

END InitPP;


PROCEDURE HandleActionStack(fUnindent: BOOLEAN): TokenType;
VAR
    topOfStack: BOOLEAN;
    stackAction: ActionType;
    stackNest: NestLevel;
    tAction, tNest: TokenType;
BEGIN
    topOfStack:= TRUE;
    tNest := TOther;

    WHILE PopAction(stackAction, tAction, stackNest) DO

        CASE stackAction OF
          IndentNext:                     PushNest(stackNest, tAction);
        | UnindentNext, UnindentCurrent:  tNest := PopNest();
        ELSE
        END; (* CASE *)

        topOfStack := FALSE;
        END; (* WHILE *)

    IF fUnindent THEN
        tNest := PopNest();
        fFutureUnindent := FALSE;
        END; (* IF unindent *)

    RETURN(tNest);
END HandleActionStack;

PROCEDURE InProc(): BOOLEAN;
VAR
    tT: TokenType;
BEGIN
    tT := actionStack[actionTail].tT;
    RETURN(tT = TProc);
END InProc;

PROCEDURE GoOverFile;
VAR
    fPendingDefModule, fModuleFound: BOOLEAN;
    fNoIndentIfNoSemi: BOOLEAN;
    fUnindentedCurrent: BOOLEAN;
    fromFound: BOOLEAN;
    i: CARDINAL;
    tokenType, tLastPop: TokenType;
    token: String;
BEGIN
    (* Now the real fun begins.  Until the end of the file do:
       1: If countnewLines > 0 then
          a) apply pending actions to level stack
          b) set leadingSpaces to current nest level
       2: Output leading spaces worth of spaces
       3: Output the token
       4: Save the token's action and nesting info for next line
       5: Get next token
       *)
    fPendingDefModule := FALSE;
    fModuleFound := FALSE;
    fromFound := FALSE;
    fNoIndentIfNoSemi := TRUE;
    countComments := 0;
    token := Strings.New(0);
    tokenType := GetToken(inStream, token);
    tLastPop := TOther;
    
    IF fFirstToken THEN
        PushNest(leadingSpaces, TOther);
        fFirstToken := FALSE;
        END;

    WHILE tokenType # TEOF DO
        fUnindentedCurrent := FALSE;
        
        (* Definitions modules DON'T indent PROCEDUREs while normal
           modules do.  SO if see a DEFINITON before MODULE then
           set procedure indents to 0
        *)
        IF NOT fModuleFound THEN
            IF tokenType = TMod THEN
                fModuleFound := TRUE;
                IF fPendingDefModule THEN
                    tokenToNestAction[TProc].nest := NoNest;
                END;
            ELSE
                fPendingDefModule := (tokenType = TDef);
            END;
        END;
        
        IF countNewLines > 0 THEN
            FOR i := 1 TO countNewLines DO
                TextIO.NewLine(sysOut);
                END;

            (* Unindent if not in a comment, not starting a comment
               but starting a nest block or starting a line with a keyword *)
            WITH tokenToNestAction[tokenType] DO
                fUnindentedCurrent :=
                  (countComments = 0) AND
                  (
                  (action = NestBlock) OR 
                  ((action = DeclarationBlock) AND (NOT InProc())) OR
                  (action = FromBlock) OR 
                  (action = EndBlock) OR
                  (
                  (fFutureUnindent) AND 
                  (tokenType # TOther) AND
                  (action # CommentOpen) AND 
                  (action # CommentClose)
                  )
                  );
                END;

            (* Under the new SRC nesting scheme, PROCEDUREs and MODULEs
               form a nesting level which gets canceled by the final END.
               Unfortunately, the only way to tell this special END is to 
               see if it matches a BEGIN and if it does, then pop off the
               nest for the PROECDURE or MODULE *)
            tLastPop := HandleActionStack(fUnindentedCurrent);
            IF (tokenType = TEnd) AND (tLastPop = TBegin) THEN
                fFutureUnindent := TRUE;
            END;
            
            leadingSpaces := CurrentNest();
            IF (NOT fNoIndentIfNoSemi) AND (countComments = 0) AND
                  (tokenType = TOther) THEN
                INC(leadingSpaces, HalfNest);
                END;
            fNoIndentIfNoSemi := FALSE;
            END; (* starting new line *)

        FOR i:= 1 TO leadingSpaces DO
            TextIO.WriteCHAR(sysOut, SpaceCh);
            END;

        (* Can't use WriteString since it interprets the "*X"s in the
           string as magic characters *)
        FOR i:= 1 TO Strings.LengthS(token) DO
            TextIO.WriteCHAR(sysOut, token^[i-1]);
            END;

        (* Only update the pending nest and action on an interesting token
           Don't have to worry about current unindent since HandleActionStack
           does it *)
        IF (countComments = 0) OR
              (tokenType = TOpen) OR (tokenType = TClose) THEN
            WITH tokenToNestAction[tokenType] DO
                CASE action OF
                  BeginBlock:
                      PushAction(IndentNext, tokenType, nest);
                      (* Make BEGIN, LOOP, REPEAT, RECORD self ending *)
                      CASE tokenType OF
                        TTry:
                          fNoIndentIfNoSemi := TRUE;
                          fFutureUnindent := TRUE;
                        | TBegin, TLoop, TRep, TRec: 
                          fNoIndentIfNoSemi := TRUE;
                        ELSE
                        END;
                  | NestBlock:
                      (* ELSE/ELSIF only matter when at start of line 
                         Otherwise don't do anything since unindent/indent
                         cancels out *)
                      IF fUnindentedCurrent THEN
                          PushAction(IndentNext, tokenType, nest);
                          fFutureUnindent := FALSE;
                          fNoIndentIfNoSemi := TRUE;
                      END;
                  | FromBlock:
                      PushAction(IndentNext, tokenType, nest);
                      fFutureUnindent := TRUE;
                      fromFound := TRUE;
                      fNoIndentIfNoSemi := TRUE;
                  | DeclarationBlock:
                      IF (tokenType # TImport) OR (NOT fromFound) THEN
                          PushAction(IndentNext, tokenType, nest);
                          fFutureUnindent := TRUE;
                          fromFound := FALSE;
                          fNoIndentIfNoSemi := TRUE;
                      END;
                  | EndBlock:
                      (* EndBlocks now UnindentCurrent, but they may not
                         start the line, so must UnindentNext then *)
                      IF (NOT fUnindentedCurrent) THEN
                          PushAction(UnindentNext, tokenType, nest);
                          END;
                  | CommentOpen:
                      IF countComments = 0 THEN
                          PushAction(IndentNext, tokenType, nest);
                      END;
                      INC(countComments);
                  | CommentClose:
                      fNoIndentIfNoSemi := TRUE;
                      IF countComments > 0 THEN
                          DEC(countComments);
                      END;
                      IF countComments = 0 THEN
                          PushAction(UnindentNext, tokenType, nest);
                          END;
                  | EndStatement:
                      fNoIndentIfNoSemi := TRUE;
                  | NoChange:
                  END;
                END;
            END;

        tokenType := GetToken(inStream, token);
        END;

    Strings.Dispose(token);
    FOR i := 1 TO countNewLines DO
        TextIO.NewLine(sysOut);
        END;
END GoOverFile;

BEGIN
    InitPP();
    HandleArgument;
    GoOverFile();

END m2pretty.
