(*
  Title:	GenMakeFile
  $Header: GenMakeFile.mod,v 1.4 85/01/04 10:39:12 mjj Exp $
  		Copyright (C) 1984 by Acorn Research Centre
*)

IMPLEMENTATION MODULE GenMakeFile;
FROM SYSTEM IMPORT ADR;
FROM Storage IMPORT ALLOCATE;
FROM Streams IMPORT Stream, Seek, Delete, ErrorCode, Put, SeekMode;
IMPORT FileStream;
IMPORT PageStream; 
IMPORT Strings;
IMPORT HashStrings;
IMPORT Slist;
FROM UtilError IMPORT ReportError;
FROM TextIO IMPORT WriteChars, NewLine, WriteCHAR;
FROM WriteF IMPORT WriteF0, WriteF1, WriteF2, WriteF3, H, S;
FROM CharCodes IMPORT SpaceCh;

FROM OSNames IMPORT GenFileName, GenPathName;
FROM GenGraph IMPORT 
  ModuleInstance, Imports, Directory, Class, M2Module,
  moduleList, directoryList, moduleTable;

IMPORT AppendUser;

TYPE
  Target = POINTER TO TargetData;
  TargetData = RECORD
      next: Target;
      instance: ModuleInstance;
    END;

  ClassSet = SET OF Class;

CONST
   TwoTabs = 17; Continuation = 3;
   MaxLineLength = 80; MaxPageLength = 32767;
   ImplOrProg = ClassSet{Impl,Prog};

VAR
  s: Stream; 
  targets: ARRAY Class OF Target;
  optionName: ARRAY Class OF Strings.String;
  classListName: ARRAY Class OF Strings.String;
  utilityId: ARRAY Class OF CHAR;

PROCEDURE Build;
  VAR
    fS: Stream;
  BEGIN
    IF AppendUser.noAppend THEN
      fS := FileStream.CreateOutput(makeFilename)
    ELSE
      fS := FileStream.CreateInOutput(makeFilename);
      Seek(fS, End, 0)
    END;
    IF fS^.status # Success THEN
      ReportError(TRUE, "failed to open makefile %s", H(makeFilename),0);
    END;
    s := PageStream.CreateOutput (fS, MaxLineLength, MaxPageLength, 
            TRUE, TRUE);
    
    WritePreamble;
    targets[Sym] := ConsiderTargets(ClassSet{Def}, Sym);
    targets[Obj] := ConsiderTargets(ImplOrProg, Obj);
    targets[Exe] := ConsiderTargets(ClassSet{Prog}, Exe);

    (* full target is for programs, objects and symbols *)
    WriteChars(s, "library:"); Indent(TwoTabs);
    WriteF3(s, "%s %s %s\n", S(classListName[Obj]),
        S(classListName[Impl]), S(classListName[Def]));
    Indent(TwoTabs);
    WriteF0(s, "echo library up to date\n\n");
    
    GenTargetList(Obj, targets[Exe]);
    GenTargetList(Def, targets[Sym]);
    GenTargetList(Impl, targets[Obj]);
    
    GenTargets(Obj,targets[Exe] );
    GenTargets(Impl,targets[Obj] );
    GenTargets(Def, targets[Sym]);
    
    Delete(s);
    Delete(fS);
  END Build;

PROCEDURE ConsiderTargets(classSet: ClassSet; targetClass: Class): Target;
  VAR
    m2: M2Module;
    is: ModuleInstance;
    target: Target;
    targets: Target;
  BEGIN
    targets := NIL;
    m2 := moduleList;
    WHILE m2 # NIL DO
      is := m2^.instances;
      WHILE is # NIL DO
        IF (is^.class IN classSet) AND MakeHere(is, targetClass) THEN
          NEW(target);
          Slist.Add(targets, target);
          target^.instance := is;
        END;
        is := is^.next;
      END;
      m2 := m2^.next;
    END;
    RETURN targets;
  END ConsiderTargets;

PROCEDURE MakeHere(is: ModuleInstance; targetClass: Class): BOOLEAN;
(* decides whether to make 'is' in the current directory.
Do so iff:
1) The object source is in a target directory OR
2) The object imports something which will be a target
*)
  VAR
    imports: Imports;
    importsIs: ModuleInstance;
  BEGIN
    IF is^.directory^.target THEN
      RETURN TRUE
    ELSE
      imports := is^.imports;
      WHILE imports # NIL DO
      (* look for an instance which will be made a target.
      We are only interested in definition modules unless
      (is^.class = Prog) AND (targetClass=Exe).
      *)
        importsIs := imports^.m2Module^.instances;
        WHILE importsIs # NIL DO
          IF ((importsIs^.class = Def) OR 
              ((is^.class=Prog) AND (targetClass=Exe))) AND
               MakeHere(importsIs, targetClass) THEN
            RETURN TRUE
          ELSE
            importsIs := importsIs^.next;
          END;
        END;
        imports := imports^.next;
      END;
      RETURN FALSE;
    END;
  END MakeHere;

PROCEDURE GenTargetList(class: Class; t: Target);
  VAR
    is: ModuleInstance;
    targetName: ARRAY [0..79] OF CHAR;
  BEGIN
    WriteF1(s, "%s:", S(classListName[class]));
    Indent(TwoTabs);
    WHILE t # NIL DO
      is := t^.instance;
      GenFileNameFromInstance(is, TRUE, CompiledFormOf(class), targetName);
      CheckFor(Strings.LengthC(targetName));
      WriteChars(s, targetName); WriteCHAR(s, SpaceCh);
      t := t^.next;
    END;
    NewLine(s); 
    Indent(TwoTabs);
    WriteF1(s, "echo %s up to date\n", S(classListName[class]));
    NewLine(s);
  END GenTargetList;

PROCEDURE GenFileNameFromInstance(is: ModuleInstance; forceHere: BOOLEAN;
    class: Class; VAR targetName: ARRAY OF CHAR);
  VAR
    pModuleName, stringCopy: Strings.String;
  BEGIN
    pModuleName := HashStrings.StringAt(moduleTable, is^.m2Module^.moduleName);
    GenFileName(pModuleName, class, targetName);
    IF NOT (is^.directory^.isCurrent OR forceHere) THEN
      stringCopy := Strings.CopyCS(targetName);
      GenPathName(is^.directory^.pathName, stringCopy,
          targetName);
      Strings.Dispose(stringCopy);
    END;
  END GenFileNameFromInstance;
  
PROCEDURE CompiledFormOf(c: Class): Class;
  BEGIN
    CASE c OF
      Def: RETURN Sym;
    | Impl, Prog: RETURN Obj;
    | Obj: RETURN Exe;
    END;
  END CompiledFormOf;

PROCEDURE DependsOnClass(c: Class): Class;
  BEGIN
    CASE c OF
      Impl, Prog, Def: RETURN Def;
    | Obj: RETURN Impl;
    END;
  END DependsOnClass;

PROCEDURE ExistsHere(is: ModuleInstance; c: Class): BOOLEAN;
(* TRUE if 'is' is in the current directory OR will
be made in the current directory.
*)
  VAR
    t: Target;
  BEGIN
    IF is^.directory^.isCurrent THEN
      RETURN TRUE;
    END;
    t := targets[c];
    WHILE t # NIL DO
      IF t^.instance = is THEN
        RETURN TRUE
      ELSE
        t := t^.next;
      END;
    END;
    RETURN FALSE;
  END ExistsHere;

PROCEDURE GenTargets(class: Class; t: Target);
  VAR
    targetName: ARRAY [0..79] OF CHAR;
    is: ModuleInstance;
    pModuleName: Strings.String;
  BEGIN
    WHILE t # NIL DO
      is := t^.instance;
      GenFileNameFromInstance(is, TRUE, CompiledFormOf(class), targetName);
      WriteF1(s, "%s:", H(targetName)); Indent(TwoTabs);
      GenFileNameFromInstance(is, ExistsHere(is, class), class, targetName);
      WriteChars(s, targetName); WriteCHAR(s, SpaceCh);
      GenDependencies(is, class);
      NewLine(s);
      Indent(TwoTabs);
      WriteF1(s, "m2%c$(system) ", utilityId[class]);
      IF class = Def THEN
        WriteChars(s, "-d ");
      END;
      pModuleName := HashStrings.StringAt(moduleTable, is^.m2Module^.moduleName);
      WriteF2(s, "$(%s) %s\n\n", S(optionName[class]), S(pModuleName));
      IF class = Obj THEN
        UnMark;
      END;
      t := t^.next;
    END;
  END GenTargets;

PROCEDURE GenDependencies(is: ModuleInstance; class: Class);
  VAR
    targetName: ARRAY [0..79] OF CHAR;
    importsIs: ModuleInstance;
    imports: Imports;
    cF: Class;
  BEGIN
      imports := is^.imports;
      WHILE imports # NIL DO
        importsIs := imports^.m2Module^.instances;
        (* find the first instance - this is what the compiler would do.
        This assumes that the instances list is ordered in the same way
        as the directoryList.
        *)
        WHILE importsIs # NIL DO
          IF (NOT importsIs^.marked) AND (importsIs^.class = DependsOnClass(class)) THEN
            cF := CompiledFormOf(importsIs^.class);
            GenFileNameFromInstance(importsIs, ExistsHere(importsIs, cF), 
                cF, targetName);
            CheckFor(Strings.LengthC(targetName));
            WriteChars(s, targetName); WriteCHAR(s, SpaceCh);
            IF class = Obj THEN
            (* a program depends on all the instances, not just those
            at the first level.
            *)
              importsIs^.marked := TRUE;    (* to suppress duplicates *)
              GenDependencies(importsIs, class);
            END;
            importsIs := NIL;
          ELSE
            importsIs := importsIs^.next;
          END;
        END;
        imports := imports^.next;
      END;
  END GenDependencies;

PROCEDURE UnMark;
  VAR
    m2: M2Module;
    is: ModuleInstance;
  BEGIN
    m2 := moduleList;
    WHILE m2 # NIL DO
      is := m2^.instances;
      WHILE is # NIL DO
        is^.marked := FALSE;
        is := is^.next;
      END;
      m2 := m2^.next;
    END;
  END UnMark;

PROCEDURE WritePreamble;
    BEGIN
        WriteF0 (s, "system =\n");
        WriteF0 (s, "compile =\n");
        WriteF0 (s, "link =\n");
        NewLine(s);
    END WritePreamble;

PROCEDURE Indent(n: CARDINAL);
    BEGIN
        IF PageStream.TellChar (s) < n THEN
            PageStream.SeekChar (s, Beginning, n);
        ELSE
            Put (s, ' ');
        END (* if *);
    END Indent;


PROCEDURE CheckFor (chars: CARDINAL);
    BEGIN
        IF PageStream.TellChar (s) + chars >= MaxLineLength - 2 THEN
            WriteF0 (s, "\\\n");
            Indent(Continuation);
        END (* if too close to the end *);
    END CheckFor;

PROCEDURE InitTargets;
  VAR
    c: Class;
  BEGIN
    FOR c := Def (* MIN(Class) *) TO Exe (* MAX(Class) *) DO
      targets[c] := NIL;
    END;
  END InitTargets;
  
BEGIN
  makeFilename := "makefile";
  utilityId[Impl] := 'c';
  utilityId[Def] := 'c';
  utilityId[Obj] := 'l';
  InitTargets;
  optionName[Impl] := Strings.CopyCS("compile");
  optionName[Def] := optionName[Impl];
  optionName[Obj] := Strings.CopyCS("link");
  classListName[Def] := Strings.CopyCS("symbols");
  classListName[Impl] := Strings.CopyCS("objects");
  classListName[Prog] := classListName[Impl];
  classListName[Obj]:= Strings.CopyCS("programs");
END GenMakeFile.

(*
$Log:	GenMakeFile.mod,v $
Revision 1.4  85/01/04  10:39:12  mjj
Modifications to reflect the 'target' and 'isCurrent' fields of
a directory entry.

Revision 1.3  84/12/04  12:40:57  mjj
Get filename generation correct in '-p' mode.

Revision 1.2  84/11/30  16:53:58  mjj
fix missing "echo" on library target

Revision 1.1  84/11/30  16:10:25  mjj
Initial revision

*)

